#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

class SidfmdVulView < ActiveRecord::Base
  has_many :sidfmd_vv_scsids, dependent: :delete_all
  has_many :sidfmd_vv_hosts, dependent: :delete_all
  has_many :sidfmd_vv_categories, dependent: :delete_all
  has_many :sidfmd_vv_due_dates, dependent: :delete_all
  has_many :sidfmd_vv_unf_due_dates, dependent: :delete_all
  has_many :sidfmd_vv_cves, dependent: :delete_all
  has_many :sidfmd_vv_ctags, dependent: :delete_all

  scope :user_vul_view, -> (is_role_admin, user_host_ids){
    unless is_role_admin then
      where(:id => SidfmdVvHost.where(:sidfmd_host_id => user_host_ids).select(:sidfmd_vul_view_id))
    end
  }

  PROC_UNIT = 50

  # 全データを同期する
  # viewのデータを全て廃棄して、現在のデータを反映する。
  # 時間がかかるので注意
  # 処理中はSidfmdVulViewでテーブルロックがかかることに注意。
  # GUI内の処理では基本的に実行しないこと（メンテナンスページを用意して実行できるようにする可能性はあり）。
  def self.sync_whole_data
    data = SidfmdIssue.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required)
    SidfmdVulView.transaction do
      SidfmdVulView.lock
      SidfmdVvCtag.delete_all
      SidfmdVvCve.delete_all
      SidfmdVvUnfDueDate.delete_all
      SidfmdVvDueDate.delete_all
      SidfmdVvCategory.delete_all
      SidfmdVvHost.delete_all
      SidfmdVvScsid.delete_all
      SidfmdVulView.delete_all
      data.each_slice(PROC_UNIT) do |data_s|
        temp_data = []
        data_s.each do |i|
          vul = SidfmdVulView.new
          build_row(vul, i)
          temp_data << vul
        end
        SidfmdVulView.import(temp_data, validate: false, recursive: true) if temp_data.any?
        temp_data.clear
      end
    end
    data
  end

  # 指定した複数のhtmlidのデータのみ同期する
  # transaction外で実行すること（メソッド内部でtransactionを使っているため）。
  # 指定したidに対応するSidfmdIssue, SidfmdVulViewの状態で以下のように動作する。
  #
  # [htmlidに対応するSidfmdIssue無し]
  # [htmlidに対応するSidfmdIssueに関連したホストが全てflg_delete=true]
  # 対応するSidfmdVulViewを削除
  #
  # [htmlidに対応するSidfmdIssueあり、対応するSidfmdVulView無し]
  # 対応するSidfmdVulViewを新規作成して追加
  #
  # [htmlidに対応するSidfmdIssue, SidfmdVulView共にあり]
  # 対応するSidfmdVulViewを更新
  # 既に取得済みのSidfmdIssue#create_all_contentのデータがあるなら、
  # 引数に入れておくと再度取得しない。
  #
  # sidfmd_issue_idやissue_idではなくhtmlidであることに注意。
  def self.sync_data(htmlids = [], ext_vul_content = nil)
    # 削除
    not_exist_htmlids =
      htmlids - SidfmdIssue.where("html_id IN (?)", htmlids).select(:html_id).distinct.map(&:html_id).uniq
    delete_htmlids = SidfmdVulView.where("html_id IN (?)", not_exist_htmlids).select(:html_id).distinct.map(&:html_id)
    #delete_htmlids |= SidfmdIssue.joins(:sidfmd_host).where(:html_id => htmlids).where(:sidfmd_hosts => { :flg_delete => true }).select(:html_id).distinct.map(&:html_id)
    delete_htmlids.each_slice(PROC_UNIT) do |delete_htmlids_s|
      SidfmdVulView.transaction do
        SidfmdVulView.where(:html_id => delete_htmlids_s).destroy_all
      end
    end

    modify_htmlids = SidfmdVulView.where(:html_id => htmlids).select(:html_id).distinct.map(&:html_id).uniq

    # 追加
    add_htmlids = htmlids - modify_htmlids - delete_htmlids
    add_htmlids.uniq!
    no_add_htmlids = []
    add_htmlids.each_slice(PROC_UNIT) do |add_htmlids_s|
      temp_data = []
      vul_content = (ext_vul_content || SidfmdIssue.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required, html_ids: add_htmlids_s))
      add_htmlids_s.each do |html_id|
        row = vul_content.find {|x| x[:html_id] == html_id }
        if row then
          vul = SidfmdVulView.new
          build_row(vul, row)
          temp_data << vul
        else
          no_add_htmlids << html_id
        end
      end
      if temp_data.any? then
        SidfmdVulView.transaction do
          SidfmdVulView.import(temp_data, validate: false, recursive: true)
        end
        temp_data.clear
      end
      vul_content.clear unless ext_vul_content
    end
    add_htmlids -= no_add_htmlids

    # 更新
    no_modify_htmlids = []
    modify_htmlids.each_slice(PROC_UNIT) do |modify_htmlids_s|
      vul_content = (ext_vul_content || SidfmdIssue.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required, html_ids: modify_htmlids_s))
      SidfmdVulView.transaction do
        SidfmdVulView.where(:html_id => modify_htmlids_s).lock.destroy_all
        temp_data = []
        modify_htmlids_s.each do |html_id|
          row = vul_content.find {|x| x[:html_id] == html_id }
          if row then
            vul = SidfmdVulView.new
            build_row(vul, row)
            temp_data << vul
          else
            no_modify_htmlids << html_id
          end
        end
        if temp_data.any? then
          SidfmdVulView.import(temp_data, validate: false, recursive: true)
          temp_data.clear
        end
        vul_content.clear unless ext_vul_content
      end
    end
    modify_htmlids -= no_modify_htmlids
      
    { :add => add_htmlids, :modify => modify_htmlids, :delete => delete_htmlids }
  end

  # 指定したhtml_idの配列について、関連するsidfmd_host_idの配列を返す。
  # ViewではなくSidfmdIssueから取得しているため、更新前と更新後の2回
  # 取得し、合わせたものをSidfmdHostView.sync_dataに与えてViewを更新する。
  def self.relate_hostids(htmlids = [])
    SidfmdIssue.where(:html_id => htmlids).select(:sidfmd_host_id).distinct.map(&:sidfmd_host_id).uniq.compact
  end

  # rule_actionで更新を行った値に対して、Viewを直接更新する。
  # [ { :sidfmd_issue_id, :html_id, :sidfmd_host_id, :due_date, :status_id } ]
  # View以外操作しないこと。
  def self.update_from_rule_action(data)
    corresp_required = Sidfm.corresp_required
    data.each_slice(PROC_UNIT) do |data_s|
      SidfmdVulView.transaction do
        data_s.each do |i|
          row = SidfmdVvScsid.lock.joins(:sidfmd_vul_view).where(:sidfmd_vul_views => { :html_id => i[:html_id] }).where(:sidfmd_host_id => i[:sidfmd_host_id]).first
          if row then
            opts = {}
            if i.has_key?(:due_date) then
              opts[:due_date] = i[:due_date]
            end
            if i.has_key?(:status_id) then
              opts[:status_id] = i[:status_id]
            end
            if opts.size > 0 then
              row.update_attributes opts
            end
          end
        end
        html_ids = data_s.collect {|x| x[:html_id] }.uniq.compact
        html_ids.each do |html_id|
          vul = SidfmdVulView.lock.find_by_html_id(html_id)
          vul.sidfmd_vv_due_dates.delete_all
          temp_data = []
          vul.sidfmd_vv_scsids.each do |i|
            due_date = SidfmdVvDueDate.new
            due_date.sidfmd_vul_view_id = vul.id
            due_date.dt = i.due_date
            temp_data << due_date
          end
          SidfmdVvDueDate.import(temp_data, validate: false) if temp_data.any?
          vul.sidfmd_vv_unf_due_dates.delete_all
          temp_data.clear
          dp_cnt = 0
          f_cnt = 0
          fp_cnt = 0
          min_status_id = 5
          max_sri = 0
          max_cvss = nil
          min_elapsed_created_on = Time.current
          vul.sidfmd_vv_scsids.each do |i|
            min_status_id = i.status_id if min_status_id > i.status_id
            max_cvss = i.cvss if max_cvss.nil? or (i.cvss and max_cvss < i.cvss)
            if SidfmdIssue::STATUS_UNFINISHED.include?(i.status_id) then
              f_cnt += 1
              dp_cnt += 1 if i.due_date
              max_sri = i.sri if max_sri < i.sri
              min_elapsed_created_on = i.created_on if (i.created_on and min_elapsed_created_on > i.created_on)
              due_date = SidfmdVvUnfDueDate.new
              due_date.sidfmd_vul_view_id = vul.id
              due_date.dt = i.due_date
              temp_data << due_date
            else
              fp_cnt += 1
            end
          end
          SidfmdVvUnfDueDate.import(temp_data, validate: false) if temp_data.any?
          opts = {
            :not_fixed_hosts => f_cnt,
            :min_status_id => min_status_id,
            :max_sri => max_sri,
            :cvss_score => max_cvss,
            :cvss_label => SidfmdIssueCve.labelling(max_cvss).to_s,
          }
          opts[:exploitable] = nil
          if f_cnt > 0 then
            opts[:duedates_percent] = (dp_cnt.to_f / f_cnt.to_f * 100.0)
            unf_issues = SidfmdIssue.joins(:issue)
                                    .joins(:sidfmd_host)
                                    .joins(:sidfmd_issue_cves)
                                    .where(:sidfmd_hosts => { :flg_delete => false }) 
                                    .where(:sidfmd_issues => { :html_id => vul.html_id })
                                    .where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
            opts[:exploitable] = if unf_issues then
                                   unf_issues.where(%| sidfmd_issue_cves.c3_exploitable IS TRUE OR sidfmd_issue_cves.c2_exploitable IS TRUE |)
                                             .exists?
                                 end
          else
            opts[:duedates_percent] = 0.0
          end
          scsids_cnt = vul.sidfmd_vv_scsids.count
          if scsids_cnt > 0 then
            opts[:fixed_percent] = (fp_cnt.to_f / scsids_cnt.to_f * 100.0)
          else
            opts[:fixed_percent] = 0.0
          end
          opts[:warning_orig_sri] = max_sri ? max_sri >= corresp_required[:sri] : false
          opts[:warning_orig_cvss] = max_cvss ? max_cvss >= corresp_required[:cvss] : true
          opts[:warning_sri] = SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) ? opts[:warning_orig_sri] : false
          opts[:warning_cvss] = SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) ? opts[:warning_orig_cvss] : false
          opts[:fixed_warning_sri] = SidfmdIssue::STATUS_FINISHED.include?(min_status_id) ? opts[:warning_orig_sri] : false
          opts[:fixed_warning_cvss] = SidfmdIssue::STATUS_FINISHED.include?(min_status_id) ? opts[:warning_orig_cvss] : false
          opts[:not_fixed_warning_sri] = opts[:warning_sri]
          opts[:not_fixed_warning_cvss] = opts[:warning_cvss]
          opts[:min_elapsed_created_on] = min_elapsed_created_on
          opts[:min_elapsed_created_on_display] = (min_elapsed_created_on ? ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i : -1).to_s
          vul.update_attributes(opts) if opts.size > 0

          vul.sidfmd_vv_ctags.delete_all
          new_ctags = SidfmdIssueCtag.joins(:sidfmd_issue)
            .where(:sidfmd_issues => { :html_id => html_id })
            .select('sidfmd_issues.sidfmd_host_id AS sidfmd_host_id', 'sidfmd_issue_ctags.sidfmd_ctag_id AS ctag_id')
            .map{|issue_ctag|
              SidfmdVvCtag.new(
                sidfmd_vul_view_id: vul.id,
                sidfmd_host_id: issue_ctag.sidfmd_host_id,
                sidfmd_ctag_id: issue_ctag.ctag_id
              )
            }
          SidfmdVvCtag.import(new_ctags)
        end
      end
    end
  end

  private

  public

  # SidfmdIssue#create_all_contentのデータ1行の値を反映する。
  def self.build_row(vul, row, modify = false)
    vul.html_id = row[:html_id]
    vul.title = row[:title]
    if row[:softek] then
      vul.softek_risk_id = row[:softek][:risk_id] if row[:softek][:risk_id]
      vul.softek_area_id = row[:softek][:area_id] if row[:softek][:area_id]
    end
    vul.short_description = row[:short_description]
    vul.exploitable = row[:exploitable]
    vul.max_sri = row[:max_sri]
    vul.min_status_id = row[:min_status_id]
    if row[:cvss] then
      vul.cvss_score = row[:cvss][:score] if row[:cvss][:score]
      vul.cvss_label = row[:cvss][:label].to_s if row[:cvss][:label]
    end
    [ 'warning_orig', 'warning', 'fixed_warning', 'not_fixed_warning' ].each do |warning|
      w_sym = warning.to_sym
      if row[w_sym] and row[w_sym].kind_of?(Hash) then
        vul.send((warning + '_sri=').to_sym, row[w_sym][:sri])
        vul.send((warning + '_cvss=').to_sym, row[w_sym][:cvss])
      end
    end
    if row[:due_dates_unfinished] then
      f_cnt = row[:due_dates_unfinished].size
      if f_cnt > 0 and row[:due_dates_unfinished].compact.size > 0 then
        cnt = row[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
        vul.duedates_percent = cnt.to_f / f_cnt.to_f * 100.0
      else
        vul.duedates_percent = 0.0
      end
    end
    if row[:status_ids] then
      f_cnt = row[:status_ids].size
      if f_cnt > 0 and row[:status_ids].compact.size > 0 then
        cnt = row[:status_ids].inject(0) {|r, x|
          r += 1 if SidfmdIssue::STATUS_FINISHED.include?(x)
          r
        }
        vul.fixed_percent = cnt.to_f / f_cnt.to_f * 100.0
      else
        vul.fixed_percent = 0.0
      end
    end
    vul.not_fixed_hosts = row[:not_fixed_hosts]
    if row[:min_created_on] then
      vul.min_created_on = row[:min_created_on][:value] if row[:min_created_on][:value]
      vul.min_created_on_display = row[:min_created_on][:display] if row[:min_created_on][:value]
    end
    if row[:min_elapsed_created_on] then
      vul.min_elapsed_created_on = row[:min_elapsed_created_on][:value] if row[:min_elapsed_created_on][:value]
      vul.min_elapsed_created_on_display = row[:min_elapsed_created_on][:display].to_s if row[:min_elapsed_created_on][:display]
    end
    if modify then
      vul.save
    end

    if modify then
      vul.sidfmd_vv_scsids.delete_all
      scsids = []
      row[:sri_cvss_status_ids].each do |j|
        scsid = SidfmdVvScsid.new
        scsid.sidfmd_vul_view_id = vul.id
        scsid.sidfmd_host_id = j[:sidfmd_host_id]
        scsid.status_id = j[:status_id]
        scsid.due_date = j[:due_date]
        scsid.created_on = j[:created_on]
        scsid.sri = j[:sri]
        scsid.cvss = j[:cvss]
        scsids << scsid
      end
      SidfmdVvScsid.import(scsids, validate: false) if scsids.any?
    else
      row[:sri_cvss_status_ids].each do |j|
        vul.sidfmd_vv_scsids.build(sidfmd_host_id: j[:sidfmd_host_id],
                                   status_id: j[:status_id],
                                   due_date: j[:due_date],
                                   created_on: j[:created_on],
                                   sri: j[:sri],
                                   cvss: j[:cvss])
      end
    end

    if modify then
      hosts = []
      vul.sidfmd_vv_hosts.delete_all
      row[:hosts].each do |j|
        host = SidfmdVvHost.new
        host.sidfmd_vul_view_id = vul.id
        host.sidfmd_host_id = j[:id]
        host.hostname = j[:hostname]
        host.ext_name = j[:ext_name]
        host.display_name = j[:display_name]
        hosts << host
      end
      SidfmdVvHost.import(hosts, validate: false) if hosts.any?
    else
      row[:hosts].each do |j|
        vul.sidfmd_vv_hosts.build(sidfmd_host_id: j[:id],
                                  hostname: j[:hostname],
                                  ext_name: j[:ext_name],
                                  display_name: j[:display_name])
      end
    end

    if modify then
      categories = []
      vul.sidfmd_vv_categories.delete_all
      row[:categories].each do |j|
        category = SidfmdVvCategory.new
        category.sidfmd_vul_view_id = vul.id
        category.sidfmd_category_id = j[:id]
        category.is_system = j[:is_system]
        category.name = j[:name]
        categories << category
      end
      SidfmdVvCategory.import(categories, validate: false) if categories.any?
    else
      row[:categories].each do |j|
        vul.sidfmd_vv_categories.build(sidfmd_category_id: j[:id],
                                       is_system: j[:is_system],
                                       name: j[:name])
      end
    end
 
    if modify then
      due_dates = []
      vul.sidfmd_vv_due_dates.delete_all
      row[:due_dates].each do |j|
        due_date = SidfmdVvDueDate.new
        due_date.sidfmd_vul_view_id = vul.id
        due_date.dt = j
        due_dates << due_date
      end
      SidfmdVvDueDate.import(due_dates, validate: false) if due_dates.any?
    else
      row[:due_dates].each do |j|
        vul.sidfmd_vv_due_dates.build(dt: j)
      end
    end

    if modify then
      due_dates = []
      vul.sidfmd_vv_unf_due_dates.delete_all
      row[:due_dates_unfinished].each do |j|
        due_date = SidfmdVvUnfDueDate.new
        due_date.sidfmd_vul_view_id = vul.id
        due_date.dt = j
        due_dates << due_date
      end
      SidfmdVvUnfDueDate.import(due_dates, validate: false) if due_dates.any?
    else
      row[:due_dates_unfinished].each do |j|
        vul.sidfmd_vv_unf_due_dates.build(dt: j)
      end
    end


    if modify then
      cves = []
      vul.sidfmd_vv_cves.delete_all
      row[:cves].each do |j|
        cve = SidfmdVvCve.new
        cve.sidfmd_vul_view_id = vul.id
        cve.cve_id = j[:cve_id]
        cve.c2_base_score = j[:c2_base_score] < 0 ? nil : j[:c2_base_score]
        cve.c3_base_score = j[:c3_base_score] < 0 ? nil : j[:c3_base_score]
        if j[:c2_vector] and j[:c2_vector].size > 0 then
          if /AV:(.)/ =~ j[:c2_vector] then
            cve.c2_av = Regexp.last_match(1)
          end
          if /AC:(.)/ =~ j[:c2_vector] then
            cve.c2_ac = Regexp.last_match(1)
          end
          if /Au:(.)/ =~ j[:c2_vector] then
            cve.c2_au = Regexp.last_match(1)
          end
          if /[^A]C:(.)/ =~ j[:c2_vector] then
            cve.c2_c = Regexp.last_match(1)
          end
          if /I:(.)/ =~ j[:c2_vector] then
            cve.c2_i = Regexp.last_match(1)
          end
          if /A:(.)/ =~ j[:c2_vector] then
            cve.c2_a = Regexp.last_match(1)
          end
        end
        if j[:c3_vector] and j[:c3_vector].size > 0 then
          if /AV:(.)/ =~ j[:c3_vector] then
            cve.c3_av = Regexp.last_match(1)
          end
          if /AC:(.)/ =~ j[:c3_vector] then
            cve.c3_ac = Regexp.last_match(1)
          end
          if /PR:(.)/ =~ j[:c3_vector] then
            cve.c3_pr = Regexp.last_match(1)
          end
          if /UI:(.)/ =~ j[:c3_vector] then
            cve.c3_ui = Regexp.last_match(1)
          end
          if /S:(.)/ =~ j[:c3_vector] then
            cve.c3_s = Regexp.last_match(1)
          end
          if /[^A]C:(.)/ =~ j[:c3_vector] then
            cve.c3_c = Regexp.last_match(1)
          end
          if /[^U]I:(.)/ =~ j[:c3_vector] then
            cve.c3_i = Regexp.last_match(1)
          end
          if /A:(.)/ =~ j[:c3_vector] then
            cve.c3_a = Regexp.last_match(1)
          end
        end
        [ :c2_create_datetime, :c2_update_datetime,
          :c3_create_datetime, :c3_update_datetime ].each do |label|
          if j[label] and j[label].size > 0 then
            cve[label] = j[label]
          end
        end
        cves << cve
      end
      SidfmdVvCve.import(cves, validate: false) if cves.any?
    else
      row[:cves].each do |j|
        cve = {}
        [ :c2_av, :c2_ac, :c2_au, :c2_c, :c2_i, :c2_a,
          :c3_av, :c3_ac, :c3_pr, :c3_ui,
          :c3_s, :c3_c, :c3_i, :c3_a ].each do |k|
          cve[k] = nil
        end
        cve[:cve_id] = j[:cve_id]
        cve[:c2_base_score] = j[:c2_base_score] < 0 ? nil : j[:c2_base_score]
        cve[:c3_base_score] = j[:c3_base_score] < 0 ? nil : j[:c3_base_score]
        if j[:c2_vector] and j[:c2_vector].size > 0 then
          if /AV:(.)/ =~ j[:c2_vector] then
            cve[:c2_av] = Regexp.last_match(1)
          end
          if /AC:(.)/ =~ j[:c2_vector] then
            cve[:c2_ac] = Regexp.last_match(1)
          end
          if /Au:(.)/ =~ j[:c2_vector] then
            cve[:c2_au] = Regexp.last_match(1)
          end
          if /[^A]C:(.)/ =~ j[:c2_vector] then
            cve[:c2_c] = Regexp.last_match(1)
          end
          if /I:(.)/ =~ j[:c2_vector] then
            cve[:c2_i] = Regexp.last_match(1)
          end
          if /A:(.)/ =~ j[:c2_vector] then
            cve[:c2_a] = Regexp.last_match(1)
          end
        end
        if j[:c3_vector] and j[:c3_vector].size > 0 then
          if /AV:(.)/ =~ j[:c3_vector] then
            cve[:c3_av] = Regexp.last_match(1)
          end
          if /AC:(.)/ =~ j[:c3_vector] then
            cve[:c3_ac] = Regexp.last_match(1)
          end
          if /PR:(.)/ =~ j[:c3_vector] then
            cve[:c3_pr] = Regexp.last_match(1)
          end
          if /UI:(.)/ =~ j[:c3_vector] then
            cve[:c3_ui] = Regexp.last_match(1)
          end
          if /S:(.)/ =~ j[:c3_vector] then
            cve[:c3_s] = Regexp.last_match(1)
          end
          if /[^A]C:(.)/ =~ j[:c3_vector] then
            cve[:c3_c] = Regexp.last_match(1)
          end
          if /[^U]I:(.)/ =~ j[:c3_vector] then
            cve[:c3_i] = Regexp.last_match(1)
          end
          if /A:(.)/ =~ j[:c3_vector] then
            cve[:c3_a] = Regexp.last_match(1)
          end
        end
        [ :c2_create_datetime, :c2_update_datetime,
          :c3_create_datetime, :c3_update_datetime ].each do |label|
          if j[label] and j[label].size > 0 then
            cve[label] = j[label]
          else
            cve[label] = nil
          end
        end
        vul.sidfmd_vv_cves.build(cve)
      end
    end

    if modify then
      vul.sidfmd_vv_ctags.delete_all
      new_ctags = row[:ctags].map{|ctag|
        SidfmdVvCtag.new(
          sidfmd_vul_view_id: vul.id,
          sidfmd_host_id: ctag[:sidfmd_host_id],
          sidfmd_ctag_id: ctag[:ctag_id]
        )
      }
      SidfmdVvCtag.import(new_ctags)
    else
      row[:ctags].each do |ctag|
        vul.sidfmd_vv_ctags.build(
          sidfmd_host_id: ctag[:sidfmd_host_id],
          sidfmd_ctag_id: ctag[:ctag_id]
        )
      end  
    end

    vul
  end

  def apply_row(row)
    self.class.build_row(self, row, true)
  end

  #UIやファイル出力で出力するタイトル
  def self.display_title(title)
    title = title.split(' / ').uniq.join(' / ')
  end
  
  #Viewを使って、脆弱性の各項目の算出した値、もしくはnilを返す(脆弱性別状況ページのサマリー表の各列の値)
  def create_summary_values(role_admin, category_id, user_category_ids, user_host_ids, corresp_required)
    calc_necessary = false #ページ上部のカテゴリのフィルタがある、または管理者権限ではないときは表示する値の再計算が必要
    if !role_admin or category_id > 0 then
      calc_necessary = true
      scsids = if category_id > 0 then
                 sidfmd_vv_scsids.where(
                   :sidfmd_host_id => SidfmdCategoriesHost.where(:sidfmd_category_id => category_id).select(:sidfmd_host_id)
                 )
               else
                 sidfmd_vv_scsids
               end
      if !role_admin then
        scsids = scsids.where(:sidfmd_host_id => user_host_ids)
      end
    else
      scsids = sidfmd_vv_scsids
    end

    #title = (self.title.split(' / ').size != self.title.split(' / ').uniq.size) ? self.title.split(' / ').uniq.join(' / ') : self.title.split(' / ').first
    title = SidfmdVulView.display_title(self.title)
 
    if !scsids.exists? then
      ret = nil #万が一何らかのタイミングでデータが無かった時はnilを返す。nilチェックは呼び出し元で行う。
    elsif calc_necessary then
      status_ids = []
      unf_cnt = 0
      dp_cnt = 0
      fp_cnt = 0
      min_status_id = 5
      max_sri = 0
      max_cvss_score = nil
      min_created_on = nil
      min_elapsed_created_on = nil
      unf_max_cvss_score = nil
 
      scsids.each do |i|
        min_created_on = i.created_on if min_created_on.nil? or (i.created_on and i.created_on < min_created_on)
        min_status_id = i.status_id if min_status_id > i.status_id
        max_cvss_score = i.cvss if max_cvss_score.nil? or (i.cvss and max_cvss_score < i.cvss)    
        status_ids << i.status_id
        if SidfmdIssue::STATUS_UNFINISHED.include?(i.status_id) then
          unf_cnt += 1
          dp_cnt += 1 if i.due_date
          max_sri = i.sri if max_sri < i.sri
          unf_max_cvss_score = i.cvss if unf_max_cvss_score.nil? or (i.cvss and unf_max_cvss_score < i.cvss)
          min_elapsed_created_on = i.created_on if min_elapsed_created_on.nil? or (i.created_on and min_elapsed_created_on > i.created_on)
        else
          fp_cnt += 1
        end        
      end

      host_ids = scsids.pluck(:sidfmd_host_id)
      c_categories = sidfmd_vv_categories.where(
                     :is_system => false,
                     :sidfmd_category_id => SidfmdCategoriesHost.where(:sidfmd_host_id => host_ids).select(:sidfmd_category_id)
                   )
      if !role_admin then
        c_categories = c_categories.where(:sidfmd_category_id => user_category_ids)
      end
      ret = {
        :id => self.id,
        :html_id => self.html_id,
        :warning_sri => (SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) and max_sri >= corresp_required[:sri]),
        :warning_cvss => (SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) and (!max_cvss_score || max_cvss_score >= corresp_required[:cvss])),
        :status_ids => status_ids,
        :min_status_id => min_status_id,
        :max_sri => max_sri,
        :cvss_label => SidfmdIssueCve.labelling(max_cvss_score).to_s,
        :cvss_score => max_cvss_score,
        :is_cves => sidfmd_vv_cves.exists?,
        :softek_area_id => self.softek_area_id,
        :exploitable => (self.html_id < 0 ? nil : (unf_max_cvss_score.nil? ? nil : self.exploitable)),
        :min_created_on_display => min_created_on ? min_created_on.to_datetime.localtime.strftime('%Y/%m/%d') : '---',
        :min_elapsed_created_on_display => min_elapsed_created_on ? ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i : -1,
        :short_description => self.short_description,
        :title => title,
        :hosts => sidfmd_vv_hosts.where(:sidfmd_host_id => host_ids).order('LOWER(display_name COLLATE "C")'),
        :system_categories => sidfmd_vv_categories.where(
                                :is_system => true,
                                :sidfmd_category_id => SidfmdCategoriesHost.where(:sidfmd_host_id => host_ids).select(:sidfmd_category_id)
                              ).order(:sidfmd_category_id),
        :custom_categories => c_categories.order('LOWER(name COLLATE "C")'),
        :duedates_percent => dp_cnt > 0 ? (dp_cnt.to_f / unf_cnt.to_f * 100.0) : 0.0,
        :fixed_percent => fp_cnt > 0 ? (fp_cnt.to_f / status_ids.size.to_f * 100.0) : 0.0
      }
    else
      min_created_on = nil
      min_elapsed_created_on = nil
      status_ids = []
      scsids.each do |i|
        status_ids << i.status_id
        min_created_on = i.created_on if min_created_on.nil? or (i.created_on and i.created_on < min_created_on)
        if SidfmdIssue::STATUS_UNFINISHED.include?(i.status_id) then
          min_elapsed_created_on = i.created_on if min_elapsed_created_on.nil? or (i.created_on and min_elapsed_created_on > i.created_on)
        end
      end

      ret = {
        :id => self.id,
        :html_id => self.html_id,
        :warning_sri => self.warning_sri,
        :warning_cvss => self.warning_cvss,
        :status_ids => status_ids,
        :min_status_id => self.min_status_id,
        :max_sri => self.max_sri,
        :cvss_label => self.cvss_label,
        :cvss_score => self.cvss_score,
        :is_cves => sidfmd_vv_cves.exists?,
        :softek_area_id => self.softek_area_id,
        :exploitable => self.cvss_label == 'notreg' ? nil : self.exploitable,
        :min_created_on_display => min_created_on ? min_created_on.to_datetime.localtime.strftime('%Y/%m/%d') : '---',
        :min_elapsed_created_on_display => min_elapsed_created_on ? ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i : -1, 
        :short_description => self.short_description,
        :title => title,
        :hosts => sidfmd_vv_hosts.order('LOWER(display_name COLLATE "C")'),
        :system_categories => sidfmd_vv_categories.where(:is_system => true).order(:sidfmd_category_id),
        :custom_categories => sidfmd_vv_categories.where(:is_system => false).order('LOWER(name COLLATE "C")'),
        :duedates_percent => self.duedates_percent,
        :fixed_percent => self.fixed_percent
      }    
    end
 
    ret
  end

  # bulk_create_summary_valueで、カテゴリのフィルタがあるまたは管理者権限でないときに
  # 計算が必要な項目を生成する
  def self.create_summary_calc_value(vul_view)
    if !vul_view.respond_to?(:cvss_label) then
      def vul_view.cvss_label
        SidfmdIssueCve.labelling(self.cvss_score).to_s
      end
    end
    if !vul_view.respond_to?(:exploitable) then
      def vul_view.exploitable
        (self.html_id < 0 ? nil : (self.unf_cvss_score.nil? ? nil : self.temp_exploitable))
      end
    end
    if !vul_view.respond_to?(:duedates_percent) then
      def vul_view.duedates_percent
        self.count_due_dates.to_i > 0 ? (self.count_due_dates.to_f / self.count_unfinished.to_f * 100.0) : 0.0
      end
    end
    if !vul_view.respond_to?(:fixed_percent) then
      def vul_view.fixed_percent
        self.count_fixed.to_i > 0 ? (self.count_fixed.to_f / self.status_ids.size.to_f * 100.0) : 0.0
      end
    end
    vul_view
  end
  #脆弱性別状況ページのサマリー表の各列を生成するために必要な値を生成する
  #html生成時には表示する値に変換が必要
  #opts = :vul_views, :role_admin, :category_id, :user_category_ids, :user_host_ids, :corresp_required
  #vul_viewsはSidfmdVulView。order等は呼び出し元で行う
  #カテゴリのフィルタがあるまたは管理者権限でない時と
  #カテゴリのフィルタがなく管理者権限である時に返る値が異なるので注意
  def self.bulk_create_summary_values(**opts)
    filter_category_host_ids = if opts[:category_id].to_i > 0 then
                                 if opts[:role_admin] then
                                   SidfmdCategoriesHost.where(:sidfmd_category_id => opts[:category_id])
                                                       .pluck(:sidfmd_host_id)
                                 else
                                    SidfmdCategoriesHost.where(
                                      :sidfmd_category_id => opts[:category_id], :sidfmd_host_id => opts[:user_host_ids]
                                    )
                                    .pluck(:sidfmd_host_id)
                                  end
                               else
                                 nil
                               end
    filter_host_ids = if filter_category_host_ids then
      filter_category_host_ids
    else
      if !opts[:role_admin] then
        opts[:user_host_ids] 
      else
        nil
      end
    end

    if filter_host_ids then
      if filter_host_ids.any? then
        vul_views = opts[:vul_views].select(
          :id,
          :html_id,
          :title,
          :softek_area_id,
          :short_description
        ).select("
          EXISTS(
            SELECT id FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND sri >= #{opts[:corresp_required][:sri]} AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
          ) AS warning_sri
        ").select("
          EXISTS(
            SELECT id FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND (cvss IS NULL OR cvss >= #{opts[:corresp_required][:cvss]}) AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
          ) AS warning_cvss
        ").select("
          (
            SELECT MIN(status_id) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id
          ) AS min_status_id
        ").select("
          (
            SELECT MAX(sri) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id
          ) AS max_sri
        ").select("
          (
            SELECT MAX(cvss) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id
          ) AS cvss_score
        ").select("
          (
            ARRAY(
              SELECT status_id FROM sidfmd_vv_scsids WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            )
          ) AS status_ids
        ").select("
          exploitable AS temp_exploitable
        ").select("
          (
            SELECT MAX(cvss) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')})
            GROUP BY sidfmd_vul_view_id
          ) AS unf_cvss_score /* exploitableを算出するために使用 */
        ").select("
          EXISTS(SELECT id FROM sidfmd_vv_cves WHERE sidfmd_vul_views.id = sidfmd_vul_view_id) AS is_cves
        ")
        .select("
          (
            SELECT MIN(created_on) FROM sidfmd_vv_scsids 
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id
          ) AS min_created_on
        ").select("
          (
            SELECT MIN(created_on) FROM sidfmd_vv_scsids 
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id
          ) AS min_elapsed_created_on
        ").select(%|
          (
            ARRAY(
              SELECT ARRAY[CAST(sidfmd_host_id AS text), hostname, ext_name, display_name] FROM sidfmd_vv_hosts 
              WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
              ORDER BY LOWER(display_name COLLATE "C")
            )
          ) AS hosts
        |).select("
          (
            ARRAY(
              SELECT ARRAY[CAST(sidfmd_categories.id AS text), sidfmd_categories.name]
              FROM sidfmd_vv_scsids
                INNER JOIN sidfmd_categories_hosts ON sidfmd_vv_scsids.sidfmd_host_id = sidfmd_categories_hosts.sidfmd_host_id
                INNER JOIN sidfmd_categories ON sidfmd_categories.id = sidfmd_categories_hosts.sidfmd_category_id
                WHERE sidfmd_vul_views.id = sidfmd_vv_scsids.sidfmd_vul_view_id AND
                  sidfmd_categories.is_system IS TRUE AND
                  sidfmd_categories_hosts.sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}]))  
              ORDER BY sidfmd_categories.id
            )
          ) AS system_categories /*　重複するため、重複を除く場合は別途処理する　*/
        ").select(%|
          (
            ARRAY(
              SELECT ARRAY[CAST(sidfmd_category_id AS text), name]
              FROM sidfmd_vv_scsids
                INNER JOIN sidfmd_categories_hosts ON sidfmd_vv_scsids.sidfmd_host_id = sidfmd_categories_hosts.sidfmd_host_id
                INNER JOIN sidfmd_categories ON sidfmd_categories.id = sidfmd_categories_hosts.sidfmd_category_id
                WHERE sidfmd_vul_views.id = sidfmd_vv_scsids.sidfmd_vul_view_id AND
                  sidfmd_categories.is_system IS FALSE AND
                  sidfmd_categories_hosts.sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}]))#{(opts[:user_category_ids] and opts[:user_category_ids].any?) ? " AND sidfmd_category_id IN (SELECT unnest(ARRAY[#{opts[:user_category_ids].join(', ')}]))" : ''}
              ORDER BY LOWER(name COLLATE "C")
            )
          ) AS custom_categories /*　重複するため、重複を除く場合は別途処理する　*/
        |).select("
          (
            SELECT COUNT(*) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND due_date IS NOT NULL AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}]))  
            GROUP BY sidfmd_vul_view_id 
          ) AS count_due_dates
        ").select("
          (
            SELECT COUNT(*) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id NOT IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}])) 
            GROUP BY sidfmd_vul_view_id 
          ) AS count_fixed
        ").select("
          (
            SELECT COUNT(*) FROM sidfmd_vv_scsids
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}) AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids.join(', ')}]))
            GROUP BY sidfmd_vul_view_id 
          ) AS count_unfinished
        ")
      else
        vul_views = opts[:vul_views].where(:id => -1).select(:id) #脆弱性がないとき
      end
    else
      vul_views = opts[:vul_views].select(
        :id,
        :html_id,
        :title,
        :warning_sri,
        :warning_cvss,
        :min_status_id,
        :max_sri,
        :cvss_label,
        :cvss_score,  
        :softek_area_id,
        :short_description,
        :duedates_percent,
        :fixed_percent 
      ).select("
        (
          ARRAY(
            SELECT status_id FROM sidfmd_vv_scsids WHERE sidfmd_vul_views.id = sidfmd_vul_view_id
          )
        ) AS status_ids
      ").select("
        EXISTS(SELECT id FROM sidfmd_vv_cves WHERE sidfmd_vul_views.id = sidfmd_vul_view_id) AS is_cves
      ")
      .select("
        CASE WHEN cvss_label = 'notreg' THEN NULL ELSE exploitable END AS exploitable
      ") 
      .select("
        (
          SELECT MIN(created_on) FROM sidfmd_vv_scsids 
          WHERE sidfmd_vul_views.id = sidfmd_vul_view_id
          GROUP BY sidfmd_vul_view_id
        ) AS min_created_on
      ").select("
        (
          SELECT MIN(created_on) FROM sidfmd_vv_scsids 
          WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(', ')})
          GROUP BY sidfmd_vul_view_id
        ) AS min_elapsed_created_on
      ").select(%|
        (
          ARRAY(
            SELECT ARRAY[CAST(sidfmd_host_id AS text), hostname, ext_name, display_name] FROM sidfmd_vv_hosts 
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id
            ORDER BY LOWER(display_name COLLATE "C")
          )
        ) AS hosts
      |).select("
        (
          ARRAY(
            SELECT ARRAY[CAST(sidfmd_category_id AS text), name]  FROM sidfmd_vv_categories
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND is_system IS TRUE
            ORDER BY sidfmd_category_id
          )
        ) AS system_categories
      ").select(%|
        (
          ARRAY(
            SELECT ARRAY[CAST(sidfmd_category_id AS text), name]  FROM sidfmd_vv_categories
            WHERE sidfmd_vul_views.id = sidfmd_vul_view_id AND is_system IS FALSE
            ORDER BY LOWER(name COLLATE "C")
          )
        ) AS custom_categories
      |)
    end

    #カスタムタグ
    #管理者でないときはログインユーザのカテゴリのタグのみ
    vul_views = vul_views.select(%|
      (
        ARRAY(
          SELECT ARRAY[keyword, color, bgcolor] 
          FROM sidfmd_ctags
            INNER JOIN sidfmd_vv_ctags ON sidfmd_vv_ctags.sidfmd_ctag_id = sidfmd_ctags.id
          WHERE sidfmd_vul_views.id = sidfmd_vv_ctags.sidfmd_vul_view_id#{(!opts[:role_admin] && opts[:user_category_ids] && opts[:user_category_ids].any?) ? " AND (sidfmd_ctags.sidfmd_category_id IS NULL OR sidfmd_ctags.sidfmd_category_id IN (SELECT unnest(ARRAY[#{opts[:user_category_ids].join(', ')}]))) AND sidfmd_vv_ctags.sidfmd_host_id IN (SELECT unnest(ARRAY[#{filter_host_ids && filter_host_ids.any? ? filter_host_ids.join(', ') : "-1"}]))" : ""}
          ORDER BY sidfmd_ctags.priority DESC, LOWER(sidfmd_ctags.keyword COLLATE "C") ASC
        )
      ) AS ctags
    |)
  end

  #サマリー表では、以下の処理をして表示するcveを絞り込みしているため、
  #同じ処理をして脆弱性の対象のcveを取得する。
  #「CVEIDが同じものがある場合は、
  #c3_update_datetimeがあればc3_update_datetime, 
  #c2_update_datetimeがあればc2_update_datetimeで最も新しいもの1つを採用する。
  #両方無い場合はidが最も大きいものを採用する。」
  def get_summary_cves
    sidfmd_vv_cves.where(%|
      id = (
        SELECT id
        FROM sidfmd_vv_cves AS cve_table2
        WHERE sidfmd_vv_cves.sidfmd_vul_view_id = cve_table2.sidfmd_vul_view_id AND sidfmd_vv_cves.cve_id = cve_table2.cve_id
        ORDER BY 
        CASE
        WHEN c3_update_datetime IS NOT NULL THEN
          100
        WHEN c2_update_datetime IS NOT NULL THEN
          200
        ELSE
          300
        END,
        c3_update_datetime DESC,
        c2_update_datetime DESC,
        id DESC
        LIMIT 1
      )
    |).order('cve_id COLLATE "C"')
  end

end
