# -*- coding: utf-8 -*-

class CreateSidfmdRules < ActiveRecord::Migration

  def change
    # PostgreSQLのテーブル名は63文字の制限があるため、一部省略していることに注意。

    # ルール
    create_table :sidfmd_rules do |t|
      t.references :user, foreign_key: true, index: true, null: true
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 優先順位。ソート・並び替え用
      t.integer :position, null: false, default: 0, index: true
      # ルールの有効・無効のフラグ
      t.boolean :flg_enable, null: false, default: false, index: true

      # ルール名称
      t.text :name, null: false, index: true

      t.timestamps null: false
    end

    # トリガー
    create_table :sidfmd_rule_triggers do |t|
      t.references :sidfmd_rule, foreign_key: true, index: true, null: false

      # 最後にチェックされた日時
      # nilはまだチェックされたことが無い状態。
      t.datetime :last_checked_at, index: true, null: true

      # ログを保存。JSON形式の配列で保存。
      t.text :logs, null: false, :default => '[]'

      t.timestamps null: false
    end

    # トリガー 定期実行-時間指定 periodic_hour
    # trigger->t, periodic->p
    create_table :sidfmd_rule_tp_hours do |t|
      t.references :sidfmd_rule_trigger, foreign_key: true, index: true, null: false

      # 実行される時間0～23時
      # なお、同じ時間に多数指定された場合は
      t.integer :hour, index: true, null: false
    end

    # トリガー 定期実行-間隔指定
    # trigger->t, periodic->p
    create_table :sidfmd_rule_tp_intervals do |t|
      t.references :sidfmd_rule_trigger, foreign_key: true, index: true, null: false

      # 間隔指定（1～24）。単位は時間。
      t.integer :hour_interval, index: true, null: false
    end

    # トリガー 脆弱性同期時
    # trigger->t
    create_table :sidfmd_rule_t_vulsyncs do |t|
      t.references :sidfmd_rule_trigger, foreign_key: true, index: true, null: false

      # 対象のカテゴリ。nilを指定すれば全ホストが対象になる。
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 対象ホスト。nilを指定するとカテゴリが有効になる。
      t.references :sidfmd_host, foreign_key: true, index: true, null: true
    end

    # 条件
    create_table :sidfmd_rule_conds do |t|
      t.references :sidfmd_rule, foreign_key: true, index: true, null: false

      # 最後にチェックされた日時
      # nilはまだチェックされたことが無い状態。
      t.datetime :last_checked_at, index: true, null: true

      # ログを保存。JSON形式の配列で保存。
      t.text :logs, null: false, :default => '[]'

      t.timestamps null: false
    end

    # 条件 脆弱性更新（issue_update）
    # cond->c
    create_table :sidfmd_rule_c_issue_updates do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # trueならor条件、falseならand条件でチェックする。
      t.boolean :cond_or, null: false, default: false

      # 条件「脆弱性の新規登録がcnt_create件以上ある」。0なら無効。
      t.integer :cnt_create, null: false, default: 0

      # 条件「脆弱性の削除（影響無へのステータス変更）がcnt_remove件以上ある」。0なら無効。
      t.integer :cnt_remove, null: false, default: 0
    end

    # 条件 SRI指定値以上 (sri_above)
    # cond->c
    create_table :sidfmd_rule_c_sri_aboves do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 条件「SRIがrisk_id以上」
      t.integer :risk_id, null: false
    end

    # 条件 SRI指定値以下 (sri_below)
    # cond->c
    create_table :sidfmd_rule_c_sri_belows do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 条件「SRIがrisk_id以下」
      t.integer :risk_id, null: false
    end

    # 条件 CVSS指定値以上 (cvss_above)
    create_table :sidfmd_rule_c_cvss_aboves do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # trueの場合、未定義を含む
      t.boolean :is_notreg, null: false, default: false

      # 条件「CVSSがcvss_value以上」
      t.float :cvss_value, null: false
    end

    # 条件 CVSS指定値以下 (cvss_below)
    # cond->c
    create_table :sidfmd_rule_c_cvss_belows do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # trueの場合、未定義を含む
      t.boolean :is_notreg, null: false, default: false

      # 条件「CVSSがcvss_value以下」
      t.float :cvss_value, null: false
    end

    # 条件 CVSS未定義 (cvss_notreg)
    # この条件は、CVSS未定義だけを検出するためのものです。
    # cond->c
    create_table :sidfmd_rule_c_cvss_notregs do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 条件「trueなら未定義、falseなら未定義以外」
      t.boolean :is_notreg, null: false, default: true
    end

    # 条件 要注意 (warning)
    # cond->c
    create_table :sidfmd_rule_c_warnings do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 条件「trueなら要注意、falseなら要注意以外が対象」
      t.boolean :is_warning, null: false, default: true

      # 条件「trueならCVSS、falseならSRIが対象」
      t.boolean :is_cvss, null: false, default: false
    end

    # 条件 履歴投稿(history_posts)
    # 履歴が投稿されたことを検出するためのものです。
    # 脆弱性更新でステータスが変更される際に、自動的に登録されるものは対象外です。
    # cond->c
    create_table :sidfmd_rule_c_history_posts do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 指定したカスタムカテゴリに属するホストを対象。nilは全てのホストが対象
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 指定したホストを対象。nilでsidfmd_categoryが有効。
      t.references :sidfmd_host, foreign_key: true, index: true, null: true

      # 指定した件数以上
      t.integer :cnt, null: false, default: 1
    end

    # 条件 対策予定日超過(duedate expired)
    # cond->c
    create_table :sidfmd_rule_c_duedate_expireds do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 指定したカスタムカテゴリに属するホストを対象。nilは全てのホストが対象
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 指定したホストを対象。nilでsidfmd_categoryが有効。
      t.references :sidfmd_host, foreign_key: true, index: true, null: true
    end

    # 条件 対策予定日接近（duedate approaching）
    # cond->c
    create_table :sidfmd_rule_c_duedate_approachings do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 指定したカスタムカテゴリに属するホストを対象。nilは全てのホストが対象
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 指定したホストを対象。nilでsidfmd_categoryが有効。
      t.references :sidfmd_host, foreign_key: true, index: true, null: true

      # 設定されている対策予定日のdays日前より後の日付
      t.integer :days, null: false
    end

    #  条件 カスタムカテゴリ適合
    # 指定したカスタムカテゴリに属するホストの脆弱性を範囲とする。
    # 他の条件と組み合わせて使うもの。
    # cond->c
    create_table :sidfmd_rule_c_match_categories do |t|
      t.references :sidfmd_rule_cond, foreign_key: true, index: true, null: false

      # 指定したカスタムカテゴリに属するホストの脆弱性を範囲とする。
      # ここはnil不可（条件無しと同じ意味になるので）。
      t.references :sidfmd_category, foreign_key: true, index: true, null: false

      # false=指定したカテゴリのみ true=指定したカテゴリ以外
      t.boolean :inv, null: false, default: false
    end

    # アクション
    create_table :sidfmd_rule_actions do |t|
      t.references :sidfmd_rule, foreign_key: true, index: true, null: false

      # 最後に実行された日時
      t.datetime :last_exec_at, index: true, null: true

      # ログを保存。JSON形式の配列で保存。
      t.text :logs, null: false, :default => '[]'

      t.timestamps null: false
    end

    # アクション カテゴリ宛メール送信(mail_category)
    # action->a
    create_table :sidfmd_rule_a_mail_categories do |t|
      t.references :sidfmd_rule_action, foreign_key: true, index: true, null: false

      # 対象のカスタムカテゴリ。指定したカスタムカテゴリに属するユーザにメールを送信する。
      # nilの場合は全員。
      t.references :sidfmd_category, foreign_key: true, index: true, null: true

      # 送信元ユーザ
      # 登録ユーザかメールアドレスかどちらかを選択する。
      # 選択しないほうは nil を設定すること。
      t.integer :from_user_id, index: true, null: true
      t.text :from_email, null: true

      # メールタイトル
      t.text :subject, null: false

      # メール本文
      t.text :body, null: false
    end
    add_foreign_key :sidfmd_rule_a_mail_categories, :users, column: :from_user_id

    # アクション ユーザ宛メール送信(mail_user)
    # action->a
    create_table :sidfmd_rule_a_mail_users do |t|
      t.references :sidfmd_rule_action, foreign_key: true, index: true, null: false

      # ユーザ(:user)かメールアドレス(:email)のいずれか一方のみが有効。
      # 無効な方はnilにすること。

      # 宛先ユーザ
      # :user か :email のどちらかを指定。指定しない方は nil を設定すること。
      t.references :user, foreign_key: true, index: true, null: true
      t.text :email, null: true

      # 送信元ユーザ
      # 登録ユーザかメールアドレスかどちらかを選択する。
      # 選択しないほうは nil を設定すること。
      t.integer :from_user_id, index: true, null: true
      t.text :from_email, null: true

      # メールタイトル
      t.text :subject, null: false

      # メール本文
      t.text :body, null: false
    end
    add_foreign_key :sidfmd_rule_a_mail_users, :users, column: :from_user_id

    # アクション ステータス変更(status_change)
    # action->a
    create_table :sidfmd_rule_a_status_changes do |t|
      t.references :sidfmd_rule_action, foreign_key: true, index: true, null: false

      # 設定前ステータス1/2
      # status_idの値を設定すると、その値だった場合にのみステータス変更を実行する。
      # 基本的には prev_status1 を使用し、「未対策」「対策中」のように2つ指定する
      # 必要がある場合に prev_status2 を併用する。
      t.integer :prev_status1, null: true, default: nil
      t.integer :prev_status2, null: true, default: nil

      # 設定対象ステータス
      t.integer :status_id, null: false
    end

    # アクション 対策予定日設定（duedate_change)
    # action->a
    create_table :sidfmd_rule_a_duedate_changes do |t|
      t.references :sidfmd_rule_action, foreign_key: true, index: true, null: false

      # trueの場合、既に対策予定日が設定されている場合に上書きする。
      t.boolean :permit_update, null: false, default: false

      # 指定日
      # refdt_daysとは排他。
      t.date :fixed_date, null: true

      # 基準日からの日数
      # fixed_dateとは排他。
      t.integer :refdt_days, null: true
    end

  end

end
