# -*- coding: utf-8 -*-
require 'date'
require 'sidfm_controller_support'

class SidfmVulnerabilitiesController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :update_session_updated_at, :except => [ :get_heat_map_data ]
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license, :except => [ :threshold_setting, :threshold_update ]
  before_action :check_roles

  before_action :get_view_state
  after_action :put_view_state
 
  before_action :get_vm_version
  before_action :clear_import_session
  before_action :clear_export_session, only: [ :index, :index_r ]
  before_action -> { get_summary_table_lines('vulnerabilities') }  

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_date', 'send_mail' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    if [ 'threshold_update' ].include?(action_name) and !@sidfm_roles.include?(:admin) then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
    @cache_prefix = @role_admin ? '/vuls/admin/0' : "/vuls/user/#{@user_category_ids.join(",")}"
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # 現在はJSON/TSV出力でのみ使用しているため、no_cache: trueを付与
  def create_all_content(category_id = 0)
    @vuls[:all_content] = SidfmdIssue.create_all_content(no_cache: true, category_id: @vuls[:category_id], role_admin: @role_admin, corresp_required: @corresp_required, user_host_ids: @user_host_ids, user_category_ids: @user_category_ids)
    @enable_host_ids = SidfmdHost.enable_hosts.collect {|x| x.id }
    @category_host_ids =
      if @vuls[:category_id] > 0 then
        SidfmdCategoriesHost
          .where(:sidfmd_category_id => @vuls[:category_id])
          .select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      else
        SidfmdCategoriesHost.select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      end
  end

  #SidfmdVulViewからcontentを作成する。sortやorderもクエリで行う
  def create_all_vul_view_content(category_id = 0)

    user_vul_view = SidfmdVulView.user_vul_view(@role_admin, @user_host_ids)

    if category_id > 0 then
      if @role_admin then
        @vuls[:all_content] = user_vul_view.where(
          :id => SidfmdVvCategory.where(:sidfmd_category_id => category_id).select(:sidfmd_vul_view_id)
        )
      else
        @vuls[:all_content] = user_vul_view.where(
          :id => SidfmdVvHost.where(:sidfmd_host_id => @user_host_ids).where(:sidfmd_host_id => SidfmdCategoriesHost.where(:sidfmd_category_id => category_id).select(:sidfmd_host_id)).select(:sidfmd_vul_view_id)
        )
      end
    else
      @vuls[:all_content] = user_vul_view.all
    end
  end

  def timeline_able?(category_id = @vuls[:category_id] > 0 ? @vuls[:category_id] : nil)
    if !@role_admin and @vuls[:category_id] == 0 then
      return false
    end
    if category_id then
      unless @role_admin then
        if SidfmdCategory.
            where(:id => category_id).
            where(:is_system => false).
            count == 0 then
          return false
        end
      end
    end
    if SidfmdVulTimeline.where(:sidfmd_category_id => category_id).count == 0 then
      return false
    end
    return true
  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month

  # タイムラインのデータを返す
  def get_timeline_data
    if @vuls[:grp] != :distribution then
      return []
    end
    category_id = @vuls[:category_id] > 0 ? @vuls[:category_id] : nil
    unless timeline_able?(category_id) then
      return []
    end
    ret = {}
    ret[:now] = Time.current
    if SidfmdVulTimeline.
        where(:sidfmd_category_id => category_id).
        where("DATE_PART('year',dt)=?", ret[:now].year).
        where("DATE_PART('month',dt)=?", ret[:now].month).
        count == 0 then
      ret[:now] = ret[:now].ago(1.month)
    end
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @vuls[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end

    flg_fixed = (SidfmdVulTimeline.column_names.include?('sri_fixed') && SidfmdVulTimeline.column_names.include?('cvss_fixed'))

    ret[:timescale].each do |timescale|
      stat = SidfmdVulTimeline.
        where(:sidfmd_category_id => category_id).
        where("DATE_PART('year',dt)=?", ret[:timescale][timescale[:step]][:time].year).
        where("DATE_PART('month',dt)=?", ret[:timescale][timescale[:step]][:time].month).
        order(:dt => :desc).
        limit(1).
        first
      if stat then
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = stat.status_notconfirmed
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = stat.status_ongoing
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = stat.status_fixed
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = stat.status_notaffect
        case @vuls[:idx]
        when :sri then
          ret[:risk_counts][5][timescale[:step]] = stat.sri_fixed if flg_fixed
          ret[:risk_counts][4][timescale[:step]] = stat.sri_critical
          ret[:risk_counts][3][timescale[:step]] = stat.sri_high
          ret[:risk_counts][2][timescale[:step]] = stat.sri_medium
          ret[:risk_counts][1][timescale[:step]] = stat.sri_low
          ret[:risk_counts][0][timescale[:step]] = stat.sri_none
        when :cvss then
          ret[:risk_counts][:fixed][timescale[:step]] = stat.cvss_fixed if flg_fixed
          ret[:risk_counts][:d4][timescale[:step]] = stat.cvss_d4
          ret[:risk_counts][:d3][timescale[:step]] = stat.cvss_d3
          ret[:risk_counts][:d2][timescale[:step]] = stat.cvss_d2
          ret[:risk_counts][:d1][timescale[:step]] = stat.cvss_d1
          ret[:risk_counts][:notreg][timescale[:step]] = stat.cvss_notreg
        end
      else
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = 0
        case @vuls[:idx]
        when :sri then
          ret[:risk_syms].each_key do |sri|
            ret[:risk_counts][sri][timescale[:step]] = 0
          end
        when :cvss then
          ret[:risk_labels].each do |cvss|
            ret[:risk_counts][cvss][timescale[:step]] = 0
          end
        end
      end
    end

    ret
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :category_id, :grp, :idx, :filters, :checked, :all_checked, :anchor)

    @vuls = { :url => request.url }

    @vuls[:category_id] = (pages_params[:category_id] || 0).to_i
    @vuls[:category_name] = '全ての脆弱性'
    if @vuls[:category_id] > 0 then
      cat = SidfmdCategory.find(@vuls[:category_id])
      if cat then
        if @role_admin then
          @vuls[:category_name] = cat.name
        else
          if cat.is_system then
            @vuls[:category_name] = cat.name
          elsif @user_category_ids.include?(@vuls[:category_id]) then
            @vuls[:category_name] = cat.name
          else
            @vuls[:category_id] = 0
          end
        end
      else
        @vuls[:category_id] = 0
      end
    else
      @vuls[:category_id] = 0
    end

    @vuls[:grp] = (pages_params[:grp] || 'count').to_sym
    @vuls[:timeline_able] = timeline_able?
    unless @vuls[:timeline_able] then
      @vuls[:grp] = :count
    end
    @vuls[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @vuls[:idx]
    @vuls[:per_page] = (@summary_table_lines || pages_params[:per_page] || 10).to_i
    @vuls[:offset] = (pages_params[:offset] || 0).to_i
    @vuls[:order] = pages_params[:order] || 'title_asc'
    @vuls[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    @vuls[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @vuls[:all_checked] = pages_params[:all_checked] || false
  end

  def create_content
    @corresp_required = Sidfm.corresp_required

    create_all_content(@vuls[:category_id])

    check_filters

    @vuls[:content_filtered] = apply_filters(@vuls[:all_content])
    @vuls[:total_size] = @vuls[:content_filtered].size
    @vuls[:content] = @vuls[:content_filtered].slice(@vuls[:offset], @vuls[:per_page])
    @vuls[:content] = [] unless @vuls[:content]


    @vuls[:all_vul_size] = @vuls[:all_content].size
    @vuls[:corresp_required_vul_size] = @vuls[:all_vul_size]
    @vuls[:fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:fixed_warning][@vuls[:idx]] ? 1 : 0 }
    @vuls[:not_fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:not_fixed_warning][@vuls[:idx]] ? 1 : 0 }

    @timeline_data = get_timeline_data
  end

  def create_vul_view_content
    @corresp_required = Sidfm.corresp_required

    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories
        .where(:id =>
               SidfmdCategoriesUser
                 .select(:sidfmd_category_id)
                 .where(:user_id => User.current.id)
                 .distinct
                 .collect {|x| x[:sidfmd_category_id] })
    end

    check_filters

    @enable_host_ids = SidfmdHost.enable_hosts.collect {|x| x.id }
    @category_host_ids =
      if @vuls[:category_id] > 0 then
        SidfmdCategoriesHost
          .where(:sidfmd_category_id => @vuls[:category_id])
          .select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      else
        SidfmdCategoriesHost.select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      end

    create_all_vul_view_content(@vuls[:category_id])
    @all_html_ids = @vuls[:all_content].pluck(:html_id)
    #ページ上部のカテゴリのフィルタがある時に使うhost_ids
    @filter_category_host_ids = if @vuls[:category_id] > 0 then
                                  if @role_admin then
                                    SidfmdCategoriesHost.where(:sidfmd_category_id => @vuls[:category_id])
                                                        .pluck(:sidfmd_host_id)
                                  else
                                    SidfmdCategoriesHost.where(:sidfmd_category_id => @vuls[:category_id], :sidfmd_host_id => @user_host_ids)
                                                        .pluck(:sidfmd_host_id)
                                  end
                                else
                                  nil
                                end
    @vuls[:content_filtered] = apply_filters_to_vul_view_content(@vuls[:all_content])
   
    @vuls[:total_size] = @vuls[:content_filtered].count
    if @vuls[:offset] && @vuls[:per_page] then
      @vuls[:content] = (@vuls[:content_filtered].limit(@vuls[:per_page]).offset(@vuls[:offset]) || [])
    end
    @vuls[:all_vul_size] = @vuls[:all_content].count
    @vuls[:corresp_required_vul_size] = @vuls[:all_vul_size] 

    #fixed_corresp_required_vul_sizeとnot_fixed_corresp_required_vul_sizeの処理は、
    #整合性を取るためSidfmdIssue#update_value_by_scsidsに合わせる
    if @vuls[:idx] == :sri then
      if @filter_category_host_ids or @user_host_ids then
        @vuls[:fixed_corresp_required_vul_size] = @vuls[:all_content].where(
          :id => SidfmdVvScsid.where(
                   :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                   :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
                 ).group(:sidfmd_vul_view_id)
                  .having("MIN(status_id) IN (?)", SidfmdIssue::STATUS_FINISHED)
                  .select(:sidfmd_vul_view_id)
        ).where(
          :id => SidfmdVvScsid.where(
            :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
            :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids),
            :status_id => SidfmdIssue::STATUS_UNFINISHED
          ).group(:sidfmd_vul_view_id)
           .having("MAX(sri) >= ?", @corresp_required[:sri])
           .select(:sidfmd_vul_view_id)
        ).count

        @vuls[:not_fixed_corresp_required_vul_size] = @vuls[:all_content].where(
          :id => SidfmdVvScsid.where(
                   :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                   :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
                 ).group(:sidfmd_vul_view_id)
                  .having("MIN(status_id) IN (?)", SidfmdIssue::STATUS_UNFINISHED) 
                  .select(:sidfmd_vul_view_id)
        ).where(
          :id => SidfmdVvScsid.where(
            :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
            :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids),
            :status_id => SidfmdIssue::STATUS_UNFINISHED
          ).group(:sidfmd_vul_view_id)
           .having("MAX(sri) >= ?", @corresp_required[:sri])
           .select(:sidfmd_vul_view_id)
        ).count
      else
        @vuls[:fixed_corresp_required_vul_size] = @vuls[:all_content].where(:fixed_warning_sri => true).count
        @vuls[:not_fixed_corresp_required_vul_size] = @vuls[:all_content].where(:not_fixed_warning_sri => true).count
      end
    else
      if @filter_category_host_ids or @user_host_ids then
        @vuls[:fixed_corresp_required_vul_size] = SidfmdVvScsid.where(
          :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
          :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
        ).group(:sidfmd_vul_view_id)
         .having("MIN(status_id) IN (?) AND (MAX(cvss) IS NULL OR MAX(cvss) >= #{@corresp_required[:cvss]})", SidfmdIssue::STATUS_FINISHED)
         .count
         .size
        @vuls[:not_fixed_corresp_required_vul_size] = SidfmdVvScsid.where(
          :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
          :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
        ).group(:sidfmd_vul_view_id)
         .having("MIN(status_id) IN (?) AND (MAX(cvss) IS NULL OR MAX(cvss) >= #{@corresp_required[:cvss]})", SidfmdIssue::STATUS_UNFINISHED)
         .count
         .size 
      else
        @vuls[:fixed_corresp_required_vul_size] = @vuls[:all_content].where(:fixed_warning_cvss => true).count
        @vuls[:not_fixed_corresp_required_vul_size] = @vuls[:all_content].where(:not_fixed_warning_cvss => true).count
      end
    end
    @timeline_data = get_timeline_data
   
  end
  
  def filter_content
    @vuls[:distribution_status] = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
    @vuls[:distribution_risk] =
      case @vuls[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @vuls[:all_content].each do |i|
      case i[:min_status_id]
      when SidfmdIssue::STATUS_NOTCONFIRMED then @vuls[:distribution_status][:notconfirmed] += 1
      when SidfmdIssue::STATUS_ONGOING then @vuls[:distribution_status][:ongoing] += 1
      when SidfmdIssue::STATUS_FIXED then @vuls[:distribution_status][:fixed] += 1
      when SidfmdIssue::STATUS_NOTAFFECT then @vuls[:distribution_status][:notaffect] += 1
      else @vuls[:distribution_status][:notaffect] += 1
      end
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:min_status_id]) then
        case @vuls[:idx]
        when :sri then
          case i[:max_sri]
          when SidfmdIssueSri::VALUES[0] then
            @vuls[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @vuls[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @vuls[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @vuls[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @vuls[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:cvss][:label]
          when :notreg then
            @vuls[:distribution_risk][:notreg] += 1
          when :d4 then
            @vuls[:distribution_risk][:d4] += 1
          when :d3 then
            @vuls[:distribution_risk][:d3] += 1
          when :d2 then
            @vuls[:distribution_risk][:d2] += 1
          when :d1 then
            @vuls[:distribution_risk][:d1] += 1
          end
        end
      else
        @vuls[:distribution_risk][:fixed] += 1
      end
    end

    @vuls[:dist_cat_vuls] = get_dist_cat_vuls(@vuls[:category_id])
    @vuls[:dist_os_vuls], @vuls[:dist_os_vuls_size] = get_dist_os_vuls(@vuls[:category_id])
    @vuls[:statuses] = SidfmdIssue.statuses

    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories
        .where(:id =>
               SidfmdCategoriesUser
                 .select(:sidfmd_category_id)
                 .where(:user_id => User.current.id)
                 .distinct
                 .collect {|x| x[:sidfmd_category_id] })
    end

    @statuses = IssueStatus.all.order(:id)
  end

  def filter_vul_view_content
    @vuls[:distribution_status] = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
    @vuls[:distribution_risk] =
      case @vuls[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end

    host_filter_necessary = @filter_category_host_ids || @user_host_ids
    if host_filter_necessary then
      vul_scsids = @vuls[:all_content].joins(:sidfmd_vv_scsids)
                                      .where(:sidfmd_vv_scsids => { :sidfmd_host_id =>  (@filter_category_host_ids || @user_host_ids) })
                                      .group("sidfmd_vul_views.id")
      @vuls[:distribution_status][:notconfirmed] = vul_scsids.having("MIN(status_id) = ?", SidfmdIssue::STATUS_NOTCONFIRMED).count.size
      @vuls[:distribution_status][:ongoing] = vul_scsids.having("MIN(status_id) = ?", SidfmdIssue::STATUS_ONGOING).count.size
      @vuls[:distribution_status][:fixed] = vul_scsids.having("MIN(status_id) = ?", SidfmdIssue::STATUS_FIXED).count.size
    else
      @vuls[:distribution_status][:notconfirmed] = @vuls[:all_content].where(:min_status_id => SidfmdIssue::STATUS_NOTCONFIRMED).count
      @vuls[:distribution_status][:ongoing] = @vuls[:all_content].where(:min_status_id => SidfmdIssue::STATUS_ONGOING).count
      @vuls[:distribution_status][:fixed] = @vuls[:all_content].where(:min_status_id => SidfmdIssue::STATUS_FIXED).count
    end
    @vuls[:distribution_status][:notaffect] = @vuls[:all_content].count - (
      @vuls[:distribution_status][:notconfirmed] +
      @vuls[:distribution_status][:ongoing] +
      @vuls[:distribution_status][:fixed]
    )
   
    if host_filter_necessary then
      case @vuls[:idx]
      when :sri then
        unfinished_vuls = vul_scsids.where(:sidfmd_vv_scsids => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
        @vuls[:distribution_risk][:critical] = unfinished_vuls.having("MAX(sri) = ?", SidfmdIssueSri::VALUES[0]).count.size
        @vuls[:distribution_risk][:high] = unfinished_vuls.having("MAX(sri) = ?", SidfmdIssueSri::VALUES[1]).count.size
        @vuls[:distribution_risk][:medium] = unfinished_vuls.having("MAX(sri) = ?", SidfmdIssueSri::VALUES[2]).count.size
        @vuls[:distribution_risk][:low] = unfinished_vuls.having("MAX(sri) = ?", SidfmdIssueSri::VALUES[3]).count.size
        @vuls[:distribution_risk][:none] = unfinished_vuls.having("MAX(sri) = ?", SidfmdIssueSri::VALUES[4]).count.size
      when :cvss then
        unfinished_vuls = vul_scsids.having("MIN(status_id) IN (?)", SidfmdIssue::STATUS_UNFINISHED)
        @vuls[:distribution_risk][:notreg] = unfinished_vuls.having("MAX(cvss) IS NULL").count.size
        @vuls[:distribution_risk][:d4] = unfinished_vuls.having("MAX(cvss) >= ?", SidfmdIssueCve::LABEL[:d4][:bottom]).count.size
        @vuls[:distribution_risk][:d3] = unfinished_vuls.having("MAX(cvss) < ? AND MAX(cvss) >= ?", SidfmdIssueCve::LABEL[:d3][:top], SidfmdIssueCve::LABEL[:d3][:bottom]).count.size
        @vuls[:distribution_risk][:d2] = unfinished_vuls.having("MAX(cvss) < ? AND MAX(cvss) >= ?", SidfmdIssueCve::LABEL[:d2][:top], SidfmdIssueCve::LABEL[:d2][:bottom]).count.size
        @vuls[:distribution_risk][:d1] = unfinished_vuls.having("MAX(cvss) < ?", SidfmdIssueCve::LABEL[:d1][:top]).count.size
      end
      @vuls[:distribution_risk][:fixed] = vul_scsids.having("MIN(status_id) NOT IN (?)", SidfmdIssue::STATUS_UNFINISHED).count.size
    else
      unfinished_vuls = @vuls[:all_content].where(:min_status_id => SidfmdIssue::STATUS_UNFINISHED)
      case @vuls[:idx]
      when :sri then
        @vuls[:distribution_risk][:critical] = unfinished_vuls.where(:max_sri => SidfmdIssueSri::VALUES[0]).count
        @vuls[:distribution_risk][:high] = unfinished_vuls.where(:max_sri => SidfmdIssueSri::VALUES[1]).count
        @vuls[:distribution_risk][:medium] = unfinished_vuls.where(:max_sri => SidfmdIssueSri::VALUES[2]).count
        @vuls[:distribution_risk][:low] = unfinished_vuls.where(:max_sri => SidfmdIssueSri::VALUES[3]).count
        @vuls[:distribution_risk][:none] = unfinished_vuls.where(:max_sri => SidfmdIssueSri::VALUES[4]).count 
      when :cvss then
        @vuls[:distribution_risk][:notreg] = unfinished_vuls.where(:cvss_label => 'notreg').count
        @vuls[:distribution_risk][:d4] = unfinished_vuls.where(:cvss_label => 'd4').count
        @vuls[:distribution_risk][:d3] = unfinished_vuls.where(:cvss_label => 'd3').count
        @vuls[:distribution_risk][:d2] = unfinished_vuls.where(:cvss_label => 'd2').count
        @vuls[:distribution_risk][:d1] = unfinished_vuls.where(:cvss_label => 'd1').count
      end
      @vuls[:distribution_risk][:fixed] = @vuls[:all_content].where.not(:min_status_id => SidfmdIssue::STATUS_UNFINISHED).count
    end
    #@vuls[:dist_cat_vuls] = get_dist_cat_vuls(@vuls[:category_id])
    #@vuls[:dist_os_vuls], @vuls[:dist_os_vuls_size] = get_dist_os_vuls(@vuls[:category_id])
    create_vul_view_dist_hosts(@vuls[:category_id])
    @vuls[:dist_cat_vuls] = get_vul_view_dist_cat_vuls 
    @vuls[:dist_os_vuls], @vuls[:dist_os_vuls_size] = get_vul_view_dist_os_vuls
    @vuls[:statuses] = SidfmdIssue.statuses

    @statuses = IssueStatus.all.order(:id)
  end

  FILTER_KEYS = {}

  # フィルター情報の初期化
  def init_filter_keys
    #@due_dates = @vuls[:all_content].inject([]) {|r, x|
    #  r += x[:due_dates]
    #}.uniq.collect {|x| x.nil? ? '' : x }.sort

    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:warning] = { :values => [ 'warning', 'others', 'all' ].map(&:freeze).freeze, :default => 'all'.freeze }
      FILTER_KEYS[:status] = {
        :values => (SidfmdIssue::STATUS_ALL.collect {|x| x.to_s } +
                    [ SidfmdIssue::STATUS_UNFINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ SidfmdIssue::STATUS_FINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ '0' ]).map(&:freeze).freeze, :default => '0'.freeze }
      FILTER_KEYS[:sris] = { :values => nil, :default => [].freeze }
      FILTER_KEYS[:cvsses] = { :values => nil, :default => [].freeze }
      FILTER_KEYS[:area] = { :values => (SidfmdArea::VALUES.collect {|x| x.to_s } + [ '' ]).map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:exploitable] = { :values => [ 'true', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:categories] = {
        :values => nil,
        :default => {
          :ids => [].freeze,
          :is_and => false,
          :is_except => false
        }
      }
      FILTER_KEYS[:due_date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:keyword] = { :values => nil, :default => '' }
      FILTER_KEYS[:date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:products] = { 
        :values => nil,
        :default => {
          :product_ids => [],
          :version_ids => [],
          :lproduct_ids => [],
          :lversion_ids => []
        }
      }
      FILTER_KEYS[:cvss_vector] = { :values => nil, :default => {} }
      FILTER_KEYS[:ctags] = { :values => nil, :default => { :ids => [] } }
    end
    @default_filters = {}
    FILTER_KEYS.each do |key, value|
      @default_filters[key] = FILTER_KEYS[key][:default]
    end
  end

  # フィルタのチェックと正規化
  def check_filters
    init_filter_keys
    f_keys = FILTER_KEYS.keys
    for i in 0..f_keys.size - 1 do
      if @vuls[:filters].has_key?(f_keys[i]) then
        if FILTER_KEYS[f_keys[i]][:values] && !FILTER_KEYS[f_keys[i]][:values].include?(@vuls[:filters][f_keys[i]]) then
          @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
        end
      else
        @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
      end
    end
    @vuls[:filters][:categories][:ids] = @vuls[:filters][:categories][:ids].map(&:to_i)
    FILTER_KEYS[:products][:default].keys.each do |key|
      @vuls[:filters][:products][key] = [] unless @vuls[:filters][:products][key]
      @vuls[:filters][:products][key] = @vuls[:filters][:products][key].map(&:to_i)
    end
  end

  DAY_SECONDS = 24 * 60 * 60

  # フィルタの適用
  # 未使用。カスタムタグ検索未実装
  def apply_filters(all_content)

    if @vuls[:filters][:cvss_vector].size > 0 && !@vuls[:filters][:cvss_vector][:is_target_unregistered] then
      #CVSS指標フィルタ用のデータ
      filter_cvss_v2 = ''
      filter_cvss_v3 = ''
      cvss_v2_vectors = SidfmdIssueCve.get_cvss_v2_vectors.keys.map{|vector| vector.to_sym}
      cvss_v3_vectors = SidfmdIssueCve.get_cvss_v3_vectors.keys.map{|vector| vector.to_sym}
      selected_cvss_v2 = @vuls[:filters][:cvss_vector][:cvss_v2]
      selected_cvss_v3 = @vuls[:filters][:cvss_vector][:cvss_v3]
      is_selected_cvss_v2 = false
      is_selected_cvss_v3 = false
      cvss_v2_vectors.each_with_index do |vector, idx|
        is_selected_cvss_v2 = true if selected_cvss_v2[vector]
        filter_cvss_v2 += '/' unless idx == 0
        filter_cvss_v2 += "#{vector}:#{(selected_cvss_v2[vector] || '.')}"
      end
      cvss_v3_vectors.each_with_index do |vector, idx|
        is_selected_cvss_v3 = true if selected_cvss_v3[vector]
        filter_cvss_v3 += '/' unless idx == 0
        filter_cvss_v3 += "#{vector}:#{(selected_cvss_v3[vector] || '.')}"
      end
      filter_cvss_v2 = Regexp.new("\\A" + filter_cvss_v2 + "\\z")
      filter_cvss_v3 = Regexp.new("\\A" + filter_cvss_v3 + "\\z")
    end

    #プロダクト検索は、クエリの発行が少なくなるように初めに該当するプロダクトだけ取得する
    prod_searched = false
    if @vuls[:filters][:product].size > 0 then
      prod_info = @vuls[:filters][:product].symbolize_keys
      prod_searched = true
      if prod_info[:pid] then
        #productが検索対象
        prod_issue_ids = SidfmdIssueAffprod.where(:product_id => prod_info[:pid]).pluck(:sidfmd_issue_id) + 
                         SidfmdIssueAffver.where(:product_id => prod_info[:pid]).pluck(:sidfmd_issue_id)
        issue_html_ids = SidfmdIssue.where(:id => prod_issue_ids).pluck(:html_id)
      elsif prod_info[:vid] then
        #versionが検索対象
        prod_issue_ids = SidfmdIssueAffver.where(:product_version_id => prod_info[:vid]).pluck(:sidfmd_issue_id)
        issue_html_ids = SidfmdIssue.where(:id => prod_issue_ids).pluck(:html_id)
      elsif prod_info[:lpid] then
        #lproductが検索対象
        issue_html_ids = SidfmdLocalVul.joins(:sidfmd_local_vul_lapps)
                                       .joins('LEFT OUTER JOIN sidfmd_issues ON sidfmd_issues.html_id = sidfmd_local_vuls.html_id')
                                       .where('sidfmd_local_vul_lapps.sidfmd_lproduct_id = ?', prod_info[:lpid])
                                       .select('sidfmd_issues.html_id AS html_id')
                                       .distinct
                                       .pluck(:html_id)
      elsif prod_info[:lvid] then
        #lversionが検索対象
        issue_html_ids = SidfmdLocalVul.joins(:sidfmd_local_vul_lapps)
                                       .joins('LEFT OUTER JOIN sidfmd_issues ON sidfmd_issues.html_id = sidfmd_local_vuls.html_id')
                                       .where('sidfmd_local_vul_lapps.sidfmd_lversion_id = ?', prod_info[:lvid])
                                       .select('sidfmd_issues.html_id AS html_id')
                                       .distinct
                                       .pluck(:html_id)
      else
        prod_searched = false #検索リセット時
      end
    end

    ret = all_content.find_all {|item|
      #プロダクト検索
      if prod_searched then
        next false unless issue_html_ids.include?(item[:html_id].to_i)
        #next false if (prods_related_host_ids - item[:hosts].map{|h| h[:id]}).size == prods_related_host_ids.size
      end
      # カテゴリ
      if @vuls[:category_id] > 0 then
        next false unless (item[:categories].collect {|x| x[:id] }).include?(@vuls[:category_id])
      end
      # 要注意
      if @vuls[:filters][:warning] == 'warning' then
        next false unless item[:warning][@vuls[:idx]]
      elsif @vuls[:filters][:warning] == 'others' then
        next false if item[:warning][@vuls[:idx]]
      end
      sri_cvss_status_ids = if @vuls[:filters][:warning] == 'warning' then
                              item[:sri_cvss_status_ids].find_all {|x| case @vuls[:idx] when :sri then x[:sri] >= @corresp_required[:sri]; when :cvss then x[:cvss] ? x[:cvss] >= @corresp_required[:cvss] : true; end }
                            else
                              item[:sri_cvss_status_ids]
                            end
      # 対応状況(new)
      # min_status_idで判断
      if @vuls[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:min_status_id] == status1 or item[:min_status_id] == status2)
        else
          next false unless item[:min_status_id] == @vuls[:filters][:status].to_i
        end
      end

      # 対応状況(old)
      # sri_cvss_status_idsで判断
=begin
      if @vuls[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (sri_cvss_status_ids.find {|x| x[:status_id] == status1} or sri_cvss_status_ids.find {|x| x[:status_id] == status2 })
        else
          next false unless
            (sri_cvss_status_ids.size == 0 and @vuls[:filters][:status].to_i == SidfmdIssue::STATUS_NOTCONFIRMED) or
            (sri_cvss_status_ids.find {|x| x[:status_id] == @vuls[:filters][:status].to_i })
        end
      end
=end
      # 最大SRI
      if @vuls[:filters][:sris].any? then
        next false unless @vuls[:filters][:sris].include?(item[:max_sri].to_s)
      end
      # 最大CVSS
      if @vuls[:filters][:cvsses].any? then
         next false unless @vuls[:filters][:cvsses].include?(item[:cvss][:label].to_s)
      end
      # CVSS指標
      if @vuls[:filters][:cvss_vector].size > 0 then
        is_hit = false
        vul_cves = get_filtered_cves(item[:cves], false)
        if @vuls[:filters][:cvss_vector][:is_target_unregistered] then
          vul_cves.each do |cve|
            break if is_hit
            is_hit = true if cve[:c2_vector].empty? && cve[:c3_vector].empty?
          end
        else
          vul_cves.each do |cve|
            if is_selected_cvss_v2 and is_selected_cvss_v3 then
              is_hit = true if filter_cvss_v2 =~ cve[:c2_vector] and filter_cvss_v3 =~ cve[:c3_vector]
            elsif is_selected_cvss_v3 then
              is_hit = true if filter_cvss_v3 =~ cve[:c3_vector]
            else
              is_hit = true if filter_cvss_v2 =~ cve[:c2_vector]
            end
            break if is_hit
          end
        end
        next false unless is_hit
      end
      # SofTek影響範囲
      if @vuls[:filters][:area] != '' then
        next false unless (item[:html_id] > 0 and @vuls[:filters][:area].to_i == item[:softek][:area_id])
      end
      # 攻撃コード
      if @vuls[:filters][:exploitable] != '' then
        next false unless item[:exploitable]
      end
      #カテゴリ(複数)
      if @vuls[:filters][:categories][:ids].any? then
        pass = false
        all_cat_ids = item[:categories].map{|cc| cc[:id]}
        if @vuls[:filters][:categories][:is_and] then
          if (@vuls[:filters][:categories][:ids] - all_cat_ids).empty? then
            pass = true
          end
        elsif @vuls[:filters][:categories][:is_except] then
          if (@vuls[:filters][:categories][:ids] - all_cat_ids).size == @vuls[:filters][:categories][:ids].size then
            pass = true
          end
        else
          if (@vuls[:filters][:categories][:ids] - all_cat_ids).size != @vuls[:filters][:categories][:ids].size then
            pass = true
          end
        end
        next false unless pass
      end
      # 対策予定日設定済
      if @vuls[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @vuls[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless item[:due_dates].include?(nil)
          else
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              log_exception(e)
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              log_exception(e)
              enddate = nil
            end
            due_date = nil
            if begindate and enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate and d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif begindate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            end
          end
        end
      end
      # キーワード検索
      if @vuls[:filters][:keyword].size > 0 then
        if /\A\s*(\d+)\s*\z/ =~ @vuls[:filters][:keyword] then
          html_id = Regexp.last_match(1).to_i
          regexp = Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE)
          next false unless (regexp =~ item[:title] or item[:cves].select{|cve| (regexp =~ cve[:cve_id]) != nil}.any? or html_id == item[:html_id])
        elsif /\A\s*KB(\d+)\s*\z/i =~ @vuls[:filters][:keyword] then
          kbid = -Regexp.last_match(1).to_i
          next false unless (Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title] or kbid == item[:html_id])
        else
          regexp = Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE)
          next false unless regexp =~ item[:title] or item[:cves].select{|cve| (regexp =~ cve[:cve_id]) != nil}.any?
        end
      end
      # 日付指定検索
      if @vuls[:filters][:date] != '::' then
        if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @vuls[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @vuls[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if begin_date <= item[:min_created_on][:value] and end_date >= item[:min_created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false if begin_date <= item[:min_created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if end_date >= item[:min_created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless begin_date <= item[:min_created_on][:value] and end_date >= item[:min_created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless begin_date <= item[:min_created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless end_date >= item[:min_created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          end
        else
          @vuls[:filters][:date] = '::'
          next false
        end
      end
      true
    }
    #ret.sort! {|x, y|
    #  SidfmdIssue::SORT_PRIORITIES[x[:min_status_id]] <=> SidfmdIssue::SORT_PRIORITIES[y[:min_status_id]] ||
    #  y[:min_created_on][:value] <=> x[:min_created_on][:value]
    #}
    case @vuls[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'statuses_asc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ x[:title].downcase, m, 500 ]
        else
          [ x[:title].downcase, '', 0 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'statuses_desc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ x[:title].downcase, m, 500 ]
        else
          [ x[:title].downcase, '', 0 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'maxsri_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:max_sri] ? x[:max_sri] : -100 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'maxsri_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:max_sri] ? x[:max_sri] : -100 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'cvss_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:html_id].abs, x[:cvss][:score] ? x[:cvss][:score] : -10, x[:html_id] < 0 ? 100 : 10 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[4] <=> yy[4]
        if r == 0 then
          r = yy[3] <=> xx[3]
          if r == 0 then
            r = xx[2] <=> yy[2]
            if r == 0 then
              r = yy[1] <=> xx[1]
              if r == 0 then
                r = xx[0] <=> yy[0]
              end
            end
          end
        end
        r
      }
    when 'cvss_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:html_id].abs, x[:cvss][:score] ? x[:cvss][:score] : -10, x[:html_id] < 0 ? 100 : 10 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[4] <=> yy[4]
        if r == 0 then
          r = yy[3] <=> xx[3]
          if r == 0 then
            r = xx[2] <=> yy[2]
            if r == 0 then
              r = yy[1] <=> xx[1]
              if r == 0 then
                r = xx[0] <=> yy[0]
              end
            end
          end
        end
        #r = yy[2] <=> xx[2]
        #if r == 0 then
        #  r = yy[1] <=> xx[1]
        #  if r == 0 then
        #    r = xx[0] <=> yy[0]
        #  end
        #end
        r
      }
    when 'area_asc' then
      temp_ev = proc {|x|
        [
         x[:title].downcase,
         x[:min_created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'area_desc' then
      temp_ev = proc {|x|
        [
         x[:title].downcase,
         x[:min_created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase,
          x[:min_created_on][:display],
          x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase,
          x[:min_created_on][:display],
          x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'elapseddays_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:min_elapsed_created_on][:display] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'elapseddays_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:min_created_on][:display], x[:min_elapsed_created_on][:display] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'title_asc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:html_id].abs ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'title_desc' then
      temp_ev = proc {|x|
        [ x[:title].downcase, x[:html_id].abs ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'hostnames_asc' then
      ret.sort! {|x, y|
        r = x[:hosts].sort_by{|h| h[:display_name].downcase}.first[:display_name].downcase <=> y[:hosts].sort_by{|h| h[:display_name].downcase}.first[:display_name].downcase
        if r == 0 then
          r = x[:title].downcase <=> y[:title].downcase
        end
        r
        #(x[:hosts].first[:ext_name] ? x[:hosts].first[:ext_name].downcase : x[:hosts].first[:hostname].downcase) <=> (y[:hosts].first[:ext_name] ? y[:hosts].first[:ext_name].downcase : y[:hosts].first[:hostname].downcase)
        #x[:hosts].first[:hostname] <=> y[:hosts].first[:hostname]
      }
    when 'hostnames_desc' then
      ret.sort! {|y, x|
        r = x[:hosts].sort_by{|h| h[:display_name].downcase}.first[:display_name].downcase <=> y[:hosts].sort_by{|h| h[:display_name].downcase}.first[:display_name].downcase
        if r == 0 then
          r = x[:title].downcase <=> y[:title].downcase
        end
        r
        #(x[:hosts].first[:ext_name] ? x[:hosts].first[:ext_name].downcase : x[:hosts].first[:hostname].downcase) <=> (y[:hosts].first[:ext_name] ? y[:hosts].first[:ext_name].downcase : y[:hosts].first[:hostname].downcase)
        #x[:hosts].first[:hostname] <=> y[:hosts].first[:hostname]
      }
    when 'categories_asc' then
      temp_ev = proc {|x|
        scat = x[:categories]  ? x[:categories].sort_by{|c| c[:id]}.find {|y| y[:is_system] } : nil
        if scat then
          [ x[:title].downcase, x[:min_created_on][:display], scat[:id] ]
        else
          if x[:categories].size > 0 then
            [ x[:title].downcase, x[:min_created_on][:display], x[:categories][0][:id] ]
          else
            [ x[:title].downcase, x[:min_created_on][:display], 99999 ]
          end
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'categories_desc' then
      temp_ev = proc {|x|
        scat = x[:categories]  ? x[:categories].sort_by{|c| c[:id]}.find {|y| y[:is_system] } : nil
        if scat then
          [ x[:title].downcase, x[:min_created_on][:display], scat[:id] ]
        else
          if x[:categories].size > 0 then
            [ x[:title].downcase, x[:min_created_on][:display], x[:categories][0][:id] ]
          else
            [ x[:title].downcase, x[:min_created_on][:display], 99999 ]
          end
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'duedates_asc' then
      temp_ev = proc {|x|
        #p = view_context.display_due_dates_percent_value(x[:due_dates])
        #[ x[:title].downcase, x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
        f_cnt = x[:due_dates_unfinished].size
        if f_cnt > 0 and x[:due_dates_unfinished].compact.size > 0 then
          cnt = x[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
          p = cnt.to_f / f_cnt.to_f * 100.0
        else
          p = 0.0
        end
        [ x[:title].downcase, x[:min_created_on][:display], p ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'duedates_desc' then
      temp_ev = proc {|x|
        #p = view_context.display_due_dates_percent_value(x[:due_dates])
        #[ x[:title].downcase, x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
        f_cnt = x[:due_dates_unfinished].size
        if f_cnt > 0 and x[:due_dates_unfinished].compact.size > 0 then
          cnt = x[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
          p = cnt.to_f / f_cnt.to_f * 100.0
        else
          p = 0.0
        end
        [ x[:title].downcase, x[:min_created_on][:display], p ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'finished_asc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]).round(1) : nil
        [ x[:title].downcase, x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    when 'finished_desc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]).round(1) : nil
        [ x[:title].downcase, x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    end
    ret
  end

  def apply_filters_to_vul_view_content(all_content)
    get_summary_prior_category unless @summary_prior_category
    #プロダクト検索
    if @vuls[:filters][:products].values.select{|v| v.size > 0 }.any? then
      prod_info = @vuls[:filters][:products].symbolize_keys
      issue_html_ids = []
      if prod_info[:product_ids] and prod_info[:product_ids].any? then
        #productが検索対象
        issues = SidfmdIssue.where(
          "id IN (?) OR id IN (?)", 
          SidfmdIssueAffprod.where(:product_id => prod_info[:product_ids]).select(:sidfmd_issue_id),
          SidfmdIssueAffver.where(:product_id => prod_info[:product_ids]).select(:sidfmd_issue_id)
        )
        unless @role_admin then
          issues = issues.where(:sidfmd_host_id => @user_host_ids)
        end
        issue_html_ids |= issues.pluck(:html_id)
      end
      if prod_info[:version_ids] and prod_info[:version_ids].any? then
        #versionが検索対象
        issues = SidfmdIssue.where(
          :id => SidfmdIssueAffver.where(:product_version_id => prod_info[:version_ids]).pluck(:sidfmd_issue_id)
        )
        unless @role_admin then
          issues = issues.where(:sidfmd_host_id => @user_host_ids)
        end
        issue_html_ids |= issues.pluck(:html_id)
      end
      if prod_info[:lproduct_ids] and prod_info[:lproduct_ids].any? then
        #lproductが検索対象
        vuls = SidfmdLocalVul.joins(:sidfmd_local_vul_lapps)
                             .joins('LEFT OUTER JOIN sidfmd_issues ON sidfmd_issues.html_id = sidfmd_local_vuls.html_id')
                             .where('sidfmd_local_vul_lapps.sidfmd_lproduct_id IN (?)', prod_info[:lproduct_ids])
                             .select('sidfmd_issues.html_id AS html_id')
        unless @role_admin then
          vuls = vuls.where(:sidfmd_issues => { :sidfmd_host_id => @user_host_ids })
        end
        issue_html_ids |= vuls.pluck(:html_id)                      
      end
      if prod_info[:lversion_ids] and prod_info[:lversion_ids].any? then
        #lversionが検索対象
        vuls = SidfmdLocalVul.joins(:sidfmd_local_vul_lapps)
                             .joins('LEFT OUTER JOIN sidfmd_issues ON sidfmd_issues.html_id = sidfmd_local_vuls.html_id')
                             .where('sidfmd_local_vul_lapps.sidfmd_lversion_id IN (?)', prod_info[:lversion_ids])
                             .select('sidfmd_issues.html_id AS html_id') 
        unless @role_admin then
          vuls = vuls.where(:sidfmd_issues => { :sidfmd_host_id => @user_host_ids })
        end
        issue_html_ids |= vuls.pluck(:html_id)
      end
     
      if issue_html_ids.any? then
        all_content = all_content.where(:html_id => issue_html_ids)
      else
        all_content = all_content.where(:html_id => 0) 
      end
    end

    # カテゴリ
    #if @vuls[:category_id] > 0 then
    #  all_content = all_content.where(
    #    :id => SidfmdVvCategory.where(:sidfmd_category_id => @vuls[:category_id]).select(:sidfmd_vul_view_id)
    #  )
    #end

    # 要注意
    if @vuls[:filters][:warning] == 'warning' || @vuls[:filters][:warning] == 'others' then
      if @filter_category_host_ids or @user_host_ids then
        #SidfmdIssue#update_value_by_scsidsの処理に合わせる
        filter_by_warning_sri = ->(content){
          if @vuls[:filters][:warning] == 'warning' then
            content.where(
              :id => SidfmdVvScsid.where(
                       :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                       :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
                     ).group(:sidfmd_vul_view_id)
                      .having("MIN(status_id) IN (?)", SidfmdIssue::STATUS_UNFINISHED)
                      .select(:sidfmd_vul_view_id)
            ).where(
              :id => SidfmdVvScsid.where(
                       :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                       :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids),
                       :status_id => SidfmdIssue::STATUS_UNFINISHED
                     ).group(:sidfmd_vul_view_id)
                      .having("MAX(sri) >= ?", @corresp_required[:sri])
                      .select(:sidfmd_vul_view_id)
            )
          else
            content.where(%|
             NOT id IN (?) OR NOT id IN (?)  
            | , SidfmdVvScsid.where(
                 :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                 :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)
               ).group(:sidfmd_vul_view_id)
                .having("MIN(status_id) IN (?)", SidfmdIssue::STATUS_UNFINISHED)
                .select(:sidfmd_vul_view_id),
               SidfmdVvScsid.where(
                 :sidfmd_vul_view_id => @vuls[:all_content].select(:id),
                 :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids),
                 :status_id => SidfmdIssue::STATUS_UNFINISHED
               ).group(:sidfmd_vul_view_id)
                .having("MAX(sri) >= ?", @corresp_required[:sri])
                .select(:sidfmd_vul_view_id) 
            ) 
          end
        }

        filter_by_warning_cvss = ->(content){
          content.where("id #{@vuls[:filters][:warning] == 'warning' ? '' : 'NOT'} IN (?)",
            SidfmdVvScsid.where(:sidfmd_host_id => (@filter_category_host_ids || @user_host_ids))
                         .group(:sidfmd_vul_view_id)
                         .having(%|
                           MIN(status_id) IN (?) AND 
                           #{@vuls[:idx] == :sri ? "MAX(sri) >= #{@corresp_required[:sri]}" : "(MAX(cvss) IS NULL OR MAX(cvss) >= #{@corresp_required[:cvss]})"}
                         |, SidfmdIssue::STATUS_UNFINISHED)
                         .select(:sidfmd_vul_view_id)
         )
        }
        all_content = if @vuls[:idx] == :sri then
                        filter_by_warning_sri.(all_content)
                      else
                        filter_by_warning_cvss.(all_content)
                      end
      else
        if @vuls[:filters][:warning] == 'warning' then
          all_content = @vuls[:idx] == :sri ? all_content.where(:warning_sri => true) : all_content.where(:warning_cvss => true)
        else
          all_content = @vuls[:idx] == :sri ? all_content.where.not(:warning_sri => true) : all_content.where.not(:warning_cvss => true)
        end
      end

    end
  
    # 対応状況
    # min_status_idで判断
    if @vuls[:filters][:status] != '0' then
      filter_by_status = ->(content, host_ids, status_ids){
        content = if host_ids.any? then
                    content.joins(%|
                      LEFT OUTER JOIN(
                        SELECT
                          sidfmd_vul_view_id AS vul_view_id,
                          MIN(status_id) AS vv_min_status_id
                        FROM
                          sidfmd_vv_scsids
                        WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))
                        GROUP BY sidfmd_vul_view_id
                      ) AS vv_min_status
                      ON sidfmd_vul_views.id = vv_min_status.vul_view_id
                    |).where('vv_min_status_id IN (?)', status_ids)
                  else
                    content.where(:id => 0)
                  end
      }
   
      if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
        status1 = Regexp.last_match(1).to_i
        status2 = Regexp.last_match(2).to_i
        if @filter_category_host_ids then
          all_content = filter_by_status.(all_content, @filter_category_host_ids, [ status1, status2 ])
        else
          all_content = if @role_admin then
                          all_content.where(:min_status_id => [ status1, status2 ])
                        else
                          filter_by_status.(all_content, @user_host_ids, [ status1, status2 ])
                        end
        end
      else
        if @filter_category_host_ids then
          all_content = filter_by_status.(all_content, @filter_category_host_ids, [ @vuls[:filters][:status] ])
        else
          all_content = if @role_admin then
                          all_content.where(:min_status_id => @vuls[:filters][:status])
                        else
                          filter_by_status.(all_content, @user_host_ids, [ @vuls[:filters][:status] ])
                        end
        end
      end
    end

    # 最大SRI
    if @vuls[:filters][:sris].any? then
      filter_by_sri = ->(content, filter_hosts){
        content = if @vuls[:filters][:sris].include?(SidfmdIssueSri::VALUE_NONE.to_s) then 
                    content.joins(%|
                      LEFT OUTER JOIN(
                        SELECT
                          sidfmd_vul_view_id AS vul_view_id,
                          MAX(sri) AS max_sri_value
                        FROM
                          sidfmd_vv_scsids
                        WHERE 
                          #{filter_hosts}
                          status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                        GROUP BY sidfmd_vul_view_id           
                      ) AS vv_max_sri
                      ON sidfmd_vul_views.id = vv_max_sri.vul_view_id 
                    |).joins(%|
                      LEFT OUTER JOIN(
                        SELECT DISTINCT
                          sidfmd_vul_view_id AS vul_view_id2
                        FROM
                          sidfmd_vv_scsids 
                        WHERE
                          #{filter_hosts}
                          status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                        GROUP BY sidfmd_vul_view_id 
                      ) AS vv_scsids 
                      ON sidfmd_vul_views.id = vv_scsids.vul_view_id2
                    |)
                    .where('max_sri_value IN (?) OR vul_view_id2 IS NULL', @vuls[:filters][:sris])
                  else
                    content.joins(%|
                      LEFT OUTER JOIN(
                        SELECT
                          sidfmd_vul_view_id AS vul_view_id,
                          MAX(sri) AS max_sri_value
                        FROM
                          sidfmd_vv_scsids
                        WHERE 
                          #{filter_hosts}
                          status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                        GROUP BY sidfmd_vul_view_id           
                      ) AS vv_max_sri
                      ON sidfmd_vul_views.id = vv_max_sri.vul_view_id 
                    |).where('max_sri_value IN (?)',@vuls[:filters][:sris])
                  end
      }
      if @filter_category_host_ids then
        if @filter_category_host_ids.any? then
          all_content = filter_by_sri.(all_content, "sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}])) AND")
        end
      elsif @role_admin || @user_host_ids.any? then
        all_content = filter_by_sri.(all_content, (@role_admin ? '' : "sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}])) AND"))
      else
        all_content = all_content.where(:id => 0) 
      end
    end

    # 最大CVSS
    if @vuls[:filters][:cvsses].any? then
      label_values = {
        :notreg => nil,
        :d4 => { :top => 10.0 , :bottom => 9.0 },
        :d3 => { :top => 8.9, :bottom => 7.0 },
        :d2 => { :top => 6.9, :bottom => 4.0 },
        :d1 => { :top => 3.9, :bottom => 0.0 },
      }
      conds = []
      @vuls[:filters][:cvsses].each do |label|
        values = label_values[label.to_sym]
        if values then
          conds << "vv_max_cvss_score >= #{values[:bottom]} AND vv_max_cvss_score <= #{values[:top]}"
        else
          conds << "vv_max_cvss_score IS NULL"   
        end
      end
      condition = conds.join(' OR ')

      filter_by_cvss = ->(content, condition, filter_hosts){
        content = content.joins(%|
                    LEFT OUTER JOIN(
                      SELECT
                        sidfmd_vul_view_id AS vul_view_id,
                        MAX(cvss) AS vv_max_cvss_score
                      FROM
                        sidfmd_vv_scsids
                      #{filter_hosts}
                      GROUP BY sidfmd_vul_view_id
                    ) AS vv_max_cvss
                    ON sidfmd_vul_views.id = vv_max_cvss.vul_view_id
                  |).where(condition)
      }
      if @filter_category_host_ids then
        if @filter_category_host_ids.any? then
          all_content = filter_by_cvss.(all_content, condition, "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}]))")
        end
      elsif @role_admin or @user_host_ids.any? then 
        all_content = filter_by_cvss.(all_content, condition, (@role_admin ? '' : "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}]))"))
      else
        all_content = all_content.where(:id => 0)
      end
    end

    #CVSS指標
    if @vuls[:filters][:cvss_vector].size > 0 then
      if @vuls[:filters][:cvss_vector][:is_target_unregistered] then
        condition = %|
          c2_av IS NULL AND
          c2_ac IS NULL AND
          c2_au IS NULL AND
          c2_c IS NULL AND
          c2_i IS NULL AND
          c2_a IS NULL AND
          c3_av IS NULL AND
          c3_ac IS NULL AND
          c3_pr IS NULL AND
          c3_ui IS NULL AND
          c3_s IS NULL AND
          c3_c IS NULL AND
          c3_i IS NULL AND
          c3_a IS NULL
        |
      else
        cvss_filter_items = []
        @vuls[:filters][:cvss_vector][:cvss_v2].each do |vector, value|
          if value then
            cvss_filter_items << "c2_#{vector.to_s.downcase} = '#{value}'"
          end
        end
        @vuls[:filters][:cvss_vector][:cvss_v3].each do |vector, value|
          if value then
            cvss_filter_items << "c3_#{vector.to_s.downcase} = '#{value}'"
          end
        end
        condition = cvss_filter_items.join(' AND ')
      end
      all_content = all_content.where(%|
        id IN (
          SELECT sidfmd_vul_view_id 
          FROM sidfmd_vv_cves AS cve_table
          WHERE( 
            #{condition} AND
            id = (
                SELECT id
                FROM sidfmd_vv_cves AS cve_table2
                WHERE cve_table.sidfmd_vul_view_id = cve_table2.sidfmd_vul_view_id AND cve_table.cve_id = cve_table2.cve_id
                ORDER BY 
                CASE
                WHEN c3_update_datetime IS NOT NULL THEN
                  100
                WHEN c2_update_datetime IS NOT NULL THEN
                  200
                ELSE
                  300
                END,
                c3_update_datetime DESC,
                c2_update_datetime DESC,
                id DESC
                LIMIT 1
            )
          )
        )
      |)
    end

    # SofTek影響範囲
    if @vuls[:filters][:area] != '' then
      all_content = all_content.where(:softek_area_id => @vuls[:filters][:area])
    end

    # 攻撃コード
    if @vuls[:filters][:exploitable] != '' then
      #CVSSが未定義の場合は、表示が---になるので該当させない。(元々のTSV出力での表示方法と同じ)
      if @filter_category_host_ids or !@role_admin then
        all_content = all_content.where(
          :id => SidfmdVvScsid.where(:status_id => SidfmdIssue::STATUS_UNFINISHED)
                              .where(:sidfmd_host_id => (@filter_category_host_ids || @user_host_ids))
                              .where.not(:cvss => nil)
                              .select(:sidfmd_vul_view_id)
        ).where(:exploitable => true)
      else
        all_content = all_content.where(:exploitable => true)
                                 .where.not(:cvss_label => 'notreg')
      end
    end

    #カテゴリ(複数)
    if @vuls[:filters][:categories][:ids].any? then
      if @filter_category_host_ids or !@role_admin then
        #管理者権限でないかページ上部のカテゴリのフィルタがある場合は、カテゴリを対象のホストのカテゴリに絞り込む
        if @vuls[:filters][:categories][:is_except] then
          all_content = all_content.where.not(
            :id => SidfmdVvHost.joins(
                     'LEFT OUTER JOIN sidfmd_categories_hosts ON sidfmd_categories_hosts.sidfmd_host_id = sidfmd_vv_hosts.sidfmd_host_id'
                   )
                   .where(:sidfmd_vv_hosts => { :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)})
                   .where('sidfmd_categories_hosts.sidfmd_category_id IN (?)', @vuls[:filters][:categories][:ids])
                   .distinct
                   .select(:sidfmd_vul_view_id)
          )
        else
          all_content = all_content.where(
            :id => if @vuls[:filters][:categories][:is_and] then
                     SidfmdVvHost.joins(
                       'LEFT OUTER JOIN sidfmd_categories_hosts ON sidfmd_categories_hosts.sidfmd_host_id = sidfmd_vv_hosts.sidfmd_host_id'
                     )
                     .where(:sidfmd_vv_hosts => { :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)})
                     .group(:sidfmd_vul_view_id)
                     .having('ARRAY_AGG(DISTINCT sidfmd_category_id) @> ARRAY[?]', @vuls[:filters][:categories][:ids])
                     .distinct
                     .select(:sidfmd_vul_view_id)
                   else
                     SidfmdVvHost.joins(
                       'LEFT OUTER JOIN sidfmd_categories_hosts ON sidfmd_categories_hosts.sidfmd_host_id = sidfmd_vv_hosts.sidfmd_host_id'
                     )
                     .where(:sidfmd_vv_hosts => { :sidfmd_host_id => (@filter_category_host_ids || @user_host_ids)})
                     .where('sidfmd_categories_hosts.sidfmd_category_id IN (?)', @vuls[:filters][:categories][:ids])
                     .distinct
                     .select(:sidfmd_vul_view_id)
                   end
          )
        end
      else
        if @vuls[:filters][:categories][:is_except] then
          all_content = all_content.where.not(
            :id => SidfmdVvCategory.where(:sidfmd_category_id => @vuls[:filters][:categories][:ids])
                                   .distinct
                                   .select(:sidfmd_vul_view_id)
          )
        else
          all_content = all_content.where(
            :id => if @vuls[:filters][:categories][:is_and] then
                     SidfmdVvCategory.group(:sidfmd_vul_view_id)
                                     .having('ARRAY_AGG(DISTINCT sidfmd_category_id) @> ARRAY[?]', @vuls[:filters][:categories][:ids])
                                     .distinct
                                     .select(:sidfmd_vul_view_id)
                   else
                     SidfmdVvCategory.where(:sidfmd_category_id => @vuls[:filters][:categories][:ids])
                                     .distinct
                                     .select(:sidfmd_vul_view_id)
                   end
          )
        end
      end
    end

    # 対策予定日設定済
    if @vuls[:filters][:due_date] != '::' then
      if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @vuls[:filters][:due_date] then
        notset = ('notset' == Regexp.last_match(1))
        begindate = Regexp.last_match(2)
        enddate = Regexp.last_match(3)
        if notset then
          if @filter_category_host_ids then
            all_content = all_content.where(
              :id => SidfmdVvScsid.where(:due_date => nil).where(:sidfmd_host_id => @filter_category_host_ids).select(:sidfmd_vul_view_id)
            )
          else
            if @role_admin then
              all_content = all_content.where(
                :id => SidfmdVvScsid.where(:due_date => nil).select(:sidfmd_vul_view_id)
              )
            else
              all_content = all_content.where(
                :id => SidfmdVvScsid.where(:due_date => nil).where(:sidfmd_host_id => @user_host_ids).select(:sidfmd_vul_view_id)
              )
            end
          end
        else
          begin 
            begindate = Date.parse(begindate)
          rescue Exception => e
            log_exception(e)
            begindate = nil
          end
          
          begin
            enddate = Date.parse(enddate)
          rescue Exception => e
            log_exception(e)
            enddate = nil
          end

          if begindate and enddate then
            if @filter_category_host_ids then
               all_content = all_content.where(
                 :id => SidfmdVvScsid.where(:due_date => begindate..enddate).where(:sidfmd_host_id => @filter_category_host_ids).select(:sidfmd_vul_view_id)
               )
            else
              if @role_admin then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where(:due_date => begindate..enddate).select(:sidfmd_vul_view_id)
                )
              else
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where(:due_date => begindate..enddate).where(:sidfmd_host_id => @user_host_ids).select(:sidfmd_vul_view_id)
                )
              end
            end
          elsif begindate then
            if @filter_category_host_ids then
              all_content = all_content.where(
                :id => SidfmdVvScsid.where('due_date >= ?', begindate).where(:sidfmd_host_id => @filter_category_host_ids).select(:sidfmd_vul_view_id)
              )
            else
              if @role_admin then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where('due_date >= ?', begindate).select(:sidfmd_vul_view_id)
                )
              else
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where('due_date >= ?', begindate).where(:sidfmd_host_id => @user_host_ids).select(:sidfmd_vul_view_id)
                )
              end
            end
          elsif enddate then
            if @filter_category_host_ids then
              all_content = all_content.where(
                :id => SidfmdVvScsid.where('due_date <= ?', enddate).where(:sidfmd_host_id => @filter_category_host_ids).select(:sidfmd_vul_view_id)
              )
            else
              if @role_admin then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where('due_date <= ?', enddate).select(:sidfmd_vul_view_id)  
                )
              else
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where('due_date <= ?', enddate).where(:sidfmd_host_id => @user_host_ids).select(:sidfmd_vul_view_id)
                )
              end
            end
          end
        end
      end
    end

    # キーワード検索
    if @vuls[:filters][:keyword].size > 0 then
      kwd = @vuls[:filters][:keyword].gsub('/', '')
      if kwd.size > 0 then
        if /\A\s*(\d+)\s*\z/ =~ kwd then
          html_id = Regexp.last_match(1)
          all_content = all_content.where(%|
            html_id = ? OR 
            title ILIKE ? OR 
            id IN (SELECT sidfmd_vul_view_id FROM sidfmd_vv_cves WHERE cve_id ILIKE ?)
          |, html_id.to_i, "%#{ActiveRecord::Base.send(:sanitize_sql_like, html_id.to_s.strip)}%", "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd.strip)}%")
        elsif /\A\s*KB(\d+)\s*\z/i =~ kwd then
          kbid = -Regexp.last_match(1).to_i
          all_content = all_content.where(
            'html_id = ? OR title ILIKE ?', kbid.to_i, "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd.strip)}%"
          )
        else
          all_content = all_content.where(%|
            id IN (SELECT sidfmd_vul_view_id FROM sidfmd_vv_cves WHERE cve_id ILIKE ?) OR 
            title ILIKE ?
          |, "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd.strip)}%", "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd.strip)}%")
        end
      else
        #/で検索すると、/でtitleが連結されているため、該当してしまう可能性がある
        #/は無効な文字として該当させない
        all_content = all_content.where(:title => '')
      end
    end

    # 日付指定検索
    if @vuls[:filters][:date] != '::' then
      if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @vuls[:filters][:date] then
        orig_flg_rev = Regexp.last_match(1)
        orig_begin_date = Regexp.last_match(2)
        orig_end_date = Regexp.last_match(3)
        if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
          @vuls[:filters][:date] = '::'
        else
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              if @filter_category_host_ids then
                all_content = all_content.where.not(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) BETWEEN ? AND ?', begin_date, end_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where.not(:min_created_on => begin_date..end_date)
                else
                  all_content = all_content.where.not(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) BETWEEN ? AND ?', begin_date, end_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            elsif begin_date then
              begin_date = begin_date.to_time
              if @filter_category_host_ids then
                all_content = all_content.where.not(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) >= ?', begin_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where.not('min_created_on >= ?', begin_date)
                else
                  all_content = all_content.where.not(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) >= ?', begin_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              if @filter_category_host_ids then
                all_content = all_content.where.not(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) <= ?', end_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where.not('min_created_on <= ?', end_date)
                else
                  all_content = all_content.where.not(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) <= ?', end_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            else
              # last resort
              @vuls[:filters][:date] = '::'
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              if @filter_category_host_ids then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) BETWEEN ? AND ?', begin_date, end_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where(:min_created_on => begin_date..end_date)
                else
                  all_content = all_content.where(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) BETWEEN ? AND ?', begin_date, end_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            elsif begin_date then
              begin_date = begin_date.to_time
              if @filter_category_host_ids then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) >= ?', begin_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where('min_created_on >= ?', begin_date)
                else
                  all_content = all_content.where(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) >= ?', begin_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              if @filter_category_host_ids then
                all_content = all_content.where(
                  :id => SidfmdVvScsid.where(:sidfmd_host_id => @filter_category_host_ids)
                                      .group(:sidfmd_vul_view_id)
                                      .having('MIN(created_on) <= ?', end_date)
                                      .select(:sidfmd_vul_view_id)
                )
              else
                if @role_admin then
                  all_content = all_content.where('min_created_on <= ?', end_date)
                else
                  all_content = all_content.where(
                    :id => SidfmdVvScsid.where(:sidfmd_host_id => @user_host_ids)
                                        .group(:sidfmd_vul_view_id)
                                        .having('MIN(created_on) <= ?', end_date)
                                        .select(:sidfmd_vul_view_id)
                  )
                end
              end
            else
              # last resort
              @vuls[:filters][:date] = '::'
            end
          end
        end
      else
        @vuls[:filters][:date] = '::'
      end
    end

    #カスタムタグ検索
    if @vuls[:filters][:ctags] && @vuls[:filters][:ctags][:ids] && @vuls[:filters][:ctags][:ids].any? then
      unless @role_admin then
        @vuls[:filters][:ctags][:ids] = @vuls[:filters][:ctags][:ids] & SidfmdCtag.user_relatable_ctag_ids(@role_admin, @user_category_ids)
      end
      all_content = all_content.where(
        :id => SidfmdVvCtag.where(:sidfmd_ctag_id => @vuls[:filters][:ctags][:ids]).select(:sidfmd_vul_view_id)
      )
    end

    #ソート処理
    
    #min_created_onでソートする前のデータ取得処理
    #_vv_min_created_onでソートする
    get_min_created_on = ->(content){
      condition = if @filter_category_host_ids and @filter_category_host_ids.any? then
                    "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}]))"
                  elsif !@role_admin and @user_host_ids.any? then
                    "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}]))"
                  else
                    ''
                  end
      content = content.joins(%|
        LEFT OUTER JOIN(
          SELECT
            sidfmd_vul_view_id AS _vv_id,
            MIN(DATE_TRUNC('DAY', created_on)) AS order_vv_min_created_on
          FROM
            sidfmd_vv_scsids
          #{condition}
          GROUP BY sidfmd_vul_view_id 
        )AS order_vv_min_created_ons
        ON sidfmd_vul_views.id = order_vv_min_created_ons._vv_id
      |)
    }

    #要対応
    case @vuls[:order]
    when 'warning_asc', 'warning_desc' then
      get_warning_values = ->(content, host_ids, idx){
        content = content.joins(%|
                    LEFT OUTER JOIN (
                      SELECT 
                        sidfmd_vul_view_id AS vul_view_id,
                        MIN(status_id) AS order_warning_target_vv_min_status_id,
                        #{idx == :sri ? "MAX(sri) AS order_warning_target_vv_max_sri" : "MAX(cvss) order_warning_target_vv_max_cvss"}
                      FROM
                        sidfmd_vv_scsids
                      WHERE
                        #{(host_ids and host_ids.any?) ? "sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}])) AND" : ''}
                        status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                      GROUP BY
                        sidfmd_vul_view_id
                      HAVING
                        #{idx == :sri ? "MAX(sri) >= #{@corresp_required[:sri]}" : "(MAX(cvss) IS NULL OR MAX(cvss) >= #{@corresp_required[:cvss]})"}
                    ) AS order_vv_warning_values
                    ON sidfmd_vul_views.id = order_vv_warning_values.vul_view_id
                  |)
      }
      all_content = get_min_created_on.(all_content)
      if @vuls[:order] == 'warning_asc' then
        if @filter_category_host_ids or !@role_admin then
          all_content = get_warning_values.(all_content, (@filter_category_host_ids || @user_host_ids), @vuls[:idx])
                                          .order(%|
                                            CASE
                                            WHEN order_warning_target_vv_min_status_id IS NULL THEN
                                              100
                                            ELSE
                                              0
                                            END
                                          |)
                                          .order('order_vv_min_created_on')
                                          .order('LOWER(title COLLATE "C")')
                                          .order("abs(html_id)") 
        else
          all_content = all_content.order("warning_#{@vuls[:idx].to_s} DESC")
                                   .order('order_vv_min_created_on')
                                   .order('LOWER(title COLLATE "C")')
                                   .order("abs(html_id)")
        end
      else
        if @filter_category_host_ids or !@role_admin then
          all_content = get_warning_values.(all_content, (@filter_category_host_ids || @user_host_ids), @vuls[:idx])
                                          .order(%|
                                            CASE 
                                            WHEN order_warning_target_vv_min_status_id IS NULL THEN
                                              0
                                            ELSE
                                              100
                                            END
                                          |)
                                          .order('order_vv_min_created_on DESC')
                                          .order('LOWER(title COLLATE "C") DESC')
                                          .order("abs(html_id) DESC")
        else
          all_content = all_content.order("warning_#{@vuls[:idx].to_s}")
                                   .order('order_vv_min_created_on DESC')
                                   .order('LOWER(title COLLATE "C") DESC')
                                   .order("abs(html_id) DESC")
        end
      end
    when 'statuses_asc', 'statuses_desc' then
      order_by_status = ->(content, host_ids){
        content = if host_ids.any? then
                    content.joins(%|
                      LEFT OUTER JOIN(
                        SELECT
                          sidfmd_vul_view_id AS vul_view_id,
                          MIN(status_id) AS order_vv_min_status_id
                        FROM
                          sidfmd_vv_scsids
                        WHERE
                          sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))
                        GROUP BY sidfmd_vul_view_id
                      ) AS order_vv_min_status
                      ON sidfmd_vul_views.id = order_vv_min_status.vul_view_id
                    |).order(%| order_vv_min_status_id #{@vuls[:order] == 'statuses_asc' ? 'ASC' : 'DESC'} |)
                    .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'statuses_asc' ? 'ASC' : 'DESC'} |)
                    .order("abs(html_id) #{@vuls[:order] == 'statuses_asc' ? 'ASC' : 'DESC'}")
                  else
                    content
                  end
      }
      all_content = if @filter_category_host_ids or !@role_admin then
                      order_by_status.(all_content, (@filter_category_host_ids || @user_host_ids))
                    else
                      if @vuls[:order] == 'statuses_asc' then
                        all_content.order(:min_status_id).order('LOWER(title COLLATE "C")').order("abs(html_id)")
                      else
                        all_content.order(:min_status_id => :desc).order('LOWER(title COLLATE "C") DESC').order("abs(html_id) DESC")
                      end
                    end      
    when 'maxsri_asc', 'maxsri_desc' then
      order_by_sri = ->(content, filter_hosts){
        content = content.joins(%|
                    LEFT OUTER JOIN(
                      SELECT
                        sidfmd_vul_view_id AS vul_view_id,
                        MAX(sri) AS order_max_sri_value
                      FROM
                        sidfmd_vv_scsids
                      WHERE
                        #{filter_hosts}
                        status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                      GROUP BY sidfmd_vul_view_id
                    ) AS order_vv_max_sri
                    ON sidfmd_vul_views.id = order_vv_max_sri.vul_view_id
                  |)
                  .order(@vuls[:order] == 'maxsri_asc' ? 'order_max_sri_value DESC NULLS LAST' : 'order_max_sri_value ASC NULLS FIRST')
                  .order(%| order_vv_min_created_on #{@vuls[:order] == 'maxsri_asc' ? 'DESC' : 'ASC'}|)
                  .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'maxsri_asc' ? 'ASC' : 'DESC'} |)
                  .order("abs(html_id) #{@vuls[:order] == 'maxsri_asc' ? 'DESC' : 'ASC'}")
      }
      if @filter_category_host_ids then
        if @filter_category_host_ids.any? then
          all_content = order_by_sri.(get_min_created_on.(all_content), "sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}])) AND")
        end
      elsif @role_admin then
        all_content = order_by_sri.(get_min_created_on.(all_content), '')
      elsif @user_host_ids.any? then
        all_content = order_by_sri.(get_min_created_on.(all_content), "sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}])) AND")
      end
    when 'cvss_asc', 'cvss_desc' then
      order_by_cvss = ->(content, host_ids){
        content = content.joins(%|
                    LEFT OUTER JOIN(
                      SELECT
                        sidfmd_vul_view_id AS vul_view_id,
                        MAX(cvss) AS order_max_cvss_score
                      FROM
                        sidfmd_vv_scsids
                      #{host_ids.blank? ? '' : "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))"}
                      GROUP BY sidfmd_vul_view_id
                    ) AS order_vv_max_cvss
                    ON sidfmd_vul_views.id = order_vv_max_cvss.vul_view_id
                  |).order(%| 
                      CASE
                      WHEN html_id < 0 THEN
                        100
                      ELSE
                        10
                      END
                      #{@vuls[:order] == 'cvss_asc' ? 'ASC' : 'DESC'}
                  |)
                  .order(%| order_max_cvss_score #{@vuls[:order] == 'cvss_asc' ? 'DESC NULLS LAST' : 'ASC NULLS FIRST'} |)
                  .order(%| abs(html_id) #{@vuls[:order] == 'cvss_asc' ? 'ASC' : 'DESC'} |)
                  .order(%| order_vv_min_created_on #{@vuls[:order] == 'cvss_asc' ? 'DESC' : 'ASC'} |)
                  .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'cvss_asc' ? 'ASC' : 'DESC'} |)
                  .order("abs(html_id) #{@vuls[:order] == 'cvss_asc' ? 'ASC' : 'DESC'}")
      }

      if @filter_category_host_ids then
        if @filter_category_host_ids.any? then
          all_content = order_by_cvss.(get_min_created_on.(all_content), @filter_category_host_ids)
        end
      elsif @role_admin then
        all_content = order_by_cvss.(get_min_created_on.(all_content), '')
      elsif @user_host_ids.any? then
        all_content = order_by_cvss.(get_min_created_on.(all_content), @user_host_ids)
      end
    when 'area_asc', 'area_desc' then
      all_content = get_min_created_on.(all_content).order(%|
                      CASE WHEN html_id > 0 THEN softek_area_id ELSE -100 END #{@vuls[:order] == 'area_asc' ? 'DESC' : 'ASC'}
                    |).order(%| order_vv_min_created_on #{@vuls[:order] == 'area_asc' ? 'DESC' : 'ASC'} |)
                    .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'area_asc' ? 'ASC' : 'DESC'} |)
                    .order(%| abs(html_id) #{@vuls[:order] == 'area_asc' ? 'ASC' : 'DESC'} |)
    when 'exploitable_asc', 'exploitable_desc' then
      all_content = get_min_created_on.(all_content)
                                      .joins(%|
        LEFT OUTER JOIN(
          SELECT
            sidfmd_vul_view_id AS order_exploit_vv_id,
            MAX(cvss) AS order_exploit_max_cvss
          FROM
            sidfmd_vv_scsids
          #{(@filter_category_host_ids and @filter_category_host_ids.any?) ? "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}])) AND status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))" : (@role_admin or @user_host_ids.blank?) ?  '' : "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}])) AND status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))"}
          GROUP BY
            sidfmd_vul_view_id
        )AS order_exploit_max_cvsses 
        ON sidfmd_vul_views.id = order_exploit_max_cvsses.order_exploit_vv_id
      |).order(%|
        CASE WHEN order_exploit_max_cvss IS NULL THEN
          CASE WHEN html_id < 0 THEN
            1000
          ELSE
            1500
          END
        ELSE
          CASE 
          WHEN exploitable IS TRUE THEN
            100
          WHEN exploitable IS NULL THEN
            1400
          ELSE
            500
          END
        END
        #{@vuls[:order] == 'exploitable_asc' ? 'ASC' : 'DESC'}
      |).order(%| order_vv_min_created_on #{@vuls[:order] == 'exploitable_asc' ? 'DESC' : 'ASC'} |)
        .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'exploitable_asc' ? 'DESC' : 'ASC'} |)
        .order(%| abs(html_id) #{@vuls[:order] == 'exploitable_asc' ? 'ASC' : 'DESC'} |)
    when 'elapseddays_asc', 'elapseddays_desc' then
      order_condition = if @filter_category_host_ids and @filter_category_host_ids.any? then
                          " AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}]))"
                        elsif !@role_admin and @user_host_ids.any? then
                          " AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}]))"
                        else
                          ''
                        end
      all_content = get_min_created_on.(all_content).joins(%|
                      LEFT OUTER JOIN(
                        SELECT
                          sidfmd_vul_view_id AS _vv_id,
                          MIN(created_on) AS vv_min_created_on_for_elapsed_days
                        FROM
                          sidfmd_vv_scsids
                        WHERE status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))#{order_condition}
                        GROUP BY sidfmd_vul_view_id 
                      )AS vv_min_created_ons_for_elapsed_days
                      ON sidfmd_vul_views.id = vv_min_created_ons_for_elapsed_days._vv_id
                    |)
                    .order(%| vv_min_created_on_for_elapsed_days #{@vuls[:order] == 'elapseddays_asc' ? 'ASC NULLS LAST' : 'DESC NULLS FIRST'} |)
                    .order(%| order_vv_min_created_on #{@vuls[:order] == 'elapseddays_asc' ? 'DESC' : 'ASC'} |)
                    .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'elapseddays_asc' ? 'DESC' : 'ASC'} |)
                    .order(%| abs(html_id) #{@vuls[:order] == 'elapseddays_asc' ? 'ASC' : 'DESC'} |)
    when 'title_asc', 'title_desc' then
      all_content = all_content.order(%| abs(html_id) #{@vuls[:order] == 'title_asc' ? 'ASC' : 'DESC'} |)
                               .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'title_asc' ? 'ASC' : 'DESC'} |)
    when 'hostnames_asc', 'hostnames_desc' then
      hosts_filter = if @filter_category_host_ids and @filter_category_host_ids.any? then
                       "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@filter_category_host_ids.join(', ')}]))"
                     elsif @role_admin then
                       ''
                     else
                       if @user_host_ids.any? then
                         "WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{@user_host_ids.join(', ')}]))"
                       else
                         ''
                       end
                     end
      all_content = all_content.joins(%|
        LEFT OUTER JOIN(
          SELECT 
            sidfmd_vul_view_id AS vul_view_id, 
            MIN(LOWER(display_name COLLATE "C")) AS min_name 
          FROM 
            sidfmd_vv_hosts
          #{hosts_filter}
          GROUP BY sidfmd_vul_view_id
        ) AS vv_hosts_min_name 
        ON sidfmd_vul_views.id = vv_hosts_min_name.vul_view_id
      |).order(%| min_name COLLATE "C" #{@vuls[:order] == 'hostnames_asc' ? 'ASC' : 'DESC'} |)
        .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'hostnames_asc' ? 'ASC' : 'DESC'} |)
        .order(%| abs(html_id) #{@vuls[:order] == 'hostnames_asc' ? 'ASC' : 'DESC'} |)
    when 'categories_asc', 'categories_desc' then
      #管理者権限でないか、ページ上部のカテゴリのフィルタがある場合は、表示されるホストのカテゴリで絞る
      filter_by_host_cat_ids = if (@filter_category_host_ids and @filter_category_host_ids.any?) or (!@role_admin and @user_host_ids.any?) then
                                 %| WHERE sidfmd_vv_hosts.sidfmd_host_id IN (SELECT unnest(ARRAY[#{(@filter_category_host_ids || @user_host_ids).join(', ')}])) |
                               else
                                 ''
                               end
      if @summary_prior_category == 'custom' then
        all_content = get_min_created_on.(all_content).joins(%|
          LEFT OUTER JOIN(
            SELECT 
              sidfmd_vv_categories.sidfmd_vul_view_id AS vul_view_id,
              MIN(LOWER(sidfmd_vv_categories.name COLLATE "C")) AS custom_cat_name
            FROM
              sidfmd_vv_categories
              INNER JOIN sidfmd_vv_hosts ON sidfmd_vv_categories.sidfmd_vul_view_id = sidfmd_vv_hosts.sidfmd_vul_view_id
            #{filter_by_host_cat_ids}
            #{filter_by_host_cat_ids.size > 0 ? 'AND' : 'WHERE'} sidfmd_vv_categories.is_system IS FALSE
            #{(@role_admin or @user_category_ids.size == 0) ? '' : "AND sidfmd_vv_categories.sidfmd_category_id IN (SELECT unnest(ARRAY[#{@user_category_ids.join(', ')}]))"}
            GROUP BY
              sidfmd_vv_categories.sidfmd_vul_view_id
          ) AS order_custom_category
          ON sidfmd_vul_views.id = order_custom_category.vul_view_id
        |).joins(%|
          LEFT OUTER JOIN(
            SELECT 
              sidfmd_vv_categories.sidfmd_vul_view_id AS vul_view_id,
              MIN(sidfmd_vv_categories.sidfmd_category_id) AS system_cat_id
            FROM
              sidfmd_vv_categories
              INNER JOIN sidfmd_vv_hosts ON sidfmd_vv_categories.sidfmd_vul_view_id = sidfmd_vv_hosts.sidfmd_vul_view_id
            #{filter_by_host_cat_ids}
            #{filter_by_host_cat_ids.size > 0 ? 'AND' : 'WHERE'} sidfmd_vv_categories.is_system IS TRUE
            GROUP BY
              sidfmd_vv_categories.sidfmd_vul_view_id
          ) AS order_system_category
          ON sidfmd_vul_views.id = order_system_category.vul_view_id
        |).order(%|
          CASE WHEN custom_cat_name IS NULL THEN
            NULL
          ELSE
            LOWER(custom_cat_name COLLATE "C")
          END
          #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'}
        |).order(%|
          CASE WHEN custom_cat_name IS NULL THEN
            system_cat_id 
          ELSE
            0
          END
          #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'}
        |).order(%| order_vv_min_created_on #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
          .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
          .order(%| abs(html_id) #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'} |)
      else
        all_content = get_min_created_on.(all_content).joins(%|
          LEFT OUTER JOIN(
            SELECT 
              sidfmd_vul_view_id AS vul_view_id, 
              MIN(sidfmd_category_id) AS cat_id 
            FROM 
              sidfmd_vv_hosts INNER JOIN sidfmd_categories_hosts ON sidfmd_vv_hosts.sidfmd_host_id = sidfmd_categories_hosts.sidfmd_host_id
            #{filter_by_host_cat_ids}  
            GROUP BY sidfmd_vul_view_id
          ) AS vul_min_system_category 
          ON sidfmd_vul_views.id = vul_min_system_category.vul_view_id
        |).order(%| cat_id #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'} |)
          .order(%| order_vv_min_created_on #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
          .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
          .order(%| abs(html_id) #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'} |) 
      end
      #all_content = get_min_created_on.(all_content).joins(%|
      #  LEFT OUTER JOIN(
      #    SELECT 
      #      sidfmd_vul_view_id AS vul_view_id, 
      #      MIN(sidfmd_category_id) AS cat_id 
      #    FROM 
      #      sidfmd_vv_categories 
      #    WHERE is_system IS TRUE#{filter_by_host_cat_ids}
      #    GROUP BY sidfmd_vul_view_id
      #  ) AS vul_min_system_category 
      #  ON sidfmd_vul_views.id = vul_min_system_category.vul_view_id
      #|).order(%| cat_id #{@vuls[:order] == 'categories_asc' ? 'ASC' : 'DESC'} |)
      #  .order(%| order_vv_min_created_on #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
      #  .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'categories_asc' ? 'DESC' : 'ASC'} |)
    when 'duedates_asc', 'duedates_desc' then
      order_by_duedate = ->(content, host_ids){
        # STATUS_FINISHEDのデータは計算に入れない
        content = content.joins(%|
                    LEFT OUTER JOIN(
                      SELECT
                        sidfmd_vul_view_id AS vul_view_id,
                        COUNT(sidfmd_host_id) AS count_unset_due_date_hosts
                      FROM 
                        sidfmd_vv_scsids
                      WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}])) 
                        AND status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                        AND due_date IS NOT NULL
                      GROUP BY sidfmd_vul_view_id
                    ) AS count_unset_duedates
                    ON sidfmd_vul_views.id = count_unset_duedates.vul_view_id
                  |).joins(%|
                    LEFT OUTER JOIN(
                      SELECT
                        sidfmd_vul_view_id AS vul_view_id,
                        COUNT(sidfmd_host_id) AS count_due_date_hosts
                      FROM
                        sidfmd_vv_scsids
                      WHERE sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))
                        AND status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_UNFINISHED.join(', ')}]))
                      GROUP BY sidfmd_vul_view_id
                    ) AS count_duedates
                    ON sidfmd_vul_views.id = count_duedates.vul_view_id
                  |).order(%|
                    CASE 
                    WHEN count_unset_due_date_hosts > 0 AND count_due_date_hosts > 0 THEN
                      ROUND(CAST(CAST(count_unset_due_date_hosts AS FLOAT) / CAST(count_due_date_hosts AS FLOAT) * 100.0 as numeric) , 1)
                    ELSE
                      0.0
                    END
                    #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'}
                  |)
                  .order(%| order_vv_min_created_on #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'} |)
                  .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'} |)
                  .order(%| abs(html_id) #{@vuls[:order] == 'duedates_asc' ? 'ASC' : 'DESC'} |)
      }
      
      all_content = if (@filter_category_host_ids and @filter_category_host_ids.any?) or (!@role_admin and @user_host_ids.any?) then
                      order_by_duedate.(get_min_created_on.(all_content), (@filter_category_host_ids || @user_host_ids))
                    elsif @role_admin then
                      get_min_created_on.(all_content).order(%|
                        ROUND(CAST(duedates_percent as numeric), 1) #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'}
                      |)
                      .order(%| order_vv_min_created_on #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'} |)
                      .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'duedates_asc' ? 'DESC' : 'ASC'} |)
                      .order(%| abs(html_id) #{@vuls[:order] == 'duedates_asc' ? 'ASC' : 'DESC'} |)
                    else
                      all_content
                    end
    when 'finished_asc', 'finished_desc' then
      order_by_fixed_percent = ->(content, host_ids){
        content = content.joins(%|
          LEFT OUTER JOIN(
            SELECT
              sidfmd_vul_view_id AS vul_view_id,
              COUNT(sidfmd_host_id) AS count_fin_host_id
            FROM 
              sidfmd_vv_scsids
            WHERE 
              status_id IN (SELECT unnest(ARRAY[#{SidfmdIssue::STATUS_FINISHED.join(', ')}]))
              AND sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))
            GROUP BY sidfmd_vul_view_id
          ) AS count_fin_host_ids
          ON sidfmd_vul_views.id = count_fin_host_ids.vul_view_id
        |).joins(%|
          LEFT OUTER JOIN(
            SELECT
              sidfmd_vul_view_id AS vul_view_id,
              COUNT(sidfmd_host_id) AS count_host_id
            FROM 
              sidfmd_vv_scsids
            WHERE
              sidfmd_host_id IN (SELECT unnest(ARRAY[#{host_ids.join(', ')}]))
            GROUP BY sidfmd_vul_view_id
          ) AS count_host_ids
          ON sidfmd_vul_views.id = count_host_ids.vul_view_id
        |).order(%|
          ROUND(CAST(CAST(COALESCE(count_fin_host_id, 0) AS FLOAT) / CAST(COALESCE(count_host_id, 0) AS FLOAT) * 100.0 as numeric), 1) 
          #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'}
        |)
        .order(%| order_vv_min_created_on #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'} |)
        .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'} |)
        .order(%| abs(html_id) #{@vuls[:order] == 'finished_asc' ? 'ASC' : 'DESC'} |)
      }

      all_content = if (@filter_category_host_ids and @filter_category_host_ids.any?) or (!@role_admin and @user_host_ids.any?) then
                      order_by_fixed_percent.(get_min_created_on.(all_content), (@filter_category_host_ids || @user_host_ids))
                    elsif @role_admin then
                      get_min_created_on.(all_content)
                                        .order("ROUND(CAST(fixed_percent as numeric), 1) #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'}")
                                        .order(%| order_vv_min_created_on #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'} |)
                                        .order(%| LOWER(title COLLATE "C") #{@vuls[:order] == 'finished_asc' ? 'DESC' : 'ASC'} |)
                                        .order(%| abs(html_id) #{@vuls[:order] == 'finished_asc' ? 'ASC' : 'DESC'} |)
                    else
                      all_content
                    end
    end

    all_content
  end

  def get_dist_cat_vuls(category_id = 0)
    ret = {}
    scats = [ :invalid, :public_server, :private_server, :mobile, :client, :shared_server, :virtual_host_server, :public_network, :private_network ]
    scats.each do |s|
      ret[s] = 0
    end
    hosts =
      SidfmdHost
      .enable_hosts
    if category_id > 0 then
      hosts =
        hosts
        .where(:id =>
               SidfmdCategoriesHost
                 .where(:sidfmd_category_id => category_id)
                 .select(:sidfmd_host_id)
                 .distinct)
    else
      hosts = hosts.categorized
      if !@role_admin then
        hosts = hosts.where(:id => @user_host_ids)
      end
    end
    hosts.each do |host|
      ret[scats[host.system_category.id]] += 1
    end
    ret
  end

  def get_dist_os_vuls(category_id = 0)
    ret = []
    ret_hash = {}
    hosts =
      SidfmdHost
      .enable_hosts
    if category_id > 0 then
      hosts =
        hosts
        .where(:id =>
               SidfmdCategoriesHost
                 .where(:sidfmd_category_id => category_id)
                 .select(:sidfmd_host_id)
                 .distinct)
    else
      hosts = hosts.categorized
      if !@role_admin then
        hosts = hosts.where(:id => @user_host_ids)
      end
    end
    hosts.each do |host|
      dist = host.get_os
      os = '（不明）'
      if dist and dist[:osname] and dist[:osname].size > 0 then
        os = dist[:osname]
      end
      if ret_hash.has_key?(os) then
        ret_hash[os] += 1
      else
        ret_hash[os] = 1
      end
    end
    ret_hash.keys.each do |k|
      ret << { :name => k, :value => ret_hash[k] }
    end
    ret_size = ret.sum {|x| x[:value] }
    temp_ev = proc {|x| [ (x[:name] == '（不明）' ? 1000 : 100), x[:value], x[:name] ]}
    ret.sort! {|x, y|
      xx = temp_ev.call(x)
      yy = temp_ev.call(y)
      r = xx[0] <=> yy[0]
      if r == 0 then
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[2] <=> yy[2]
        end
      end
      r
    }
    [ ret, ret_size ]
  end

  def create_vul_view_dist_hosts(category_id = 0)
    @vul_view_dist_hosts = 
      if category_id > 0 then
        SidfmdHostView.where(
          :id => SidfmdHvCategory.where(:sidfmd_category_id => category_id)
                                 .select(:sidfmd_host_view_id)
        )
      else
        if @role_admin then
          SidfmdHostView.where(
           :id => SidfmdHvCategory.select(:sidfmd_host_view_id)
          )
        else
          SidfmdHostView.where(
            :id => SidfmdHvCategory.where(:sidfmd_category_id => @user_category_ids)
                                   .select(:sidfmd_host_view_id)
          )
        end
      end
  end

  def get_vul_view_dist_cat_vuls
    ret = {}
    scats = [ :invalid, :public_server, :private_server, :mobile, :client, :shared_server, :virtual_host_server, :public_network, :private_network ]
    scats.each_with_index do |scat, idx|
      ret[scat] = @vul_view_dist_hosts.where(:system_category_id => idx).count
    end
    ret
  end

  def get_vul_view_dist_os_vuls
    hosts_size = @vul_view_dist_hosts.count
    oss = @vul_view_dist_hosts.order('LOWER(osname COLLATE "C")')
                              .select(:osname, 'COUNT(id) AS count_host')
                              .group(:osname)
                              .map{|os| { :name => (os.osname ? os.osname : '（不明）'), :value => os.count_host } }
    [ oss, hosts_size ]   
  end

  #脆弱性のcve(s)として採用するcve(s)を取得する
  ##CVEIDが同じものがある場合は、c3_update_datetimeがあればc3_update_datetime
  ##c2_update_datetimeがあればc2_update_datetimeで最も新しいもの1つを採用する。
  ##両方無い場合は配列の順番で後のものを採用する。
  #※Viewを使っていないところで使用する。
  #sidfmd_host_controller#get_filtered_cvesと同じもの
  def get_filtered_cves(vul_cves, is_for_summary = true)
    filtered_cves = []
    if vul_cves.any? then
      target_cve = nil
      create_cve_data = ->(cve){
        cvss_v2 = if !cve[:c2_vector].empty? then
                    cve[:c2_vector].split('/').map{|c| c.split(':')}.to_h
                  else
                    nil
                  end
        cvss_v3 = if !cve[:c3_vector].empty? then
                    cve[:c3_vector].split('/').map{|c| c.split(':')}.to_h
                  else
                    nil
                  end
        return {
          :cve_id => cve[:cve_id],
          :cvss_v2 => cvss_v2,
          :cvss_v3 => cvss_v3,
        }
      }
      none_update_time_vuls = {}
      vul_cves.each do |cve|
        if cve[:c3_update_datetime].empty? && cve[:c2_update_datetime].empty? then
          none_update_time_vuls[cve[:cve_id]] = [] unless none_update_time_vuls[cve[:cve_id]]
          none_update_time_vuls[cve[:cve_id]] << cve
        end
      end

      vul_cves.sort_by{|cve| cve[:cve_id]}.each_with_index do |cve, idx|
        if idx == 0 then
          target_cve = cve
        else
          if target_cve[:cve_id] == cve[:cve_id] then
            if !cve[:c3_update_datetime].empty?  then
              if target_cve[:c3_update_datetime].empty? then
                target_cve = cve
              else
                target_cve = cve if target_cve[:c3_update_datetime] < cve[:c3_update_datetime]
              end
            elsif target_cve[:c3_update_datetime].empty? && !cve[:c2_update_datetime].empty? then
              if target_cve[:c2_update_datetime].empty? then
                target_cve = cve
              else
                target_cve = cve if target_cve[:c2_update_datetime] < cve[:c2_update_datetime]
              end
            elsif target_cve[:c3_update_datetime].empty? && target_cve[:c2_update_datetime].empty?
              target_cve = none_update_time_vuls[[:cve_id]].last
            end
          else
            if is_for_summary then
              filtered_cves << create_cve_data.(target_cve)
            else
              filtered_cves << target_cve
            end
            target_cve = cve
          end
        end
        if idx == vul_cves.size - 1 then
          if is_for_summary then
            filtered_cves << create_cve_data.(target_cve)
          else
            filtered_cves << target_cve
          end
        end
      end
    end
    return filtered_cves
  end

  def create_summary_vul_html(vul, **opts)
    hosts = (
      vul.hosts ? 
      vul.hosts.map{|host|
        {
          :sidfmd_host_id => host[0].to_i,
          :hostname => host[1],
          :ext_name => host[2],
          :display_name => host[3]
        }
      } :
      []
    )
    system_categories = (
      vul.system_categories ?
      vul.system_categories.map{|s_cat|
        { :sidfmd_category_id => s_cat[0].to_i, :name => s_cat[1] }
      } :
      []
    ).uniq

    custom_categories = (
      vul.custom_categories ?
      vul.custom_categories.map{|c_cat|
        { :sidfmd_category_id => c_cat[0].to_i, :name => c_cat[1] }
      } :
      []
    ).uniq

    ctags = vul.ctags.map{|ctag|
      { :keyword => ctag[0], :text_color => ctag[1], :bg_color => ctag[2] }
    }
    if opts[:category_id] .to_i > 0 or !@role_admin then
      SidfmdVulView.create_summary_calc_value(vul)
    end
    filter_inherit = {
      :'pages[idx]' => opts[:idx],
      :'pages[category_id]' => opts[:category_id],
      :'pages[filters]' => opts[:filters].reject {|k, v|
        [ :keyword ].include?(k)
      }.to_json
    }
    data = {
      :chkbox => view_context.create_summary_vul_chkbox(vul.html_id, opts[:checked]),
      :warning => view_context.create_summary_taiou(opts[:idx] == :sri ? vul.warning_sri : vul.warning_cvss),
      :statuses => view_context.create_summary_status(vul.status_ids, vul.min_status_id.to_i),
      :maxsri => view_context.create_summary_vul_sri(vul.max_sri.to_i, vul.html_id),
      :cvss => view_context.create_summary_vul_cvss(vul.id, vul.html_id, vul.cvss_score, vul.cvss_label, vul.is_cves),
      :area => view_context.create_summary_vul_area(vul.softek_area_id, vul.html_id),
      :exploitable => view_context.create_summary_vul_exploit(vul.exploitable, vul.cvss_label, vul.html_id),
      :elapseddays => view_context.create_summary_elapsed_days(
        (vul.min_created_on ? vul.min_created_on.to_datetime.localtime.strftime('%Y/%m/%d') : '---'),
        (vul.min_elapsed_created_on ? ((Time.current - vul.min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i : -1)
      ),
      :title => view_context.create_summary_vul_title(
        SidfmdVulView.display_title(vul.title),
        vul.short_description,
        vul.html_id,
        filter_inherit
      ),
      :detail => view_context.create_summary_link_to_vul_detail(vul.html_id),
      :hostnames => view_context.create_summary_hosts(hosts, vul.html_id, opts[:category_id]),
      :categories => view_context.create_summary_vul_categories(
        system_categories,
        custom_categories,
        vul.html_id,
        opts[:category_id],
        opts[:summary_prior_category]
      ),
      :duedates => view_context.create_summary_duedates_percent(vul.duedates_percent),
      :finished => view_context.create_summary_vul_fixed_percent(vul.fixed_percent),
      :ctags => view_context.create_summary_custom_tag(ctags, vul.html_id, vul.title, @sidfm_roles),
    }
  end

  public

  def index
    parse_params
  end

  def index_r
    get_summary_prior_category
    parse_params
    create_vul_view_content
    filter_vul_view_content
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    @filter_inherit = {
      :'pages[idx]' => @vuls[:idx],
      :'pages[category_id]' => @vuls[:category_id],
      :'pages[filters]' => @vuls[:filters].reject {|k, v|
        [ :keyword ].include?(k)
      }.to_json,
    }
  
    @display_not_system_categories = SidfmdCategory.where(:id => @not_system_categories.map{|c| c.id}).order('name COLLATE "C"')
 
    @cvss_v2_vectors = SidfmdIssueCve.get_cvss_v2_vectors
    @cvss_v3_vectors = SidfmdIssueCve.get_cvss_v3_vectors
    #ブラウザバックかどうか判断するための処理
    #if !session['chkbok_ids'] then
    #  session['chkbok_ids'] = {}
    #end
    #if !session['chkbok_ids']['vuls'] then
    #  session['chkbok_ids']['vuls'] = {}
    #end
    #session['chkbok_ids']['vuls']['is_browser_back'] = false
  end

  def threshold_setting
    @sri_risks = [ 4, 3, 2 , 1 ]
    @corresp_required = Sidfm.corresp_required
  end

  def threshold_update
    #filtered_params = params.fetch(:params, {}).permit(:sri, :cvss)
    filtered_params = params.permit(:sri, :cvss)
    if filtered_params[:sri] then
      Sidfm.corresp_required_sri = filtered_params[:sri].to_i
    end
    if filtered_params[:cvss] then
      Sidfm.corresp_required_cvss = filtered_params[:cvss].to_f
    end
    host_ids = SidfmdHost.pluck(:id)
    html_ids = SidfmdIssue.distinct.pluck(:html_id)
    Rails.cache.clear
 
    render :layout => false, :json => { :host_ids => host_ids, :html_ids => html_ids }
    #redirect_to :action => :threshold_setting, :notice => '設定しました。'
  end

  def update_status
=begin    f_params = params.fetch(:status, {}).permit(:htmlids, :prev_status, :new_status, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    htmlids = nil
    begin
      htmlids = JSON.parse(f_params[:htmlids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !htmlids.respond_to?(:'[]') or !htmlids.respond_to?(:size) or htmlids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then 
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    if new_status == 0 then
      dump[:notice] = '設定後ステータスが設定されなかったため、操作は行われておりません。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        old_status_ids = []
        issues = SidfmdIssue
          .joins(:issue)
          .where(:html_id => htmlids)
        if prev_status == 0 then
          issues = issues
            .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          old_status_ids = SidfmdIssue::STATUS_UNFINISHED
        else
          issues = issues.where(:issues => { :status_id => prev_status })
          old_status_ids << prev_status
        end
        #issues.each do |issue|
        #  issue.with_lock do
        #    issue.change_status_with_journal(User.current.id, new_status, comment)
        #  end
        #end
        SidfmdIssue.change_statuses_with_journal(issues.map{|i|i.id}, old_status_ids, new_status, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids,:shift_num, :ids,:prev_status, :new_status, :comment, :category_id)
    err = false
    err_info = {}
    htmlids = nil
    shift_num = f_params[:shift_num]
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      htmlids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      err = true
    end
    if !htmlids.respond_to?(:'[]') or !htmlids.respond_to?(:size) or htmlids.size == 0 then
      err = true
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      err = true
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      err = true
    end
    if new_status == 0 then
      err = true
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    update_data = []
    begin
      SidfmdIssue.transaction do
        old_status_ids = []
        if prev_status == 0 then
          old_status_ids = SidfmdIssue::STATUS_UNFINISHED
        else
          old_status_ids << prev_status
        end
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .joins(:issue)
            .where(:html_id => htmlids)
          if !@role_admin then
            issues = issues.where(:sidfmd_host_id => @user_host_ids)
          end
          if f_params[:category_id].to_i > 0 then
            issues = issues.where(
              :sidfmd_host_id => SidfmdCategoriesHost.where(:sidfmd_category_id => f_params[:category_id].to_i)
                                                     .select(:sidfmd_host_id)
            )
          end
          if prev_status == 0 then
            issues = issues.where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          else
            issues = issues.where(:issues => { :status_id => prev_status })
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        SidfmdIssue.where(:issue_id => p_issue_ids).each do |i|
          update_data << {
            :sidfmd_issue_id => i.id,
            :html_id => i.html_id,
            :sidfmd_host_id => i.sidfmd_host_id,
            :status_id => new_status
          }
        end
        SidfmdIssue.change_statuses_with_journal(p_issue_ids, old_status_ids, new_status, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      logger.info(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.update_from_rule_action(update_data)
      SidfmdHostView.update_from_rule_action(update_data)
    end
    render :layout => false, :json => return_value
  end

  def update_due_dates
=begin    f_params = params.fetch(:schedule, {}).permit(:htmlids, :force, :due_date, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index_r
    htmlids = nil
    begin
      htmlids = JSON.parse(f_params[:htmlids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    unless htmlids then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    unless due_date then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .enable_hosts
          .categorized
          .joins(:issue)
          .where(:html_id => htmlids)
          .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        if !force then
          issue = issues
            .where(:issues => { :due_date => nil })
        end
        #issues.each do |issue|
        #  issue.with_lock do
        #    issue.change_due_date_with_journal(User.current.id, due_date, comment)
        #  end
        #end
        issue_ids = issues.map{|i| i.issue.id}
        SidfmdIssue.change_due_dates_with_journal(issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids, :shift_num, :ids, :force, :due_date, :comment, :category_id)
    err = false
    err_info = {}
    htmlids = nil
    shift_num = f_params[:shift_num]
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      htmlids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      logger.info(e)
      err = true
    end
    unless htmlids then
      err = true
    end
    force = f_params[:force] == 'true' ? true : false
    due_date = f_params[:due_date]
    unless due_date then
      err = true
    end
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        logger.info(e)
        err = true
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          logger.info(e)
          err = true
        end
      end
      if temp then
        due_date = temp
      else
        err = true
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    update_data = []
    begin
      SidfmdIssue.transaction do
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .enable_hosts
            .categorized
            .joins(:issue)
            .where(:html_id => htmlids)
            .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          if !@role_admin then
            issues = issues.where(:sidfmd_host_id => @user_host_ids)
          end
          if f_params[:category_id].to_i > 0 then
            issues = issues.where(
              :sidfmd_host_id => SidfmdCategoriesHost.where(:sidfmd_category_id => f_params[:category_id].to_i)
                                                     .select(:sidfmd_host_id)
            )
          end
          if !force then
            issues = issues.where(:issues => { :due_date => nil })
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        SidfmdIssue.where(:issue_id => p_issue_ids).each do |i|
          update_data << {
            :sidfmd_issue_id => i.id,
            :html_id => i.html_id,
            :sidfmd_host_id => i.sidfmd_host_id,
            :due_date => due_date
          }
        end
        SidfmdIssue.change_due_dates_with_journal(p_issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      logger.info(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.update_from_rule_action(update_data)
      SidfmdHostView.update_from_rule_action(update_data)
    end
    render :layout => false, :json => return_value
  end

  #カスタムタグの更新
  def update_ctags
    if !@sidfm_roles.include?(:use) then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    f_params = params.permit(:html_id, :ctag_ids)
    if !f_params[:html_id] or !f_params[:ctag_ids] then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    begin
      ctag_ids = JSON.parse(f_params[:ctag_ids])
    rescue JSON::ParserError => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    
    html_id = f_params[:html_id].to_i
    begin
      issues = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:html_id => html_id)
      SidfmdIssueCtag.udpate_issue_ctags(
        ctag_ids: ctag_ids,
        sidfmd_issue_rows: issues,
        role_admin: @role_admin,
        user_category_ids: @user_category_ids,
        user_host_ids: @user_host_ids
      )
    rescue Exception => e
      logger.info(e)
      render :layout => false, :json => { :err_message => 'カスタムタグの更新に失敗しました。' } and return
    end

    render :layout => false, :json => { :err_message => '' }
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

=begin
  #旧ダウンロード
  def download
    pages_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :htmlids)
    format = pages_params[:format]
    if ![ 'json', 'tsv' ].include?(format) then
      redirect_to :action => :index ,:notice => '不正な入力です。' and return
    end
    @vuls = { :url => request.url }
    @vuls[:category_id] = (pages_params[:category_id] || 0).to_i
    if @vuls[:category_id] > 0 then
      cat  = SidfmdCategory.find(@vuls[:category_id])
      if cat then
        @vuls[:category_name] = cat.name
      else
        @vuls[:category_id] = 0
        @vuls[:category_name] = '全ての脆弱性'
      end
    else
      @vuls[:category_name] = '全ての脆弱性'
    end
    @vuls[:idx] = (pages_params[:idx] || 'sri').to_sym
    @vuls[:htmlids] = []
    begin
      @vuls[:htmlids] = JSON.parse(pages_params[:htmlids] || '[]')
    rescue JSON::ParserError => e
      logger.error(s)
    end
    @corresp_required = Sidfm.corresp_required
    create_all_content(@vuls[:category_id])
    @vuls[:order] = pages_params[:order] || 'title_asc'
    @vuls[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    check_filters
    @vuls[:content_filtered] = apply_filters(@vuls[:all_content])

    case format
    when 'json' then
      json = {}
      temp_order = @vuls[:order].to_s.split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @vuls[:filters]
      json[:category] = { :id => @vuls[:category_id], :name => @vuls[:category_name], }
      json[:idx] = @vuls[:idx]
      json[:content] = []

      @vuls[:content_filtered].each do |vul|
        if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
          next
        end
        item = {}
        item[:html_id] = vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}"
        vul_title = if vul[:title].split(' / ').size != vul[:title].split(' / ').uniq.size then
                      vul[:title].split(' / ').uniq.join(' / ')
                    else
                      vul[:title].split(' / ').first
                    end
        item[:title] = vul_title
        item[:short_description] = (vul[:short_description] ? vul[:short_description] : '')
        item[:warning] = vul[:warning][@vuls[:idx]]
        count_status_ids = SidfmdIssue.count_status_ids(vul[:status_ids])
        item[:statuses] = count_status_ids
          .sort {|x, y| x[:id] <=> y[:id] }
          .collect {|x| { :name => x[:name], :value => x[:value] } }
        item[:max_sri] = SidfmdIssueSri::RISKS[vul[:max_sri]]
        t_cvss = vul[:cvss]
        item[:max_cvss] = (t_cvss[:label] == :notreg ? '' :sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
        cves = get_filtered_cves(vul[:cves], false)
        cvss_2_cves = []
        cvss_3_cves = []
        cve_ids = []
        cves.each do |cve|
          cve_ids << cve[:cve_id]
          cve[:c2_vector] = nil if cve[:c2_vector] && cve[:c2_vector].size == 0
          cve[:c3_vector] = nil if cve[:c3_vector] && cve[:c3_vector].size == 0
          cvss_2_cves << { :id => cve[:cve_id], :vector => cve[:c2_vector] }
          cvss_3_cves << { :id => cve[:cve_id], :vector => cve[:c3_vector] }
        end
        item[:cve_ids] = cve_ids
        item[:cvss_v2] = cvss_2_cves
        item[:cvss_v3] = cvss_3_cves
        item[:max_area] = SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name
        item[:exploitable] = (t_cvss[:label] != :notreg ? vul[:exploitable] : nil)
        item[:min_elapsed_days] = (vul[:min_elapsed_created_on][:display] >= 0 ? vul[:min_elapsed_created_on][:display] : '')
        item[:min_created_on] = vul[:min_created_on][:display]
        item[:hosts] = vul[:hosts].collect {|x| x[:display_name] }
        item[:categories] = vul[:categories].collect { |x| x[:name] }
        f_cnt = vul[:due_dates_unfinished].size
        if f_cnt > 0 and vul[:due_dates_unfinished].compact.size > 0 then
          cnt = vul[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
          item[:due_dates_percent] = display_percent(cnt.to_f / f_cnt.to_f * 100.0)
        else
          item[:due_dates_percent] = '0.0%'
        end

        item[:statuses_percent] = (vul[:status_ids].size > 0 ? display_finished_percent(vul[:status_ids]) : '0.0%')
        json[:content] << item
      end

      send_data(json.to_json, :filename => "vmlist-vulnerabilities-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
      return
    when 'tsv' then
      header = [ 'HTML_ID', '要対応', '対応状況', '最大SRI', 'CVSS', 'CVE ID', 'CVSS V2', 'CVSS V3', '影響範囲', '攻撃コード', '経過日数', '検出日', '脆弱性タイトル', '概要', 'ホスト名', 'カテゴリ', '対策予定日設定済', '対処率' ]
      # TSV, UTF-16LE, BOMあり, 拡張子 .csv
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @vuls[:content_filtered].each do |vul|
          if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
            next
          end
          item.clear
          item << (vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}").encode('UTF-16LE')
          item << vul[:warning][@vuls[:idx]].to_s.encode('UTF-16LE')
          count_status_ids = SidfmdIssue.count_status_ids(vul[:status_ids])
          item << count_status_ids.sort {|x, y| x[:id] <=> y[:id] }.collect {|x| "#{x[:name]}:#{x[:value]}" }.join(',').encode('UTF-16LE')
          item << (vul[:max_sri] ? SidfmdIssueSri::RISKS[vul[:max_sri]] : '').encode('UTF-16LE')
          t_cvss = vul[:cvss]
          item << (t_cvss[:label] == :notreg ? '' :sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          cves = get_filtered_cves(vul[:cves], false)
          cvss_2_cves = []
          cvss_3_cves = []
          cve_ids = []
          cves.each do |cve|
            cve_ids << cve[:cve_id]
            cvss_2_cves << "#{cve[:cve_id]}|#{cve[:c2_vector]}" if cve[:c2_vector] && cve[:c2_vector].size > 0
            cvss_3_cves << "#{cve[:cve_id]}|#{cve[:c3_vector]}" if cve[:c3_vector] && cve[:c3_vector].size > 0
          end
          item << cve_ids.join(',').encode('UTF-16LE')
          item << cvss_2_cves.join(',').encode('UTF-16LE')
          item << cvss_3_cves.join(',').encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name.encode('UTF-16LE')
          item << (t_cvss[:label] != :notreg ? vul[:exploitable].to_s : '').encode('UTF-16LE')
          item << (vul[:min_elapsed_created_on][:display] >= 0 ? vul[:min_elapsed_created_on][:display].to_s : '').encode('UTF-16LE')
          item << vul[:min_created_on][:display].encode('UTF-16LE')
          vul_title = if vul[:title].split(' / ').size != vul[:title].split(' / ').uniq.size then
                        vul[:title].split(' / ').uniq.join(' / ')
                      else
                        vul[:title].split(' / ').first
                      end
          item << vul_title.encode('UTF-16LE')
          item << (vul[:short_description] ? vul[:short_description] : '').encode('UTF-16LE')
          item << vul[:hosts].collect {|x| x[:display_name] }.join(',').encode('UTF-16LE')
          item << vul[:categories].collect { |x| escape_category_tsv(x[:name]) }.join(',').encode('UTF-16LE')
          f_cnt = vul[:due_dates_unfinished].size
          if f_cnt > 0 and vul[:due_dates_unfinished].compact.size > 0 then
            cnt = vul[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
            item << display_percent(cnt.to_f / f_cnt.to_f * 100.0).encode('UTF-16LE')
          else
            item << '0.0%'.encode('UTF-16LE')
          end
          item << (vul[:status_ids].size > 0 ? display_finished_percent(vul[:status_ids]) : '0.0%').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlist-vulnerabilities-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end

    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end
=end
#=begin

  MAX_CELL_LENGTH = 30000

  def read_download_data
    GC.start
    f_params = params.permit(:target, :executed_size, :process, :category_id, :idx, :order, :filters, :ids)
    if !f_params[:ids] or !f_params[:target] or !f_params[:order] or !f_params[:filters] or 
       !f_params[:executed_size] or !f_params[:process] or !f_params[:idx] or !f_params[:category_id]
    then
      logger.info('params error')
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    format = f_params[:target]
    if ![ 'json', 'tsv', 'affect-products-packages-tsv' ].include?(format) then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    executed_size = f_params[:executed_size].to_i
    get_summary_prior_category

    @vuls = { :url => request.url }
    @vuls[:category_id] = (f_params[:category_id] || 0).to_i
    if @vuls[:category_id] > 0 then
      cat = SidfmdCategory.find_by_id(@vuls[:category_id])
      if cat then
        @vuls[:category_name] = cat.name
      else
        @vuls[:category_id] = 0
        @vuls[:category_name] = '全ての脆弱性'
      end
    else
      @vuls[:category_name] = '全ての脆弱性'
    end
    @vuls[:idx] = (f_params[:idx] || 'sri').to_sym
    @vuls[:htmlids] = []
    begin
      @vuls[:ids] = JSON.parse(f_params[:ids] || '[]') #processがinitの時(最初の時)はhtml_idが入り、2回目以降はSidfmdVulVeiwのidが入ることに注意
    rescue JSON::ParserError => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    @vuls[:order] = f_params[:order] || 'title_asc'
    @vuls[:filters] = JSON.parse((f_params[:filters] || '{}'), symbolize_names: true)
    create_vul_view_content

    case format
    when 'json' then
      if f_params[:process] == 'init' then
        @vuls[:content_filtered] = @vuls[:content_filtered].where(:html_id => @vuls[:ids]) if @vuls[:ids].any? 
        temp_order = @vuls[:order].to_s.split(/_/)
        init_json_summary_table_progress_download(
          :ids => @vuls[:content_filtered].pluck(:id),
          :order_col_name => temp_order[0],
          :order_type => temp_order[1],
          :filters => @vuls[:filters],
          :max_download_size => @vuls[:content_filtered].count,
          :category => { :id => @vuls[:category_id], :name => @vuls[:category_name] },
          :idx => @vuls[:idx]
        )
      else
        create_json_data = Proc.new{|id|
          vul = @vuls[:content_filtered].find_by(:id => id)
          next unless vul
          vul_values = vul.create_summary_values(@role_admin, @vuls[:category_id], @user_category_ids, @user_host_ids, @corresp_required)
          next unless vul_values
          item = {}
          item[:html_id] = vul_values[:html_id] > 0 ? "#{vul_values[:html_id]}" : "KB#{-vul_values[:html_id]}"
          item[:title] = vul_values[:title]
          item[:short_description] = (vul_values[:short_description] ? vul_values[:short_description] : '')
          item[:warning] = vul_values[:"warning_#{@vuls[:idx]}"]
          count_status_ids = SidfmdIssue.count_status_ids(vul_values[:status_ids])
          item[:statuses] = count_status_ids
            .sort {|x, y| x[:id] <=> y[:id] }
            .collect {|x| { :name => x[:name], :value => x[:value] } }
          item[:max_sri] = SidfmdIssueSri::RISKS[vul_values[:max_sri]]
          item[:max_cvss] = (vul_values[:cvss_label] == 'notreg' ? '' : sprintf('%.1f', (vul_values[:cvss_score] ? vul_values[:cvss_score] : 0).to_f))
          cves = vul.get_summary_cves
          cvss_2_cves = []
          cvss_3_cves = []
          cve_ids = []
          cves.each do |cve|
            cve_ids << cve.cve_id
            cvss_2_cves << { :id => cve.cve_id, :vector => cve.c2_vector_index }
            cvss_3_cves << { :id => cve.cve_id, :vector => cve.c3_vector_index }
          end
          item[:cve_ids] = cve_ids
          item[:cvss_v2] = cvss_2_cves
          item[:cvss_v3] = cvss_3_cves
          item[:max_area] = SidfmdArea.where(:area_id => vul_values[:softek_area_id]).first.name
          item[:exploitable] = vul_values[:exploitable]
          item[:min_elapsed_days] = (vul_values[:min_elapsed_created_on_display] >= 0 ? vul_values[:min_elapsed_created_on_display] : '')
          item[:min_created_on] = vul_values[:min_created_on_display]
          item[:hosts] = vul_values[:hosts].pluck(:display_name)
          item[:categories] = (
            @summary_prior_category == 'custom' ?
            vul_values[:custom_categories].pluck(:name) + vul_values[:system_categories].pluck(:name) :
            vul_values[:system_categories].pluck(:name) + vul_values[:custom_categories].pluck(:name)
          )
          item[:due_dates_percent] = display_percent(vul_values[:duedates_percent])
          item[:statuses_percent] = display_percent(vul_values[:fixed_percent])
          item
        }
        json_tsv_summary_table_progress_download(@vuls[:ids], create_json_data, executed_size)
      end
    when 'tsv' then
      header = [ 'HTML_ID', '要対応', '対応状況', '最大SRI', 'CVSS', 'CVE ID', 'CVSS V2', 'CVSS V3', '影響範囲', '攻撃コード', '経過日数', '検出日', '脆弱性タイトル', '概要', 'ホスト名', 'カテゴリ', '対策予定日設定済', '対処率' ]
      if f_params[:process] == 'init' then
        @vuls[:content_filtered] = @vuls[:content_filtered].where(:html_id => @vuls[:ids]) if @vuls[:ids].any?
        init_tsv_summary_table_progress_download(@vuls[:content_filtered].pluck(:id), header, @vuls[:content_filtered].count)
      else
        create_tsv_data = Proc.new{|id|
          vul = @vuls[:content_filtered].find_by(:id => id)
          next unless vul
          vul_values = vul.create_summary_values(@role_admin, @vuls[:category_id], @user_category_ids, @user_host_ids, @corresp_required)
          next unless vul_values 
          item = []
          item << (vul_values[:html_id] > 0 ? "#{vul_values[:html_id]}" : "KB#{-vul_values[:html_id]}")
          item << vul_values[:"warning_#{@vuls[:idx]}"].to_s
          count_status_ids = SidfmdIssue.count_status_ids(vul_values[:status_ids])
          item << count_status_ids.sort {|x, y| x[:id] <=> y[:id] }.collect {|x| "#{x[:name]}:#{x[:value]}" }.join(',')
          item << SidfmdIssueSri::RISKS[vul_values[:max_sri]]
          item << (vul_values[:cvss_label] == 'notreg' ? '' : sprintf('%.1f', (vul_values[:cvss_score] ? vul_values[:cvss_score] : 0).to_f))
          cves = vul.get_summary_cves  
          cvss_2_cves = []
          cvss_3_cves = []
          cve_ids = []
          cves.each do |cve|
            cve_ids << cve.cve_id
            cvss_2_cves << "#{cve.cve_id}|#{cve.c2_base_score}|#{cve.c2_vector_index}" if cve.c2_av
            cvss_3_cves << "#{cve.cve_id}|#{cve.c3_base_score}|#{cve.c3_vector_index}" if cve.c3_av
          end
          item << cve_ids.join(',')
          item << cvss_2_cves.join(',')
          item << cvss_3_cves.join(',')
          item << SidfmdArea.where(:area_id => vul_values[:softek_area_id]).first.name
          item << vul_values[:exploitable].to_s
          item << (vul_values[:min_elapsed_created_on_display] >= 0 ? vul_values[:min_elapsed_created_on_display].to_s : '')
          item << vul_values[:min_created_on_display]
          item << vul_values[:title]
          item << (vul_values[:short_description] ? vul_values[:short_description] : '')
          item << vul_values[:hosts].pluck(:display_name).join(',')
          item << (
            @summary_prior_category == 'custom' ?
            vul_values[:custom_categories].pluck(:name) + vul_values[:system_categories].pluck(:name) :
            vul_values[:system_categories].pluck(:name) + vul_values[:custom_categories].pluck(:name)
          ).collect { |cat_name| escape_category_tsv(cat_name) }
           .join(',')
          item << display_percent(vul_values[:duedates_percent])
          item << display_percent(vul_values[:fixed_percent])
          item
        }
        json_tsv_summary_table_progress_download(@vuls[:ids], create_tsv_data, executed_size)
      end
    when 'affect-products-packages-tsv' then
      header = [ 'HTML_ID', 'ホストID', 'ホスト名', '表示名', '製品名', 'パッケージ名' ]
      if f_params[:process] == 'init' then
        @vuls[:content_filtered] = @vuls[:content_filtered].where(:html_id => @vuls[:ids]) if @vuls[:ids].any?
        init_tsv_summary_table_progress_download(@vuls[:content_filtered].pluck(:id), header, @vuls[:content_filtered].count)
      else
        create_cell_data = -> (items, cell_over_items){
          cell = ''
          is_over_cell = false
          while target = items.shift() do
            if target.size <= MAX_CELL_LENGTH then
              if (cell.size > 0 ? cell.size + 1 : 0) + target.size > MAX_CELL_LENGTH then
                cell_over_items << target
                is_over_cell = true
              else
                if !is_over_cell then
                  cell += ',' if cell.size > 0
                  cell += target
                else
                  cell_over_items << target
                end
              end
            end
          end
          cell
        }
        create_tsv_data = Proc.new{|id|
          vul = @vuls[:content_filtered].find_by(:id => id)
          next unless vul
          vul_values = vul.create_summary_values(@role_admin, @vuls[:category_id], @user_category_ids, @user_host_ids, @corresp_required)
          next unless vul_values
          html_id = (vul_values[:html_id] > 0 ? "#{vul_values[:html_id]}" : "KB#{-vul_values[:html_id]}")
          local_vul_lapps = SidfmdLocalVulLapp.joins(:sidfmd_local_vul).where(:sidfmd_local_vuls => { :html_id => vul_values[:html_id] })
          vul_values[:hosts].to_a.uniq{|host| host.sidfmd_host_id}.each do |vv_host| #タイミングによって、ホストが重複することがあったため、to_aでuniqにする。
            host = SidfmdHost.find_by(:id => vv_host.sidfmd_host_id)
            next unless host
            sidfmd_issue = SidfmdIssue.where(:html_id => vul_values[:html_id], :sidfmd_host_id => host.id).first
            next unless sidfmd_issue
            item = []
            products = []
            lproducts = []
            pkgs = []
            item << html_id
            item << host.id.to_s
            item << host.hostname
            item << host.display_name
            dup_item = item.dup
            #プロダクト
            products |= sidfmd_issue.sidfmd_issue_affprods.select(:name).map{|prod| escape_category_tsv(prod.name) }
            products |= sidfmd_issue.sidfmd_issue_affvers.select(:fullname).map{|ver| escape_category_tsv(ver.fullname) }
            products = products.sort()

            #ローカルプロダクト
            host_lapps = host.sidfmd_host_lapps
            lproducts |= SidfmdLproduct.where(
              :id => local_vul_lapps.select(:sidfmd_lproduct_id)
            ).where(
              :id => host_lapps.where("sidfmd_lversion_id IS NULL OR sidfmd_lversion_id = 0").select(:sidfmd_lproduct_id)
            )
            .select(:name)
            .map{|lprod| escape_category_tsv(lprod.name) }

            lproducts |= SidfmdLversion.where(
              :id => local_vul_lapps.select(:sidfmd_lversion_id)
            ).where(
              :id => host_lapps.where("sidfmd_lversion_id IS NOT NULL AND sidfmd_lversion_id != 0").select(:sidfmd_lversion_id)
            )
            .select(:fullname)
            .map{|lver| escape_category_tsv(lver.fullname) }
            lproducts = lproducts.sort()
            all_products = (products + lproducts)
            cell_over_products = [] #最大文字数がMAX_CELL_LENGTHを超えた分を格納する
            products_cell = create_cell_data.(all_products, cell_over_products)
            item << products_cell
 
            #パッケージ
            pkgs = sidfmd_issue.affect_packages('').map{|pkg| escape_category_tsv(pkg[:fullname])}
            cell_over_pkgs = [] #最大文字数がMAX_CELL_LENGTHを超えた分を格納する
            pkgs_cell = create_cell_data.(pkgs, cell_over_pkgs)
            item << pkgs_cell

            session[:export_data][:content] << item
            
            is_continue = cell_over_products.size > 0 || cell_over_pkgs.size > 0
            while is_continue do
              row = dup_item.dup
              continue_cell_over_products = []
              continue_cell_over_pkgs = []
              products_cell = create_cell_data.(cell_over_products, continue_cell_over_products)
              row << products_cell
              pkgs_cell = create_cell_data.(cell_over_pkgs, continue_cell_over_pkgs)
              row << pkgs_cell
              session[:export_data][:content] << row
              cell_over_products = continue_cell_over_products
              cell_over_pkgs = continue_cell_over_pkgs
              is_continue = cell_over_products.size > 0 || cell_over_pkgs.size > 0
            end
          end 
          item = nil
        }
        json_tsv_summary_table_progress_download(@vuls[:ids], create_tsv_data, executed_size)
      end
    end

  end

  def download
    f_params = params.fetch(:download, {}).permit(:format)
    case f_params[:format]
    when 'json' then
      send_summary_table_json_file("vmlist-vulnerabilities-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json")
    when 'tsv' then
      send_summary_table_tsv_file("vmlist-vulnerabilities-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    when 'affect-products-packages-tsv' then
      send_summary_table_tsv_file("vmlist-vulnerabilities-affect-products-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv")
    end
  end
#=end

  def send_mail
    filtered_params = params.permit(:recipients, :subject, :message)
  begin
    mail = {}
    mail[:from] = User.current.mail
    email_with_name = []
    recipients = JSON.parse(filtered_params[:recipients])
    recipients.each do |address, name|
      email_name = %("#{name}" <#{address}>)
      email_with_name << email_name
    end
    #mail[:to] = email_with_name
    mail[:subject] = filtered_params[:subject]
    mail[:message] = filtered_params[:message]
    send_mail_num = SidfmdMailer::SEND_MAIL_NUM
    time_out_second = SidfmdMailer::SEND_MAIL_TIMEOUT_SECOND
    while (send_mail_targets = email_with_name.shift(send_mail_num)).size > 0 do
      threads = send_mail_targets.map{|fullname_and_address|
        Thread.new(mail.deep_dup, fullname_and_address){|cp_mail, data|
          ret = {}
          cp_mail[:to] = data
          begin
            SidfmdMailer.send_mail(cp_mail).deliver_now
          rescue Exception => e
            ret[:exception] = e
          end
          ret
        }
      }
      threads.each do |result|
        result = result.join(time_out_second)
        if result then
          result = result.value
          if result[:exception] then logger.error("failed send mail. #{result[:exception]}") end
        else
          logger.error("time-out of sending a mail")
        end
      end
    end
    #SidfmdMailer.send_mail(mail).deliver_now
  rescue Exception => e
    log_exception(e)
    render :layout => false, :json => { :err_message => 'メールの送信に失敗しました。' } and return
    #redirect_to :action => :index, :notice => e and return
  end
    render :layout => false, :json => { :err_message => '' }
  end

=begin
  def api_get_display_summary_rows
    GC.start
    f_params = params.permit(:ids, :category_id, :idx, :checked, :filters)
    category_id = (f_params[:category_id] || 0).to_i
    html_ids = JSON.parse(f_params[:ids])
    logger.info('&&&&&&&&' + html_ids.inspect)
    filters = JSON.parse(f_params[:filters], :symbolize_names => true)
    checked = JSON.parse(f_params[:checked]).map(&:to_i)
    @vuls = {}
    idx = f_params[:idx].to_sym
    corresp_required = Sidfm.corresp_required
    get_summary_prior_category
    create_all_vul_view_content(category_id)
    display_content = SidfmdVulView.bulk_create_summary_values(
      :vul_views => @vuls[:all_content].where(:html_id => html_ids),
      :role_admin => @role_admin,
      :category_id => category_id,
      :user_category_ids => @user_category_ids,
      :user_host_ids => @user_host_ids,
      :corresp_required => corresp_required
    )
    ret = {}
    display_content.each do |vul|
      ret[vul.html_id] = create_summary_vul_html(
        vul,
        {
          :category_id => category_id,
          :idx => idx,
          :corresp_required => corresp_required,
          :checked =>  checked,
          :filters => filters,
          :summary_prior_category => @summary_prior_category,
        }
      )
    end
    
    render :layout => false, :json => ret
  end
=end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:per_page)#params.fetch(:pages, {}).permit(:order, :offset, :per_page, :category_id, :idx, :filters, :checked, :all_checked)
    parse_params
    @vuls[:per_page] = f_params[:per_page].to_i
    get_summary_prior_category
    @corresp_required = Sidfm.corresp_required
    check_filters
    create_all_vul_view_content(@vuls[:category_id])
    @filter_category_host_ids = if @vuls[:category_id] > 0 then
                                  if @role_admin then
                                    SidfmdCategoriesHost.where(:sidfmd_category_id => @vuls[:category_id])
                                                        .pluck(:sidfmd_host_id)
                                  else
                                    SidfmdCategoriesHost.where(:sidfmd_category_id => @vuls[:category_id], :sidfmd_host_id => @user_host_ids)
                                                        .pluck(:sidfmd_host_id)
                                  end
                                else
                                  nil
                                end
    @vuls[:content_filtered] = apply_filters_to_vul_view_content(@vuls[:all_content])
    @vuls[:content] = (@vuls[:content_filtered].limit(@vuls[:per_page]).offset(@vuls[:offset]) || [])
    display_content = SidfmdVulView.bulk_create_summary_values(
      :vul_views => @vuls[:content], 
      :role_admin => @role_admin,
      :category_id => @vuls[:category_id],
      :user_category_ids => @user_category_ids,
      :user_host_ids => @user_host_ids,
      :corresp_required => @corresp_required
    )
    ret = {
      :rows => display_content.map{|vul|
        create_summary_vul_html(
          vul,
          {
            :category_id => @vuls[:category_id],
            :idx => @vuls[:idx],
            :corresp_required => @corresp_required,
            :checked =>  @vuls[:checked],
            :filters => @vuls[:filters],
            :summary_prior_category => @summary_prior_category,
          }
        )
      },
      :total => @vuls[:content_filtered].count,
      :all_ids => @vuls[:content_filtered].pluck(:html_id),
      :sidfm_roles => @sidfm_roles,
      :user_cats => @user_category_ids
    }
    render :layout => false, :json => ret
  end

end
