#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

# ローカル脆弱性コンテンツ
class SidfmdLocalVul < ActiveRecord::Base
  
  # html_idの最小・最大値
  MIN_HTMLID = 90000000
  MAX_HTMLID = 99999999

  has_many :sidfmd_local_vul_logs, dependent: :destroy
  has_many :sidfmd_local_vul_apps, dependent: :destroy
  has_many :sidfmd_local_vul_lapps, dependent: :destroy
  has_many :sidfmd_local_vul_cves, dependent: :destroy
  has_many :sidfmd_local_vul_sris, dependent: :destroy

  validates :html_id, numericality: { :only_integer => true, :greater_than_or_equal_to => MIN_HTMLID, :less_than_or_equal_to => MAX_HTMLID }
  validates :html_id, uniqueness: true

  before_validation :apply_htmlid, on: :create

  def apply_htmlid
    unless self.html_id then
      r = self.class.maximum(:html_id)
      if r then
        r += 1
        if r >= MIN_HTMLID and r <= MAX_HTMLID then
          self.html_id = r
        end
      else
        self.html_id = MIN_HTMLID
      end
    end
  end

  #更新されたローカル脆弱性とSidfmdIssueの整合性をとる
  def self.establish_consistency(local_vul, sidfmd_issue, issue, products, prods_info, versions_info)
    is_change = false
    if local_vul.updated_at > issue.updated_on then
      is_change = true #CVEの比較でc3_update_datetimeが更新されるため、このスコープに入ったら、更新履歴の追加のみでも変更ありとする
      #SidfmdIssueの比較
      if sidfmd_issue.title != local_vul.title || sidfmd_issue.area_id != local_vul.area_id || sidfmd_issue.risk_id != local_vul.risk_id || sidfmd_issue.short_description != local_vul.short_description then
        sidfmd_issue.update(
          title: local_vul.title,
          area_id: local_vul.area_id,
          risk_id: local_vul.risk_id,
          short_description: local_vul.short_description
        )
      end
     
      #CVEの比較
      sidfmd_issue_cves = sidfmd_issue.sidfmd_issue_cves
      cve_ids = []
      SidfmdLocalVulCve.where(:sidfmd_local_vul_id => local_vul.id).each do |cve|
        cve_ids << cve.cve_id
        target_cve = sidfmd_issue_cves.where(:cve_id => cve.cve_id).first
        if target_cve then
          target_cve.update(  
            c3_vector: cve.c3_vector,
            c3_base_score: cve.c3_base_score,
            c3_update_datetime: cve.updated_at,
            c3_exploitable: cve.c3_exploitable
          )
        else
          SidfmdIssueCve.create(
            sidfmd_issue_id: sidfmd_issue.id,
            cve_id: cve.cve_id,
            c3_vector:cve.c3_vector,
            c3_base_score: cve.c3_base_score,
            c3_create_datetime: cve.created_at,
            c3_update_datetime: cve.updated_at,
            c3_exploitable: cve.c3_exploitable
          )
        end
      end
      sidfmd_issue_cves.where.not(:cve_id => cve_ids).destroy_all #削除されたcve_idを削除
    
      #SRIの比較
      sidfmd_issue_sris = sidfmd_issue.sidfmd_issue_sris
      SidfmdLocalVulSri.where(:sidfmd_local_vul_id => local_vul.id).each do |sri|
        target_sri = sidfmd_issue_sris.where(:category_id => sri.category_id).first
        if target_sri && (target_sri.risk_id != sri.risk_id || target_sri.memo != sri.memo) then
          target_sri.update(
            risk_id: sri.risk_id,
            memo: sri.memo,
          )
        end
      end
    end

    #プロダクトの比較
    affprods = sidfmd_issue.sidfmd_issue_affprods 
    affvers = sidfmd_issue.sidfmd_issue_affvers
    aff_product_ids = affprods.pluck(:product_id)
    aff_version_ids = affvers.pluck(:product_version_id)
    prod_ids = []
    ver_ids = []
    new_affprods = []
    new_affvers = []
    products.each do |product|
      if product.vid && product.vid != 0 then
        unless ver_ids.include?(product.vid) then
          ver_ids << product.vid
          v_info = versions_info[product.vid]
          unless aff_version_ids.include?(product.vid) then
            new_affvers << SidfmdIssueAffver.new(
              sidfmd_issue_id: sidfmd_issue.id,
              product_version_id: product.vid,
              product_id: product.pid,
              fullname: v_info[:fullname],
              version: v_info[:version]
            )
          end
        end
      else
        unless prod_ids.include?(product.pid) then
          prod_ids << product.pid
          unless aff_product_ids.include?(product.pid) then
            new_affprods << SidfmdIssueAffprod.new(
              sidfmd_issue_id: sidfmd_issue.id,
              product_id: product.pid,
              name: prods_info[product.pid][:name]
            )
          end
        end
      end
    end
    SidfmdIssueAffprod.import(new_affprods, validate: false) if new_affprods.any?
    SidfmdIssueAffver.import(new_affvers, validate: false) if new_affvers.any?
    #削除されたプロダクトを削除
    affprods.where.not(:product_id => prod_ids).destroy_all
    affvers.where.not(:product_version_id => ver_ids).destroy_all

    is_change
  end
  #脆弱性チケットの登録・更新を行う。
  #返り値として{
  #:add=>新たに追加された脆弱性のsidfmd_issue_idの配列, 
  #:modify=>ローカル脆弱性内容の更新もしくは対応状況を「未対策」に変更した脆弱性のsidfmd_issue_idの配列, 
  #:delete=>この処理で対応状況を「対策済」「影響無」に変更した脆弱性のsidfmd_issue_idの配列
  #}を返す。
  
  ADMIN_USER = User.find_by_login('sidfmadm')
  VM_TRACKER = Tracker.find_by_name('SIDfm脆弱性管理')
  VM_PROJECT = Project.find_by_name('脆弱性情報管理')
  DEFAULT_PRIORITY = IssuePriority.find_by_position_name('default')
  ISSUE_STATUS_NOTCONFIRMED = IssueStatus.find(1).name
  ISSUE_STATUS_FIXED = IssueStatus.find(3).name
  ISSUE_STATUS_NOTAFFECT = IssueStatus.find(4).name
  NOTE_NO_LONGER_DETECTED = "脆弱性情報との照合の結果、脆弱性が検出されなくなったため、ステータスを#{ISSUE_STATUS_FIXED}に変更しました。" 
  NOTE_CHANGE_PRODVER = "対象の脆弱性について、影響のあるプロダクトの情報が更新されたため、ステータスを#{ISSUE_STATUS_NOTCONFIRMED}に変更しました。お手数ですが脆弱性の内容を再度ご確認の上、対応をお願いします。"
  NOTE_DISABLED = "脆弱性情報との照合の結果、脆弱性が無効になったため、ステータスを#{ISSUE_STATUS_NOTAFFECT}に変更しました。" 
  NOTE_ENABLED = "脆弱性情報との照合の結果、脆弱性が有効になったため、ステータスを#{ISSUE_STATUS_NOTCONFIRMED}に変更しました。"
  
  def self.apply_vuls_to_hosts(host_ids)
    new_sidfmd_issue_ids = []
    modified_sidfmd_issue_ids = []
    deleted_sidfmd_issue_ids = []
    
    #ローカル脆弱性に紐づく手動登録プロダクトが登録されているホストとプロダクトの情報
   
    host_product_ids_of_all = {}
    host_all_product_ids = {}
    host_all_version_ids = {}
    SidfmdHostApp.where(:sidfmd_host_id => host_ids).each do|app|
      if !app.version_id || app.version_id == 0 then
        host_product_ids_of_all[app.sidfmd_host_id] = [] unless host_product_ids_of_all[app.sidfmd_host_id]
        host_product_ids_of_all[app.sidfmd_host_id] << app.product_id
      else
        host_all_version_ids[app.sidfmd_host_id] = [] unless host_all_version_ids[app.sidfmd_host_id]
        host_all_version_ids[app.sidfmd_host_id] << app.version_id
      end
      host_all_product_ids[app.sidfmd_host_id] = [] unless host_all_product_ids[app.sidfmd_host_id]
      host_all_product_ids[app.sidfmd_host_id] << app.product_id  
    end

    product_related_host_vuls = {}
    host_pids = []
    host_vids = []
    SidfmdHost.joins(:sidfmd_host_apps)
              .joins('LEFT OUTER JOIN sidfmd_local_vul_apps ON sidfmd_local_vul_apps.pid = sidfmd_host_apps.product_id')
              .joins('LEFT OUTER JOIN sidfmd_local_vuls ON sidfmd_local_vul_apps.sidfmd_local_vul_id = sidfmd_local_vuls.id')
              .where('sidfmd_host_apps.sidfmd_host_id IN (?)', host_ids)
              .where('sidfmd_local_vuls.id IS NOT NULL')
              .select(
                'sidfmd_local_vuls.id AS id',
                'sidfmd_local_vuls.flg_enable AS flg_enable',
                'sidfmd_local_vuls.html_id AS html_id',
                'sidfmd_local_vuls.title AS title',
                'sidfmd_local_vuls.risk_id AS risk_id',
                'sidfmd_local_vuls.area_id AS area_id',
                'sidfmd_local_vuls.short_description AS short_description',
                'sidfmd_local_vuls.created_at AS created_at',
                'sidfmd_local_vuls.updated_at AS updated_at',
                'sidfmd_hosts.id AS host_id',
                'sidfmd_hosts.hostname AS hostname', 
                'sidfmd_local_vul_apps.pid AS vul_pid',
                'sidfmd_local_vul_apps.vid AS vul_vid',
                'sidfmd_host_apps.product_id AS pid',
                'sidfmd_host_apps.version_id AS vid',
              )
              .each do |vul|
                unless product_related_host_vuls[vul.host_id] then
                  product_related_host_vuls[vul.host_id] = []
                end
                pass = false
                if !vul.vul_vid || vul.vul_vid == 0 then
                  #バージョンが全ての場合
                  host_product_ids = host_all_product_ids[vul.host_id]
                  if host_product_ids && host_product_ids.include?(vul.vul_pid) then
                    pass = true
                  end
                else
                  #バージョンが全てでない場合
                  host_version_ids = host_all_version_ids[vul.host_id]
                  if host_version_ids && host_version_ids.include?(vul.vul_vid) then
                    pass = true
                  else
                    #ホストのバージョンに無い場合は、ホストの「全て」のバージョンに含まれていれば対象となる
                    host_product_ids = host_product_ids_of_all[vul.host_id]
                    if host_product_ids && host_product_ids.include?(vul.vul_pid) then
                      pass = true
                    end
                  end
                end 
                if pass then
                  product_related_host_vuls[vul.host_id] << vul
                  if !vul.vid || vul.vid == 0 then
                    host_pids << vul.pid
                  else
                    host_vids << vul.vid
                  end
                end
              end
    prods_info = {}
    versions_info = {}
    SidfmdProduct.where(:pid => host_pids).each do |p|
      prods_info[p.pid] = {:name => p.name}
    end
    SidfmdVersion.where(:vid => host_vids).each do |v|
      versions_info[v.vid] = {:version => v.version, :fullname => v.fullname}
    end 
    #ローカル脆弱性に紐づくローカルプロダクトが登録されているホスト
    
    host_lproduct_ids_of_all = {}
    host_all_lproduct_ids = {}
    host_all_lversion_ids = {}
    SidfmdHostLapp.where(:sidfmd_host_id => host_ids).each do |lapp|
      if !lapp.sidfmd_lversion_id || lapp.sidfmd_lversion_id == 0 then
        host_lproduct_ids_of_all[lapp.sidfmd_host_id] = [] unless host_lproduct_ids_of_all[lapp.sidfmd_host_id]
        host_lproduct_ids_of_all[lapp.sidfmd_host_id] << lapp.sidfmd_lproduct_id
      else
        host_all_lversion_ids[lapp.sidfmd_host_id] = [] unless host_all_lversion_ids[lapp.sidfmd_host_id]
        host_all_lversion_ids[lapp.sidfmd_host_id] << lapp.sidfmd_lversion_id
      end
      host_all_lproduct_ids[lapp.sidfmd_host_id] = [] unless host_all_lproduct_ids[lapp.sidfmd_host_id]
      host_all_lproduct_ids[lapp.sidfmd_host_id] << lapp.sidfmd_lproduct_id
    end

    lproduct_related_host_vuls = {}
    SidfmdHost.joins(:sidfmd_host_lapps)
              .joins(
                'LEFT OUTER JOIN sidfmd_local_vul_lapps ON sidfmd_local_vul_lapps.sidfmd_lproduct_id = sidfmd_host_lapps.sidfmd_lproduct_id'
              )
              .joins('LEFT OUTER JOIN sidfmd_local_vuls ON sidfmd_local_vul_lapps.sidfmd_local_vul_id = sidfmd_local_vuls.id') 
              .where('sidfmd_host_lapps.sidfmd_host_id IN (?)', host_ids) 
              .where('sidfmd_local_vuls.id IS NOT NULL')
              .select(
                'sidfmd_local_vuls.id AS id',
                'sidfmd_local_vuls.flg_enable AS flg_enable',
                'sidfmd_local_vuls.html_id AS html_id',
                'sidfmd_local_vuls.title AS title',
                'sidfmd_local_vuls.risk_id AS risk_id',
                'sidfmd_local_vuls.area_id AS area_id',
                'sidfmd_local_vuls.short_description AS short_description',
                'sidfmd_local_vuls.created_at AS created_at',
                'sidfmd_local_vuls.updated_at AS updated_at',
                'sidfmd_local_vul_lapps.sidfmd_lproduct_id AS lpid',
                'sidfmd_local_vul_lapps.sidfmd_lversion_id AS lvid',
                'sidfmd_hosts.id AS host_id',
                'sidfmd_hosts.hostname AS hostname'
              )
              .each do |vul|
                unless lproduct_related_host_vuls[vul.host_id] then
                  lproduct_related_host_vuls[vul.host_id] = []
                end
                pass = false
                if !vul.lvid || vul.lvid == 0 then
                  #バージョンが全ての場合
                  host_lproduct_ids = host_all_lproduct_ids[vul.host_id]
                  if host_lproduct_ids && host_lproduct_ids.include?(vul.lpid) then
                    pass = true
                  end
                else
                  #バージョンが全てでない場合
                  host_lversion_ids = host_all_lversion_ids[vul.host_id]
                  if host_lversion_ids && host_lversion_ids.include?(vul.lvid) then
                    pass = true
                  else
                    #ホストのバージョンに無い場合は、ホストの「全て」のバージョンに含まれていれば対象となる
                    host_lproduct_ids = host_lproduct_ids_of_all[vul.host_id]
                    if host_lproduct_ids && host_lproduct_ids.include?(vul.lpid) then
                      pass = true
                    end
                  end
                end
                lproduct_related_host_vuls[vul.host_id] << vul if pass
              end
    
    checked_host_html_ids = {}
    host_ids = host_ids.uniq
    host_ids.each do |host_id|
      begin
        host_id = host_id.to_i
        product_related_vuls = product_related_host_vuls[host_id] 
        product_related_vuls = [] unless product_related_vuls

        lproduct_related_vuls = lproduct_related_host_vuls[host_id]
        lproduct_related_vuls = [] unless lproduct_related_vuls
        
        vuls = product_related_vuls + lproduct_related_vuls
        #ホストの紐づきが無くなったSidfmdIssueは影響無にする
        SidfmdIssue.where(:sidfmd_host_id => host_id)
                   .where('html_id >= ?', SidfmdLocalVul::MIN_HTMLID)
                   .where.not(:html_id => vuls.map{|v| v.html_id })
                   .where(:issue_id => Issue.where(:status_id => [ SidfmdIssue::STATUS_NOTCONFIRMED, SidfmdIssue::STATUS_ONGOING ]).select(:id))
                   .each do |sidfmd_issue|
                     vul = SidfmdLocalVul.find_by(:html_id => sidfmd_issue.html_id)
                     if vul then
                       establish_consistency(vul, sidfmd_issue, sidfmd_issue.issue, [], {}, {})
                     end
                     sidfmd_issue.change_status_with_journal(ADMIN_USER.id, SidfmdIssue::STATUS_FIXED, NOTE_NO_LONGER_DETECTED)
                     deleted_sidfmd_issue_ids << sidfmd_issue.id
                   end
        vuls.uniq{|vul| vul.id}.each do |vul|
          checked_host_html_ids[host_id] = [] unless checked_host_html_ids[host_id] 
          checked_host_html_ids[host_id] << vul.html_id
          sidfmd_issues = SidfmdIssue.where(:sidfmd_host_id => host_id).where(:html_id => vul.html_id)
          if sidfmd_issues.empty? then
            if vul.flg_enable then
              #新規登録 登録が無ければ登録する
              issue = Issue.create(
                tracker_id: VM_TRACKER.id,
                project_id: VM_PROJECT.id,
                subject: "[#{vul.hostname}]#{vul.title}",
                description: '新規脆弱性登録',
                author_id: ADMIN_USER.id,
                priority_id: DEFAULT_PRIORITY.id,
                status_id: SidfmdIssue::STATUS_NOTCONFIRMED
              )
              sidfmd_issue = SidfmdIssue.create(
                issue_id: issue.id,
                sidfmd_host_id: host_id,
                html_id: vul.html_id, 
                title: vul.title,
                area_id: vul.area_id,
                risk_id: vul.risk_id,
                short_description: vul.short_description
              )
              new_sidfmd_issue_ids << sidfmd_issue.id
  
              new_cves = []
              SidfmdLocalVulCve.where(:sidfmd_local_vul_id => vul.id).each do |cve|
                new_cves << SidfmdIssueCve.new(
                  sidfmd_issue_id: sidfmd_issue.id,
                  cve_id: cve.cve_id,
                  c3_vector:cve.c3_vector,
                  c3_base_score: cve.c3_base_score,
                  c3_create_datetime: cve.created_at,
                  c3_update_datetime: cve.updated_at,
                  c3_exploitable: cve.c3_exploitable
                )
              end
              SidfmdIssueCve.import(new_cves, validate: false)
              
              new_sris = []
              SidfmdLocalVulSri.where(:sidfmd_local_vul_id => vul.id).each do |sri|
                new_sris << SidfmdIssueSri.new(
                  sidfmd_issue_id: sidfmd_issue.id,
                  category_id: sri.category_id,
                  risk_id: sri.risk_id,
                  memo: sri.memo,
                ) 
              end
              SidfmdIssueSri.import(new_sris, validate: false)
            
              products = product_related_vuls.select{|p| p.html_id == vul.html_id}
               
              #手動登録プロダクトの場合はSidfmdIssueAffprodもしくはSidfmdIssueAffverを登録
              new_affvers = []
              new_affprods = []
              pids = []
              vids = []
              products.each do |product|
                if product.vid && product.vid != 0 then
                  unless vids.include?(product.vid) then
                    vids << product.vid
                    v_info = versions_info[product.vid]
                    new_affvers << SidfmdIssueAffver.new(
                      sidfmd_issue_id: sidfmd_issue.id,
                      product_version_id: product.vid,
                      product_id: product.pid,
                      fullname: v_info[:fullname],
                      version: v_info[:version]
                    )
                  end
                else
                  unless pids.include?(product.pid) then
                    pids << product.pid
                    new_affprods << SidfmdIssueAffprod.new(
                      sidfmd_issue_id: sidfmd_issue.id,
                      product_id: product.pid,
                      name: prods_info[product.pid][:name]
                    )
                  end
                end
              end
              SidfmdIssueAffver.import(new_affvers, validate: false) if new_affvers.any?
              SidfmdIssueAffprod.import(new_affprods, validate: false) if new_affprods.any?
            end
          else 
            #更新
            products = product_related_vuls.select{|p| p.html_id == vul.html_id}
            sidfmd_issues.each do |sidfmd_issue|
              issue = sidfmd_issue.issue
              is_adding_mod_ids_necessary = establish_consistency(vul, sidfmd_issue, issue, products, prods_info, versions_info)
              if vul.flg_enable && vul.updated_at > issue.updated_on && issue.status_id == SidfmdIssue::STATUS_FIXED then
                #対策済みを未対策に変更
                sidfmd_issue.change_status_with_journal(ADMIN_USER.id, SidfmdIssue::STATUS_NOTCONFIRMED, NOTE_CHANGE_PRODVER)
                modified_sidfmd_issue_ids << sidfmd_issue.id
                is_adding_mod_ids_necessary = false
              elsif !vul.flg_enable && issue.status_id != SidfmdIssue::STATUS_NOTAFFECT then
                #影響無に変更
                sidfmd_issue.change_status_with_journal(ADMIN_USER.id, SidfmdIssue::STATUS_NOTAFFECT, NOTE_DISABLED)
                deleted_sidfmd_issue_ids << sidfmd_issue.id
                is_adding_mod_ids_necessary = false
              elsif vul.flg_enable && issue.status_id == SidfmdIssue::STATUS_NOTAFFECT then
                #無効から有効になったものを未対策に変更
                # FIX:無効から未対策への変更はしない。
                #sidfmd_issue.change_status_with_journal(ADMIN_USER.id, SidfmdIssue::STATUS_NOTCONFIRMED, NOTE_ENABLED)
                #modified_sidfmd_issue_ids << sidfmd_issue.id
              end
              modified_sidfmd_issue_ids << sidfmd_issue.id if is_adding_mod_ids_necessary
            end
          end
        end
      rescue Exception => e
        logger.info(e)
      end
    end
  
    #検出されなくなった未対策・対策中のものを対策済に変更
    noaffect_sidfmd_issues = SidfmdIssue.where("html_id >= #{MIN_HTMLID}")
                                        .where(:sidfmd_host_id => host_ids)
                                        .where(
                                          :issue_id => Issue.where(
                                                         :status_id => [ SidfmdIssue::STATUS_NOTCONFIRMED, SidfmdIssue::STATUS_ONGOING ]
                                                       ).select(:id)
                                        )
                                        .select{|i| 
                                          !checked_host_html_ids[i.sidfmd_host_id] || 
                                          !checked_host_html_ids[i.sidfmd_host_id].include?(i.html_id)
                                        }
    if noaffect_sidfmd_issues.any? then
      vuls = SidfmdLocalVul.where(:html_id => noaffect_sidfmd_issues.map{|i| i.html_id})
                           .map{|vul| [ vul.html_id, vul ]}
                           .to_h
      noaffect_sidfmd_issues.each do |sidfmd_issue|
        issue = sidfmd_issue.issue
        vul = vuls[sidfmd_issue.html_id]
        if vul then
          establish_consistency(vul, sidfmd_issue, issue, [], {}, {})
        end
        if issue.status_id == SidfmdIssue::STATUS_NOTCONFIRMED || issue.status_id == SidfmdIssue::STATUS_ONGOING then 
          sidfmd_issue.change_status_with_journal(ADMIN_USER.id, SidfmdIssue::STATUS_FIXED, NOTE_NO_LONGER_DETECTED)
          deleted_sidfmd_issue_ids << sidfmd_issue.id
        end
      end
    end

    return {
      :add => new_sidfmd_issue_ids,
      :modify => modified_sidfmd_issue_ids,
      :delete => deleted_sidfmd_issue_ids
    }
  end

  #リンクのキーワードをリンクに変換する 変換対象のテキストと、エラーメッセージ表示用の項目名を渡す
  def self.get_link_converted_text(text, item_name = '')
    dump = {}
    is_hit = true
    start_idx = 0
    prefix = "%LINK:"
    item_name = '「' + item_name + '」'
    while is_hit do
      hit_idx = text.index(prefix, start_idx)
      if hit_idx then
        hit_last_idx = text.index('%', hit_idx + prefix.size - 1)
      else
        break
      end
      if hit_last_idx then
        match_str = text.slice(hit_idx..hit_last_idx)
        is_match = match_str.match(/#{prefix}.+\$\$\$.+%/)
      else
        break
      end
      if is_match then
        if match_str.index(/<|>|"|\|&/) then
          dump[:err] = 'リンクに<>"\&は使用できません。' + item_name + 'を確認して下さい。'
          break
        end
      else
        start_idx = hit_last_idx + 1
        next
      end
      str_link_top_idx = match_str.index('$$$') + 3
      link_name = match_str.slice(prefix.size..(str_link_top_idx - ('$$$'.size) - 1))
      link = match_str.slice(str_link_top_idx.. match_str.size - 2)
      begin
        p_link = URI.parse(link)
      rescue Exception => e
        logger.info('invalid link: ' + link)
        dump[:err] = '入力したリンクに誤りがあります。' + item_name + 'を確認して下さい。'
        break
      end
      if link =~ /^http|^https|^ftp/ then
        text = (hit_idx == 0 ? '' : text.slice(0..(hit_idx - 1))) +
               '<a href="' + link + '" target="_blank">' + link_name + '</a>' +
               (hit_last_idx + 1 == text.size ? '' : text.slice((hit_last_idx + 1)..-1))
        start_idx = hit_last_idx + 1
      else
        dump[:err] = 'プロトコルはhttp/https/ftpのみ使用可能です。' + item_name + 'を確認して下さい。'
        break
      end
    end
    text = text.gsub("\n", "<br />")
    return { :text => text, :err => dump[:err] }
  end

  def escape_special_chars(text)
    text = text.gsub(/<|>|"|'|&/, '<' => '&lt;', '>' => '&gt;', '"' => '&quot;', "'" => '&#39;', '`' => '&#x60;', '&' => '&amp;')
  end

  # 脆弱性詳細画面の情報を返す
  def information_content
    local_vuln = self
    vuln = {}
    vuln[:id] = local_vuln.html_id
    vuln[:title] = local_vuln.title
    vuln[:area_id] = local_vuln.area_id
    vuln[:create_datetime] = local_vuln.created_at
    vuln[:update_datetime] = local_vuln.updated_at
    vuln[:risk_id] = local_vuln.risk_id
    #影響を受ける製品
    affect_products = []
    affect_versions = {}
    affect_lproducts = []
    affect_lversions = {}
    pushed_pid = []
    pushed_lpid = []
    local_vuln.sidfmd_local_vul_apps.each do |app|
      pid = app.pid
      vid = app.vid
      prod = SidfmdProduct.find_by(:pid => pid)
      next unless prod
      if !pushed_pid.include?(pid) then
        affect_products << {
          :product => {:id => pid, :name => prod.name},
          :create_date => prod.create_datetime,
          :update_date => prod.update_datetime
        }
        pushed_pid << pid
      end
      if vid != 0 then #vid != 0 はバージョン
        version = SidfmdVersion.find_by(:vid => vid)
        next unless version
        affect_versions[pid] = [] unless affect_versions[pid]
        affect_versions[pid] << version.version
      end
    end
    local_vuln.sidfmd_local_vul_lapps.each do |lapp|
      pid = lapp.sidfmd_lproduct_id
      vid = lapp.sidfmd_lversion_id
      lprod = SidfmdLproduct.find_by(:id => pid)
      next if !lprod || !lprod.flg_enable
      if !pushed_lpid.include?(pid) then
        affect_lproducts << {
          :product => {:lproduct_id => pid, :name => lprod.name},
          :create_date => lprod.created_at,
          :update_date => lprod.updated_at
        }
        pushed_lpid << pid
      end
      if vid != 0 then #vid != 0 はバージョン
        version = SidfmdLversion.find_by(:id => vid)
        next unless version
        affect_lversions[pid] = [] unless affect_lversions[pid]
        affect_lversions[pid] << version.version
      end
    end
    vuln[:affect_products] = (affect_products + affect_lproducts).sort_by{|p| p[:product][:name].downcase}
    #content
    vuln[:content] = {}
    #影響を受ける製品
    if vuln[:affect_products].any? then
      affect_prods_elm = '<ul>'
      vuln[:affect_products].each do |p|
        version_names = []
        if p[:product][:id] then
          prod_versions = affect_versions[p[:product][:id]]
          version_names = prod_versions if prod_versions
        else
          lprod_versions = affect_lversions[p[:product][:lproduct_id]]
          version_names = lprod_versions if lprod_versions
        end
        product = '<li class=\"d1\">' + "\n" + '<p><b>' + escape_special_chars(p[:product][:name]) + "</b></p>\n"
        versions = ''
        version_names.each do |v_name|
          versions += '<p>' + escape_special_chars(v_name) + "</p>\n"
        end
        product += versions + "</li>\n"
        affect_prods_elm += product
      end
      affect_prods_elm += '</ul>'
    else
      affect_prods_elm = ''
    end
    vuln[:content][:affect] = affect_prods_elm
    #概要
    description = escape_special_chars(local_vuln.description)
    description = SidfmdLocalVul.get_link_converted_text(description)
    if description[:err] then #エラーは登録時に確認するため、通常は発生しない
      logger.info('converting link is failed: ' + description[:err])
      description = local_vuln.description
    else
      description = description[:text]
    end
    vuln[:content][:description] = '<p>' + description + '</p>'
    #影響などの確認方法
    verification  = escape_special_chars(local_vuln.confirmation)
    verification  = SidfmdLocalVul.get_link_converted_text(verification)
    if verification[:err] then
      logger.info('converting link is failed: ' + verification[:err])
      verification = local_vuln.confirmation
    else
      verification = verification[:text]
    end
    vuln[:content][:verification] = '<p>' + verification + '</p>'
    #対処方法
    solution = escape_special_chars(local_vuln.workaround)
    solution = SidfmdLocalVul.get_link_converted_text(solution)
    if solution[:err] then
      logger.info('converting link is failed: ' + solution[:err])
      solution = local_vuln.workaround
    else
      solution = solution[:text]
    end
    vuln[:content][:solution] = solution
    #関連情報
    related_info = escape_special_chars(local_vuln.related)
    related_info = SidfmdLocalVul.get_link_converted_text(related_info)
    if related_info[:err] then
      logger.info('converting link is failed: ' + related_info[:err])
      related_info = local_vuln.related
    else
      related_info = related_info[:text]
    end
    vuln[:content][:link] = related_info
    #更新履歴
    logs = local_vuln.sidfmd_local_vul_logs.order(:id)
    if logs.any? then
      history = '<ul>'
      logs.each do |log|
        conv_log = escape_special_chars(log.mes)
        conv_log = SidfmdLocalVul.get_link_converted_text(conv_log)
        log = (conv_log[:err] ? log : conv_log[:text])
        history += '<li>' + "\n" + '<p><font color=\"#000000\">' + log + "</font></p>\n" + "</li>\n"
      end
      history += '</ul>'
    else
      history = ''
    end
    vuln[:content][:history] = history
    #CVES
    cves = []
    local_vuln.sidfmd_local_vul_cves.each do |cve|
      cves << {
        :cve_id => cve.cve_id,
        :cvss3 => {
          :vector => cve.c3_vector,
          :base_score => cve.c3_base_score ? cve.c3_base_score : -1,
          :create_datetime => cve.created_at,
          :update_datetime => cve.updated_at,
          :exploitable => cve.c3_exploitable
        }
      }
    end
    vuln[:cves] = cves
    #sris
    sris = []
    local_vuln.sidfmd_local_vul_sris.order(:category_id).each do |sri|
      sris << {
        :category_id => sri.category_id,
        :risk_id => sri.risk_id,
        :memo => (sri.memo.size > 0 ? sri.memo : nil)
      }
    end
    vuln[:sris] = sris
    vuln
  end

  # 影響を受ける製品名を返す
  def affect_product_names
    ret = SidfmdProduct.where(
      :pid => sidfmd_local_vul_apps.select(:pid) 
    ).order('LOWER(name COLLATE "C")')
     .pluck(:name) +
    SidfmdLproduct.where(
      :id => sidfmd_local_vul_lapps.select(:sidfmd_lproduct_id)
    ).order('LOWER(name COLLATE "C")')
     .pluck(:name)
=begin
    local_vul_lapps = sidfmd_local_vul_lapps
    ret += SidfmdLproduct.where(
      :id => local_vul_lapps.where('sidfmd_local_vul_lapps.sidfmd_lversion_id IS NULL OR sidfmd_local_vul_lapps.sidfmd_lversion_id = 0')
                            .select('sidfmd_local_vul_lapps.sidfmd_lproduct_id')

    )
    .pluck(:name)

    if version_necessary then
      ret += SidfmdLversion.where(
        :id => local_vul_lapps.where('sidfmd_local_vul_lapps.sidfmd_lversion_id IS NOT NULL AND sidfmd_local_vul_lapps.sidfmd_lversion_id != 0')
                              .select('sidfmd_local_vul_lapps.sidfmd_lversion_id')
      )
      .pluck(:fullname)
    end
    ret = ret.sort_by{|name| name.downcase }.uniq
=end
  end

end
